#include "DecodeManager.h"

#include <unistd.h>

#include <QDebug>
#include <QDir>
#include <QSemaphore>
#include <QThread>
#include <QTime>

#include "histmanager.h"
#include "DecodeWorker.h"
#include "config/adcconfig.h"
#include "adcm_mem.h"
#include "adcm_dec.h"
#include "adcm_df.h"
#include "adcm_ev.h"
#include "adcm_ids.h"
#include "timer.h"
#include "storage.h"
#include "adcm_el.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_types.h"
#include "EventListFiller.h"

const int DecodeManager::numberWorkers = QThread::idealThreadCount ();
//const int DecodeManager::numberWorkers = 1;

#define PREFIX "DecodeManager: "

namespace {
const int updateHistogramsTime = 700; // in ms

const int MIN_EVENTS_PACK_SIZE = 100;
const int EVENTS_PACK_SIZE = 1000;
const int EVENT_LIST_SIZE = EVENTS_PACK_SIZE*DecodeManager::numberWorkers*5;
const int RUN_LOOP_PERIOD_MS = 100;
const int RUN_LOOP_TIMER_MS = 1000;
//const size_t DECODE_MAX_EVENTS_ONCE = 100;
//const size_t DECODE_MAX_WORDS_ONCE = 1024*256; // 1MByte
//const size_t DECODE_MAX_FRAMES_ONCE = 1000;

// 4 threads, full mode, 1000 frames; debug:up to 100ms, release:up to 50ms
//const uint waitWorkersTimeoutframesPackSize =EVENTS_PACK_SIZE/DecodeManager::numberWorkers;
}

DecodeManager::DecodeManager(QObject *parent):
    QObject(NULL),
    paused(false),
    channalNumber(ADCM_N_CH),
    adcConfig(new AdcConfig),
    storageConfig(new StorageConfig),
    decoderConfig(new DecoderConfig),
    hmgr(new HistManager(this)),
    counters(new adcm_counters_t),
    stor(new storage),
    idleWorkers(new QSemaphore(numberWorkers)),
    freeEvents(new QSemaphore(EVENT_LIST_SIZE)),
    managerEvents(new QSemaphore(0)),
    pauseSem(new QSemaphore(0)),
    lastEventListSize(EVENT_LIST_SIZE),
    eventList(new adcm_el_t(freeEvents, managerEvents, EVENT_LIST_SIZE)), // 4*numberWorkers*EVENTS_PACK_SIZE
    summedDecStat(new adcm_decstat_t),
    lastHistSumTime(new QTime),
    workStatTimer(new QTime)
{
    Q_ASSERT(numberWorkers>0);
    qRegisterMetaType<adcm_decstat_t>();
    qRegisterMetaType<EventListIt>();
    for(int i=0; i<numberWorkers; ++i){
        workers.append(new DecodeWorker(i, idleWorkers, stor, decoderConfig, eventList, storageConfig->enable));
        workerThreads.append(new QThread(this));

        workers[i]->moveToThread(workerThreads[i]);
        workerThreads[i]->start();
        workerThreads[i]->setPriority(QThread::LowestPriority);

        if(i==0){
            qRegisterMetaType<AdcmEvent>();
            connect(workers[i], SIGNAL(monitorEvent(AdcmEvent)), parent, SLOT(displayDecodedEvent(AdcmEvent)));
        }

        connect(workers[i], SIGNAL(statisticsUpdated(adcm_decstat_t, int)), SLOT(addWorkerStatistics(adcm_decstat_t, int)));
        qRegisterMetaType<CalZeroConfig>();
        connect(workers[i], SIGNAL(decodedCalZero(CalZeroConfig)), parent, SLOT(addCalZero(CalZeroConfig)));
        qRegisterMetaType<CalAvgConfig>();
        connect(workers[i], SIGNAL(decodedCalAvg(CalAvgConfig)), parent, SLOT(addCalAvg(CalAvgConfig)));
    }
    connect(hmgr, SIGNAL(histListModified(const HistList &)), SIGNAL(histListModified(const HistList &)));

//    connect (&updateHistogramsTimer, SIGNAL (timeout ()), SLOT (update_mainhist_view ()));
//    updateHistogramsTimer.start (updateHistogramsTime);

    qDebug()<<QString("Decode manager has %1 worker(s); eventList size=%2; max pack size=%3; min pack size=%4")
              .arg(numberWorkers).arg(EVENT_LIST_SIZE)
              .arg(EVENTS_PACK_SIZE).arg(MIN_EVENTS_PACK_SIZE);

    lastHistSumTime->start();
    workStatTimer->start();
}


DecodeManager::~DecodeManager()
{
    paused = true;
    foreach(QThread *workerThread, workerThreads){
        workerThread->quit();
    }
    foreach(QThread *workerThread, workerThreads){
        if(!workerThread->wait(100/numberWorkers))
            qDebug()<<"worker aborted";
        delete workerThread;
    }
    workerThreads.clear();

    delete adcConfig;
    delete storageConfig;
    delete decoderConfig;
    delete hmgr;
    delete counters;
    delete stor;

    delete idleWorkers;
    delete freeEvents;
    delete managerEvents;
    delete eventList;
    delete lastHistSumTime;
    delete workStatTimer;
}


EventListFiller * DecodeManager::createFiller()
{
    EventListFiller * filler = new EventListFiller(eventList, this);
    connect(filler, SIGNAL(statisticsUpdated(adcm_decstat_t)), SLOT(addWorkerStatistics(adcm_decstat_t)));
    return filler;
}


void DecodeManager::pause()
{
    paused = true;
    pauseSem->acquire(); // Enshure that loop is not running
    idleWorkers->acquire(numberWorkers); // Wait worker finished their jobs
    idleWorkers->release(numberWorkers);
    clenUpCompletedJobs();
    if(!currentJobs.empty()){
        qDebug()<<"missing events detected";
        info();
        Q_ASSERT(false);
    }
}


void DecodeManager::resume()
{
    paused = false;
    QMetaObject::invokeMethod(this, "loop", Qt::QueuedConnection);
}


void DecodeManager::setAdcConfMem(const conf_mem_t &conf_mem)
{
    idleWorkers->acquire(numberWorkers);

    foreach(DecodeWorker *worker, workers)
        worker->setAdcConfMem(conf_mem);
    hmgr->setAdcConfMem(conf_mem);
    channalNumber = conf_mem.nch;
    clearCounters();

    idleWorkers->release(numberWorkers);
}


void DecodeManager::reset()
{
    pause();

    int eventSize = managerEvents->available();
    if(eventSize){
        managerEvents->acquire(eventSize);
        freeEvents->release(eventSize);

        Q_ASSERT(eventList->getStartWorkerPos()==eventList->getEndWorkerPos());

        EventListIt endIt=eventList->getEndWorkerPos();
        eventList->next(endIt, eventSize);
        eventList->setStartWorkerPos(endIt);
        eventList->setEndWorkerPos(endIt);
        Q_ASSERT(eventList->getManagerEvNum()==endIt->first);
    }

    workersDecStat.clear();
    summedDecStat->clear();
    hmgr->clear();
    stor->reset ();

    if(storageConfig->enable)
        writeChMap ();

    clearCounters();

    checkEvListSize();
    foreach(DecodeWorker *worker, workers){
        worker->reset();
    }

    resume();
}


void DecodeManager::clear()
{
    pause();
    foreach(DecodeWorker *worker, workers){
        worker->lockWorker();
        worker->clear();
        worker->unlockWorker();
    }

    workersDecStat.clear();
    summedDecStat->clear();
    hmgr->clear();

    endBurst();
    resume();
}


void DecodeManager::loop()
{
    if(paused){
        pauseSem->release();
        return;
    }
    workStatTimer->restart();
    uint loopCount=0;
    while(!paused){
        ++loopCount;

        if(!idleWorkers->tryAcquire(1, RUN_LOOP_PERIOD_MS)){
            qDebug()<<"Workers work too long";
            if(workStatTimer->elapsed() > RUN_LOOP_TIMER_MS)
                break;
            continue;
        }
        idleWorkers->release();

        if( paused || ((!startJob() || loopCount%10 == 0) &&
                       (workStatTimer->elapsed() > RUN_LOOP_TIMER_MS)) ){
                break;
        }
    }

//    qDebug()<<"loop count:"<<loopCount;
//    info();
//    if(!paused) checkEvListSize();

    if(histEnabled && lastHistSumTime->elapsed()>updateHistogramsTime){
        hmgr->lock();
        hmgr->clear();
        foreach (DecodeWorker* worker, workers){
            worker->lockWorker();
            hmgr->addHist(worker->getHistManager());
            worker->unlockWorker();
        }
        hmgr->unlock();
        lastHistSumTime->restart();
    }

    emit summedStatisticsUpdated(*summedDecStat);
    summedDecStat->clear();

    QMetaObject::invokeMethod(this, "loop", Qt::QueuedConnection);
}


void DecodeManager::setAdcConfig(const AdcConfig &r)
{
    idleWorkers->acquire(numberWorkers);
    *adcConfig = r;
    foreach(DecodeWorker *worker, workers){
        worker->setAdcConfig(*adcConfig);
    }
    hmgr->setAdcConfig(*adcConfig);
    idleWorkers->release(numberWorkers);
}


void DecodeManager::setDecoderConfig(const DecoderConfig &r)
{
    idleWorkers->acquire(numberWorkers);
    *decoderConfig = r;
    foreach(DecodeWorker *worker, workers)
        worker->decoderConfigModified(r);
    idleWorkers->release(numberWorkers);
}


void DecodeManager::setStorageConfig(const StorageConfig &r)
{
    idleWorkers->acquire(numberWorkers);
    *storageConfig = r;
    const QString newFileName = storageConfig->bn + QDir::separator() + storageConfig->fn;
    stor->setFileName(newFileName);
    stor->setopt(storageConfig->enable, storageConfig->mode);
    stor->reset ();

    foreach(DecodeWorker *worker, workers)
        worker->setStorageConfig(storageConfig->enable);
    idleWorkers->release(numberWorkers);
}


bool DecodeManager::appendChannelCounters(const adcm_counters_t &cnt)
{
    int nch = std::min(counters->n, cnt.n);
    for (int i=0; i<nch; i++)
        counters->rawhits[i] += cnt.rawhits[i];
    counters->time += cnt.time;
    return true;
}


bool DecodeManager::endBurst()
{
    stor->lock();
    bool ret = true;
    quint16 d;
    d = STOR_ID_CNTR;
    ret &= stor->write (&d, sizeof(d)); // write type
    d = 2 * sizeof(d) + sizeof(adcm_counters_t);
    ret &= stor->write (&d, sizeof(d)); // write total length=size(type)+size(length)+size(data)
    ret &= stor->write (counters, sizeof(adcm_counters_t)); // write data

    /*
  {
    int i;
    double t;

    t = 1e-8 * ts_last;
    for (i = 0; i < ADC_NCH; i++)
      if (cnt->rawhits[i])
        printf ("Hits [%2d]: %10u   %10.3f kHz\n", i, cnt->rawhits[i], 1e-3 * cnt->rawhits[i] / t);
  }
*/

    stor->flush();
    writeChMap();
    stor->unlock();

    clearCounters();

    return ret;
}


void DecodeManager::clearHist(int index)
{
    hmgr->clear(index);
    foreach (DecodeWorker* worker, workers){
        worker->lockWorker();
        worker->getHistManager()->clear(index);
        worker->unlockWorker();
    }
}


void DecodeManager::setHistEnabled(bool val)
{
    histEnabled = val;
    foreach (DecodeWorker* worker, workers) {
        worker->setHistEnabled(val);
    }
}


// Gets workers and filler stats
void DecodeManager::addWorkerStatistics(const adcm_decstat_t &stat, int index)
{
    summedDecStat->add(stat);
//    qDebug()<<"Got stat:"<<index;
//    if(workersDecStat.contains(index)) {
//        qDebug()<<"Double statistic for worker:"<<index;
//        workersDecStat[index].add(stat);
//    } else
//        workersDecStat[index] = stat;

//    if(workersDecStat.size() == numberWorkers+1) {
//        adcm_decstat_t summaryStat;
//        foreach (adcm_decstat_t var, workersDecStat.values()) {
//            summaryStat.add(var);
//        }
//        workersDecStat.clear();
//        emit summedStatisticsUpdated(summaryStat);
//    }
}


void DecodeManager::clearCounters()
{
    counters->clear();
    counters->n=channalNumber;
}


bool DecodeManager::startJob()
{
    bool noDelay=true;
    uint eventSize;
//    info();

    if(managerEvents->tryAcquire(MIN_EVENTS_PACK_SIZE, RUN_LOOP_PERIOD_MS)){
        eventSize = std::min(managerEvents->available(), EVENTS_PACK_SIZE-MIN_EVENTS_PACK_SIZE);
        managerEvents->acquire(eventSize);
        eventSize += MIN_EVENTS_PACK_SIZE;
    } else {
        clenUpCompletedJobs();
        eventSize = managerEvents->available();
        if(eventSize==0){
//            info();
            return false;
        }
        managerEvents->acquire(eventSize);
        noDelay = false;
    }

    Q_ASSERT(idleWorkers->available());
    idleWorkers->acquire();

    DecodeWorker* idleWorker=getIdleWorker();
    Q_ASSERT(idleWorker);
    idleWorker->lockWorker();

    EventListIt startIt=eventList->getEndWorkerPos();
    EventListIt endIt=startIt;
    eventList->next(endIt, eventSize);

    if(noDelay) clenUpCompletedJobs(idleWorker->getWorkerIndex());

    if(eventList->getStartWorkerPos() != startIt){
        Q_ASSERT(currentJobs.size());
        bool found=false;
        foreach (WorkerJob job, currentJobs) {
            if(job.endPos==startIt){
                found=true;
                break;
            }
        }
        Q_ASSERT(found);
    }
    eventList->setEndWorkerPos(endIt);
    currentJobs.append(WorkerJob(idleWorker->getWorkerIndex(), startIt, endIt, eventSize));
//    qDebug()<<"Add job #"<<idleWorker->getWorkerIndex()<<":" << eventSize << "; " << startIt->first << "-" << endIt->first;
    Q_ASSERT(startIt->first<=0xFFFF);

//    info();
//    qDebug()<<"\n\n\n";
    QMetaObject::invokeMethod(idleWorker, "decodeEventList",
                              Qt::QueuedConnection,
                              Q_ARG(EventListIt, startIt),
                              Q_ARG(EventListIt, endIt));
    return noDelay;
}


void DecodeManager::clenUpCompletedJobs(int index)
{
    uint workerStatus=0;
    foreach (DecodeWorker *worker, workers) {
        if((int)worker->getWorkerIndex()==index || worker->isIdle())
            workerStatus |= 1<<worker->getWorkerIndex();
    }
    EventListIt workerPos = eventList->getStartWorkerPos();
    int completedEvents=0;
    for(int i=0;i<currentJobs.size();){
        WorkerJob &job = currentJobs[i];
//        if(index>=0 && job.workerIndex == (uint)index){
//            job.completed = true;
//        } else {
            if(1<<job.workerIndex & workerStatus)
                job.completed = true;
//        }


        if(job.startPos==workerPos && job.completed){
            workerPos=job.endPos;
            completedEvents += job.size;
//            qDebug()<<"\t\t\t\tClear   #"<<job.workerIndex<<":"<<job.startPos->first<<"-"<<job.endPos->first
//                   <<" "<< job.size;
            currentJobs.removeAt(i);
            continue;
        }
        ++i;
    }

    if(completedEvents){
        //        qDebug()<<"================";
        //        info();

        eventList->setStartWorkerPos(workerPos);
//        qDebug()<<"\t\t\t\tStart worker:"<<workerPos->first;
        freeEvents->release(completedEvents);
        Q_ASSERT(eventList->getStartWorkerPos() == workerPos);
        //        qDebug()<<"Clean size:"<<completedEvents;
        //        info();

        //        qDebug()<<"================";
    }
    Q_ASSERT(currentJobs.size()==0 || currentJobs[0].startPos == workerPos);
}


void DecodeManager::info()
{
    EventListIt sW = eventList->getStartWorkerPos();
    EventListIt eW = eventList->getEndWorkerPos();
    uint mP = eventList->getManagerEvNum();
    uint begin = eventList->getBeginEvNum();
    qDebug()<<QString("eventList begin:%1; workers:%2-%3; manager:%4")
              .arg(begin)
              .arg(sW->first).arg(eW->first)
              .arg(mP);
    int m = managerEvents->available();
    int f = freeEvents->available();
    int jobs=0;
    if(currentJobs.size()){
    foreach (WorkerJob job, currentJobs) {
        jobs += job.size;
        qDebug()<<QString("job #%1: %2-%3 %4").arg(job.workerIndex)
                  .arg(job.startPos->first).arg(job.endPos->first)
                  .arg(job.size);
    }
    } else {
        qDebug()<<"No pending jobs";
    }
    qDebug()<<QString("man ev:%1; free ev:%2; jobs:%3; sum:%4")
              .arg(m).arg(f).arg(jobs).arg(m+f+jobs);
}


void DecodeManager::checkEvListSize()
{
    return;
    {
    QList<WorkerJob> jobs = currentJobs;
    EventListIt sp = eventList->getStartWorkerPos();
    EventListIt ep = eventList->getEndWorkerPos();
    while(jobs.size()){
        int i;
        for(i=0;i<jobs.size();++i){
            if(sp==jobs[i].startPos){
                sp=jobs[i].endPos;
                break;
            }
        }
        if(i==jobs.size()){
            info();
            qDebug()<<"NotFound";
            break;
        } else {
            jobs.removeAt(i);
        }
    }
    if(sp!=ep){
        info();
        qDebug()<<"chain is broken";
    }
    }

    uint freeSize=0;
    freeSize=std::max(freeEvents->available()-10,0);
    while(!freeEvents->tryAcquire(freeSize)){
        freeSize-=10;
        if(freeSize<=0){
            freeSize=0;
            break;
        }
    }
    while(freeEvents->tryAcquire()){
        ++freeSize;
    }
    usleep(10);
    eventList->flushManEvents();
    uint manSize=managerEvents->available();
    uint jobs=0;
    foreach(WorkerJob job, currentJobs){
        jobs+=job.size;
    }
    uint sum = freeSize+manSize+jobs;
    if(qAbs(sum-lastEventListSize)>0){
        qDebug()<<"================"<<lastEventListSize<<"-"<<sum<<"="<<lastEventListSize-sum<<"================";
        lastEventListSize=sum;
    }
    freeEvents->release(freeSize);
}


DecodeWorker* DecodeManager::getIdleWorker()
{
    for(QVector<DecodeWorker*>::Iterator it=workers.begin();it!=workers.end();++it)
        if((*it)->isIdle())
            return *it;
    return NULL;
}


void DecodeManager::writeChMap()
{
    adcm_cmap_t chMap;
    memset (&chMap, 0, sizeof (adcm_cmap_t));

    chMap.n = adcConfig->size();
    int chMapSize = sizeof(quint32) + chMap.n*sizeof(quint8);
    for (uint c = 0; c < chMap.n; c++) {
        chMap.map[c] = adcConfig->trig[c];
        if(decoderConfig->mode == DM_ZERO)
            chMap.map[c] |= ADCM_CCTRL_BIT_TEST;
        else
            chMap.map[c] &= ~ADCM_CCTRL_BIT_TEST;
    }

    bool ret = true;
    quint16 d;
    d = STOR_ID_CMAP;

    ret &= stor->write (&d, sizeof(d));
    d = 2 * sizeof(d) + chMapSize;
    ret &= stor->write (&d, sizeof(d));
    ret &= stor->write (&chMap, chMapSize);
}
