#include "DecodeWorker.h"

#include <exception>

#include <QDebug>
#include <QTimer>
#include <QSemaphore>

#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"

#include "storage.h"
#include "adcm_df.h"
#include "adcm_el.h"
#include "adcm_ev.h"
#include "adcm_dec.h"
#include "adcm_raw.h"
#include "adcm_dws.h"
#include "adchardw.h"
#include "timer.h"
#include "histmanager.h"
#include "config/decoderconfig.h"
#include "decoder/waveformdecoderpeaks.h"
#include "decoder/waveformdecodersine.h"
#include "decoder/decoder_profiling_macros.h"


#define PREFIX "DecodeWorker: "

namespace {
const int monitorEventTimeMs = 100;
const int STOR_PH_SIZE = sizeof(stor_packet_hdr_t);
const int MAX_STOR_EV_SIZE = STOR_PH_SIZE + sizeof(stor_ev_hdr_t) + sizeof(stor_puls_t) * ADC_NCH;
}

//______________________________________________________________________________

DecodeWorker::DecodeWorker (int _workerIndex,
                  QSemaphore * const _idleWorkers,
                  storage * const _stro,
                  DecoderConfig * const decConf,
                  adcm_el_t * _eventList,
                  bool storEnabled) :
    QObject (NULL),
    workerIndex(_workerIndex),
    idleWorkers(_idleWorkers),
    stateMutex(new QMutex),
    stor(_stro),
    hmgr(new HistManager(this)),
    eventList(_eventList),
    decst(new adcm_decstat_t),
#ifdef FIT_SINE
    decoder(new WaveformDecoderSine),
#else
    decoder(new WaveformDecoderPeaks),
#endif /* FIT_SINE */
    monitorEventTime(new QTime()),
    statisticsTimer(new QTimer(this)),
    calResultTimer(new QTimer(this)),
    decodeMode(decConf->mode),
    storageEnabled(storEnabled),
    conf_mem(new conf_mem_t),
    newCalAvgConfig(new CalAvgConfig),
    newCalZeroConfig(new CalZeroConfig),
    prevEvTs(0)
{
    setObjectName("decoder thread handler");

    connect(statisticsTimer, SIGNAL(timeout()), SLOT(sendStatistics()));
    statisticsTimer->setInterval(1000);
    statisticsTimer->start();

    connect(calResultTimer,SIGNAL(timeout()), SLOT(sendCalResult()));
    calResultTimer->setInterval(300);
    if(decodeMode == DM_ZERO || decodeMode == DM_AVG )
        calResultTimer->start();
    monitorEventTime->start();
}

DecodeWorker::~DecodeWorker ()
{
    delete newCalZeroConfig;
    delete newCalAvgConfig;
    delete conf_mem;
    delete calResultTimer;
    delete statisticsTimer;
    delete monitorEventTime;
    delete decoder;
    delete decst;
    delete hmgr;
    delete stateMutex;
}


void DecodeWorker::makeCalZero(const AdcmEvent &ev)
{
    int n = conf_mem->nch;
    CalZeroConfig cal(n);
    cal.print();
    for (int ch = 0; ch < n; ch++) {
        const adcm_chbuf_t * const cbuf = ev.get_ev_chbuf_const(ch);
        if (!cbuf) continue;
        if (!(cbuf->flags & ADCM_CH_EVAL_RMS))
            continue;
        if (!(cbuf->flags & ADCM_CH_HAS_DATA))
            continue;
        cal.base[ch] += cbuf->dc;
        cal.rms[ch] += cbuf->rms;
        cal.nent[ch]++;
    }
    *newCalZeroConfig += cal;
}


void DecodeWorker::makeCalAvg(const AdcmEvent &ev)
{
    int n = conf_mem->nch;
    CalAvgConfig cal;
    cal.resize(n);
    for (int ch = 0; ch < n; ch++) {
        const adcm_chbuf_t * const cbuf = ev.get_ev_chbuf_const(ch);
        if (!cbuf) continue;

        if (!(cbuf->flags & ADCM_CH_HAS_THR)) {
//                  qWarning (QString("avg: not test and no threshold on ch %1, flags: 0x%2").arg(c).arg(cb->flags, 8, 16));
            continue;
        }
        if (cbuf->flags & ADCM_CH_HAS_CLIP)
            continue;
        if (cbuf->flags & ADCM_CH_HAS_PU)
            continue;
        const int len = cbuf->len();
        cal.data[ch].resize(len);
        for (int i = 0; i < len; i++)
            cal.data[ch][i] += cbuf->w[i];
        cal.nent[ch]++;
    }
    *newCalAvgConfig += cal;
}


int DecodeWorker::processEvent (AdcmEvent * ev)
{
    Q_ASSERT(ev);

    //    qDebug() << QString("process_event: job stages 0x%1").arg(job_flags, 8, 16);
    //    ev->print_event();

    decoder->decode(ev, decst);
    if (ev->getFlag(AdcmEvent::ADCM_EV_DT_CUT)) {
        return 0; // drop event
    }

    switch (decodeMode) {
    case DM_ZERO:
        makeCalZero(*ev);
        break;
    case DM_AVG:
        makeCalAvg(*ev);
        break;

    case DM_REC:
        ftimer_t t;
        if (DECODER_PROFILING) {
            timer_start (&t);
        }
        if(histEnabled){
            hmgr->fill (*ev);
            if(prevEvTs)
                hmgr->fillEvDt(*ev, prevEvTs);
            prevEvTs = ev->getEventTimestamp();
        }

        if (DECODER_PROFILING) {
            decst->prof.ta[decst_n_hist] += timer_read_clear (&t);
        }
        break;
    default:
        ;
    }
    if (DECODER_PROFILING) hmgr->fill_status (*decst);

//    if (ev->getEventNumber()<10) {
//        printf("-----------------------------------\n");
//        ev->print_event();
//        ev->print_flags();
//    }

    if (storageEnabled)
        writeEvent (*ev);

    return (1);
}

void DecodeWorker::sendStatistics()
{
//    requestForStat=false;
    emit statisticsUpdated(*decst, workerIndex);
    decst->clear();
}


void DecodeWorker::decodeEventList (EventListIt start, EventListIt end)
{
    EventListIt eventIt=start;
    AdcmEvent* ev;

    // Decode events from Event List
    while(eventIt != end){
        ev = eventIt->second;
        processEvent (ev);
        if(workerIndex==0){
            if(histEnabled && monitorEventTime->elapsed() > monitorEventTimeMs) {
                monitorEventTime->restart();
                emit monitorEvent(*ev);
            }
        }
        eventList->next(eventIt);
//        if(requestForStat) sendStatistics();
    }

//    sendStatistics();
    unlockWorker();
    idleWorkers->release();
}


void DecodeWorker::clear ()
{
    WaveformDecoder::adcm_decoder_clear_stat (decst);
    hmgr->clear();
}

void DecodeWorker::reset()
{
//    statisticsTimer->start();
}


void DecodeWorker::sendCalResult()
{
    if(decodeMode == DM_ZERO){
        emit decodedCalZero(*newCalZeroConfig);
        newCalZeroConfig->clear();
    } else if(decodeMode == DM_AVG){
        emit decodedCalAvg(*newCalAvgConfig);
        newCalAvgConfig->clear();
    } else
        qWarning("Callibration timer works in decoder thread, but it should be stopped");
}


bool DecodeWorker::isIdle()
{
    bool ret=false;
    if(stateMutex->tryLock()){
        ret=true;
        stateMutex->unlock();
    }
    return ret;
}


void DecodeWorker::lockWorker()
{
    stateMutex->lock();
}

void DecodeWorker::unlockWorker()
{
    stateMutex->unlock();
}

void DecodeWorker::setHistEnabled(bool val)
{
    histEnabled = val;
}

void DecodeWorker::pause()
{
    statisticsTimer->stop();
    calResultTimer->stop();
}

void DecodeWorker::resume()
{
    statisticsTimer->start();
    if(decodeMode == DM_ZERO || decodeMode == DM_AVG )
        calResultTimer->start();
    else
        calResultTimer->stop();
}

//void DecodeWorker::requestStats()
//{
//    requestForStat=true;
//    if(stateMutex->tryLock()){
//        sendStatistics();
//        stateMutex->unlock();
//    }
//}


bool DecodeWorker::writeEvent (const AdcmEvent &ev)
{
    unsigned char buf[MAX_STOR_EV_SIZE];
    stor_packet_hdr_t *ph = (stor_packet_hdr_t *) &buf[0];
    stor_ev_hdr_t *eh = (stor_ev_hdr_t *) &buf[STOR_PH_SIZE];

    ph->id = STOR_ID_EVNT;
    ph->size = STOR_PH_SIZE; // ph->size will be like curent write possition

    eh->ts = ev.getEventTimestamp(); // FIXME: 64 to 32 truncation
    eh->reserved1 = 0;
    eh->reserved2 = 0;
    eh->np = 0;
    ph->size += sizeof(stor_ev_hdr_t);

    for (unsigned int c = 0; c < ADC_NCH; c++) {
        const struct adcm_chbuf_t *cb = ev.get_ev_chbuf_const(c);
        if (!cb) continue;
        unsigned int required_flags = ADCM_CH_EVAL_BL | ADCM_CH_EVAL_INT | ADCM_CH_EVAL_TFAST;
        if ((cb->flags & required_flags) != required_flags)
            continue;
        {
            stor_puls_t *p = (stor_puls_t *) & buf[ph->size];
            p->ch = c;
            p->a = cb->i - cb->b; // Mean value relatively baseline
            p->t = cb->pt*ADC_REALSTEP*1e9; // Peak time, ns
            p->w = cb->pw*ADC_REALSTEP*1e9; // Peak width, ns
            ph->size += sizeof (stor_puls_t);
            eh->np++;
        }
    }
    return stor->write (buf, ph->size);
}


void DecodeWorker::setAdcConfig(const AdcConfig &adcConfig)
{
    decoder->adcConfigModified(adcConfig);
    hmgr->setAdcConfig(adcConfig);
}


void DecodeWorker::setAdcConfMem(const conf_mem_t &value)
{
    *conf_mem = value;
    decoder->setAdcConfMem(*conf_mem);
    hmgr->setAdcConfMem(*conf_mem);
}


void DecodeWorker::decoderConfigModified(const DecoderConfig &decConf)
{
    decodeMode = decConf.mode;
    decoder->decoderConfigModified(decConf);
    if(decodeMode == DM_ZERO || decodeMode == DM_AVG )
        calResultTimer->start();
    else
        calResultTimer->stop();
}


void DecodeWorker::setStorageConfig(bool enabled)
{
    storageEnabled = enabled;
}
