#include "EventListFiller.h"

#include <QDebug>
#include <QSemaphore>
#include <DecodeManager.h>

#include "adcmraw/adcm_raw.h"
#include "adcm_el.h"
#include "adcm_dws.h"
#include "adcm_ids.h"
#include "adcm_mem.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_types.h"
#include "timer.h"

namespace{
const int updateProgressTimeout = 300; // in ms
}

EventListFiller::EventListFiller(adcm_el_t * const evList, DecodeManager * const decodeManager) :
    QObject(NULL),
    eventList(evList),
    adcmDataBlockDecoder(new AdcmDataBlockDecoder(0)),
    fillerDecStat(new adcm_decstat_t),
    manager(decodeManager),
    lastProgressTime (new QTime),
    workTime(new QTime),
    statisticTimer(new QTimer)
{
    lastProgressTime->start();
    workTime->start();
    connect(statisticTimer, SIGNAL(timeout()), SLOT(sendStatistics()));
    statisticTimer->setInterval(1000);
    statisticTimer->start();
}

EventListFiller::~EventListFiller()
{
    delete adcmDataBlockDecoder;
    delete fillerDecStat;
    delete lastProgressTime;
    delete workTime;
    delete statisticTimer;
}

void EventListFiller::appendFrameList(const FrameList &frameList)
{
    if (DEBUG_PRINT_DATA_RATE){
        workTime->start();
    }

    quint64 bytesDecoded=0;
    framesDecoded=0;
    frameListSize=frameList.size();
    FrameList::const_iterator currentFrame = frameList.begin();
    FrameList::const_iterator end = frameList.end();

    updateProgress(false);
    while(currentFrame != end){
        appendFrame(*currentFrame);
        if (DEBUG_PRINT_DATA_RATE) bytesDecoded += currentFrame->size();
        ++currentFrame;
        if(++framesDecoded%500==0){
            updateProgress();
        }
    }
    eventList->flushManEvents();
    updateProgress(false);
    if (DEBUG_PRINT_DATA_RATE){
        double rate=workTime->elapsed();
        if(rate){
            rate=bytesDecoded/1024/1024 / (rate*1e-3);
        } else {
            rate=-1;
        }
        qDebug() << "Decoder throughput:" << rate << "MB/s";
    }

//    qDebug()<<"Frame list completed; size="<<frameList.size();
    sendStatistics();
    emit completed(frameList.seq);
}


void EventListFiller::setAdcConfMem(const conf_mem_t &conf_mem)
{
    manager->pause();
    bool val = conf_mem.master_board_id() == ADC64_BOARD_ID && conf_mem.boards()[0].hw.fw_rev <= 15256;
    adcmDataBlockDecoder->setTimeStampBitsFlag(val);
    manager->setAdcConfMem(conf_mem);
    manager->resume();
}


void EventListFiller::clear()
{
    fillerDecStat->clear();
    manager->clear();
}


void EventListFiller::reset()
{
    manager->reset();
    statisticTimer->start();
    eventList->reset();

    adcmDataBlockDecoder->reset();

    fillerDecStat->clear();

    lastProgressTime->restart();
    workTime->restart();
}

void EventListFiller::sendStatistics()
{
//    emit requestWorkerStatistic();
    emit statisticsUpdated(*fillerDecStat);
    emit eventsProceedUpdated(eventList->getEventsProceed());
    fillerDecStat->clear();
}


void EventListFiller::appendFrame(const QByteArray &buf)
{
    // Sort Frames to Data Blocks, fill Event List
#if DECODER_PROFILING
    ftimer_t timerRawDecode;
    timer_start (&timerRawDecode);
#endif
    decodeRawFrame(reinterpret_cast<const unsigned int *>(buf.constData()), buf.size()/4);
#if DECODER_PROFILING
    fillerDecStat->prof.ta[decst_n_rawdec] += timer_read_clear (&timerRawDecode);
#endif
//    qDebug() << QString("process_raw_events: %1 data block(s)").arg(ret);
}


int EventListFiller::decodeRawFrame(const unsigned int *frame, size_t flen)
{
    fillerDecStat->nw += flen;
    int decoded_data_blocks=0;
    const frame_hdr_t * const fh = (frame_hdr_t *) frame;

    size_t min_len = ML_FRAME_HEADER_WORDS + ML_FRAME_TRAILER_WORDS;
    switch (fh->type) {
    case FRAME_TYPE_ECHO:
        fillerDecStat->nf_echo++;
        break;
    case FRAME_TYPE_ADC_DATA:
    {
        ++fillerDecStat->nf_dadc;
        const int dlen = flen - min_len;
        if (dlen > 0) {
            const unsigned int *dbuf = (unsigned int *) &frame[ML_FRAME_HEADER_WORDS];
            int ret = adcmDataBlockDecoder->adcm_decode_data_blocks (dbuf, dlen, eventList, fillerDecStat);
            if (ret > 0) {
                decoded_data_blocks += ret;
            }
        }
    }
        break;
    case FRAME_TYPE_CTRL_REQ:
        fillerDecStat->nf_creq++;
        break;
    case FRAME_TYPE_CTRL_ACK:
        ++fillerDecStat->nf_cack;
        //    printf("decode_data_blocks: frame CTRL_ACK\n");
        break;
    case FRAME_TYPE_STAT:
        ++fillerDecStat->nf_stat;
        //    printf("decode_data_blocks: frame STAT\n");
        break;
    default:
        ++fillerDecStat->nf_unkn;
        qDebug() << QString("unexpected frame type 0x%1").arg(fh->type, 0, 16);
//        adcm_frame_print (frame, flen);
        break;
    }
    return decoded_data_blocks;
}


void EventListFiller::updateProgress(bool force)
{
    if(!force && lastProgressTime->elapsed() < updateProgressTimeout)
        return;
    lastProgressTime->restart();
    emit progressChanged(frameListSize ? 1.*framesDecoded/frameListSize : 1);
}
