//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "SettingsDialog.h"

#include <cmath>
#include <iostream>

#include <QLabel>
#include <QSpinBox>
#include <QCheckBox>
#include <QSettings>

#include "adcm16.h"
#include "ch_label.h"
#include "adcmraw/adcm_mem.h"
#include "runcontrol.h"

SettingsDialog::SettingsDialog(const conf_mem_t &i_conf_mem, QWidget *parent) :
    QDialog(parent),
    conf_mem(i_conf_mem)
{
    setupUi(this);
    textLabel_M->setText(ustr_ch_label_M);
    textLabel_S->setText(ustr_ch_label_S);

    spinBox_trig_min->setMinValue(1);
    spinBox_trig_max->setMinValue(1);
    spinBox_trig_min->setMaxValue(ADC_NCH);
    spinBox_trig_max->setMaxValue(ADC_NCH);

    spinBox_samples_base->setLineStep(2);
    spinBox_samples_base->setMinValue(ADCM_MIN_N_SAMP);
    //  spinBox_samples_base->setMaxValue(ADCM_MAX_N_SAMP);

    spinBox_samples_M->setEnabled(false);
    spinBox_samples_S->setEnabled(false);

    doubleSpinBoxTime->setMinimum(0.01);
    doubleSpinBoxTime->setMaximum(1.0*86400*365);

    spinBoxEventCount->setMinValue(1);
    spinBoxEventCount->setMaxValue(1000000000);
}

void SettingsDialog::apply()
{
    writeSettings();
    emit applied();
}

void SettingsDialog::restoreDefaults()
{

    // Upper check-boxes
    RunControl defaultRunControl;
    checkBox_EnableToolTips->setChecked(false); // unused
    checkBox_Recover->setChecked(defaultRunControl.getAutoRecover());
    checkBox_Autostart->setChecked(defaultRunControl.getAutoRun());
    checkBox_Autosave->setChecked(false); // there is no default presets, stores just in QSettings

    // Waveform length
    AdcConfig defAdcConf;
    spinBox_samples_base->setValue(defAdcConf.fragm_len_max);
//    spinBox_samples_M; // unused
//    spinBox_samples_S; // unused

    // Trigger Multiplicity
    spinBox_trig_min->setValue(defAdcConf.mult_min);
    spinBox_trig_max->setValue(defAdcConf.mult_max);

    // Stop conditions
    RunConfig defRunConfig = RunConfig();
    groupBoxStopCondition->setChecked(defRunConfig.stopCondition);
    checkBoxTime->setChecked(defRunConfig.stopByTimer);
    doubleSpinBoxTime->setValue(defRunConfig.stopTimerValue);
    checkBoxEventCount->setChecked(defRunConfig.stopByEventCount);
    spinBoxEventCount->setValue(defRunConfig.stopEventCountValue);

    //Timestamp
    if(defAdcConf.timestamp_is_utc)
        radioButtonTimestampUTC->setChecked(true);
    else
        radioButtonTimestampRelative->setChecked(true);
}

void SettingsDialog::setData( const SettingsData & in )
{
    data = in;
    QSettings s;
    checkBox_EnableToolTips->setChecked(s.value("tooltips").toBool());
    checkBox_Recover->setChecked(s.value("autorecover").toBool());
    checkBox_Autostart->setChecked(s.value("autostart").toBool());
    checkBox_Autosave->setChecked(s.value("autosave").toBool());

    //  data.sds.print();
    spinBox_trig_min->setMaxValue(conf_mem.nch);
    spinBox_trig_max->setMaxValue(conf_mem.nch);
    spinBox_trig_min->setValue(data.sds.trig_min);
    spinBox_trig_max->setValue(data.sds.trig_max ? data.sds.trig_max : conf_mem.nch);

    //  std::cout <<"SettingsDialog::setData base="<< data.sds.samples_base <<std::endl;
    updateLimits();
    spinBox_samples_base->setValue(data.sds.samples_base);
    spinBox_samples_M->setMinValue(ADCM_MIN_N_SAMP);
    int m = (data.sds.samples_mult.size() > 0) ? data.sds.samples_base*data.sds.samples_mult[0] : 0;
    spinBox_samples_M->setValue(m);
    spinBox_samples_S->setMinValue(ADCM_MIN_N_SAMP);
    spinBox_samples_M->setEnabled(false);
    spinBox_samples_S->setEnabled(false);

    // Stop Condition
    groupBoxStopCondition->setChecked(data.sds.cfg.run.stopCondition);

    checkBoxTime->setChecked(data.sds.cfg.run.stopByTimer);
    doubleSpinBoxTime->setValue(data.sds.cfg.run.stopTimerValue);

    checkBoxEventCount->setChecked(data.sds.cfg.run.stopByEventCount);
    spinBoxEventCount->setValue(data.sds.cfg.run.stopEventCountValue);

    radioButtonTimestampRelative->setChecked(!data.sds.timestamp_is_utc);
    radioButtonTimestampUTC->setChecked(data.sds.timestamp_is_utc);

    updateEnabledStatus();
}

void SettingsDialog::updateLimits()
{
    unsigned int maxsize = data.sds.cfg.adc.adc_cfg_get_recwin_max();
    spinBox_samples_base->setMaxValue(maxsize);
    spinBox_samples_M->setMaxValue(maxsize);
    spinBox_samples_S->setMaxValue(maxsize);
}

SettingsData SettingsDialog::getData() const
{
    return data;
}

void SettingsDialog::writeSettings()
{
    QSettings s;
    s.setValue("tooltips", checkBox_EnableToolTips->isChecked());
    s.setValue("autorecover", checkBox_Recover->isChecked());
    s.setValue("autostart", checkBox_Autostart->isChecked());
    s.setValue("autosave", checkBox_Autosave->isChecked());
    s.sync();
}

void SettingsDialog::accept()
{
    writeSettings();
    QDialog::accept();
}

void SettingsDialog::on_spinBox_samples_base_valueChanged( int base )
{
    data.sds.samples_base = base;
    int maxval = (int)(1.*ADCM_MAX_N_SAMP/base)*base;
    int origval = spinBox_samples_M->value();
    //  double fval = 1.*origval/base;
    //  int val = (int)(((int)fval)*base);
    int val = (int)(1.*origval/base)*base;
    //  qDebug(QString("base=%1, val=%2, fval=%3, newval=%4").arg(base).arg(origval).arg(fval).arg(val));
    spinBox_samples_M->setMinValue(base);
    spinBox_samples_M->setMaxValue(maxval);
    spinBox_samples_M->setLineStep(base);
    spinBox_samples_M->setValue(val);
    spinBox_samples_S->setMinValue(base);
    spinBox_samples_S->setMaxValue(maxval);
    spinBox_samples_S->setLineStep(base);
    spinBox_samples_S->setValue(val);
}


void SettingsDialog::on_spinBox_samples_S_valueChanged( int val )
{
    Q_UNUSED(val);
}

void SettingsDialog::on_spinBox_samples_M_valueChanged( int val )
{
    Q_UNUSED(val);
    //  for(int c=0;c<ADCM_N_CH;c++)
    //    data.sds.samples_mult[c] = val/data.sds.samples_base;
}

void SettingsDialog::on_spinBox_trig_min_valueChanged( int val )
{
    data.sds.trig_min = val;
}

void SettingsDialog::on_spinBox_trig_max_valueChanged( int val )
{
    data.sds.trig_max = val;
    data.sds.cfg.adc.mult_max = data.sds.trig_max;
    updateLimits();
}

// Stop Condition
void SettingsDialog::on_doubleSpinBoxTime_valueChanged( double value )
{
    data.sds.cfg.run.stopTimerValue = value;
}

void SettingsDialog::on_spinBoxEventCount_valueChanged( int value )
{
    data.sds.cfg.run.stopEventCountValue = value;
}

void SettingsDialog::on_checkBoxTime_toggled( bool value )
{
    data.sds.cfg.run.stopByTimer = value;
    updateEnabledStatus();
}

void SettingsDialog::on_checkBoxEventCount_toggled( bool value )
{
    data.sds.cfg.run.stopByEventCount = value;
    updateEnabledStatus();
}

void SettingsDialog::on_groupBoxStopCondition_toggled( bool value )
{
    data.sds.cfg.run.stopCondition = value;
    updateEnabledStatus();
}

void SettingsDialog::on_radioButtonTimestampUTC_toggled( bool )
{
  data.sds.timestamp_is_utc = radioButtonTimestampUTC->isChecked();
}

void SettingsDialog::updateEnabledStatus()
{
    doubleSpinBoxTime->setEnabled(data.sds.cfg.run.stopCondition && data.sds.cfg.run.stopByTimer);
    spinBoxEventCount->setEnabled(data.sds.cfg.run.stopCondition && data.sds.cfg.run.stopByEventCount);
}

void SettingsDialog::on_buttonBox_clicked(QAbstractButton* button)
{
    switch(buttonBox->standardButton(button)) {
    case QDialogButtonBox::Apply :
        apply();
        break;
    case QDialogButtonBox::RestoreDefaults :
        restoreDefaults();
        break;
    default:;
    }
}
