//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QFileDialog>
#include <QFileInfo>
#include <QLineEdit>
#include <QMessageBox>

#include "SetupDataOutputDialog.h"
#include "storage.h"
#include "config/storageconfig.h"

SetupDataOutputDialog::SetupDataOutputDialog(StorageConfig sets, int storageMode, bool runMode, QWidget * parent) :
    QDialog(parent),
    runControlInRunState(runMode)
{
    setupUi(this);

    setCaption(storageMode == SET_DATAFILE_RAW ? "Setup RAW Data Output..." : "Setup Decoded Event Data Output...");
    setRawEnable(sets.enable);
    setMode(sets.mode);
    period = sets.period;
    spinBox_period->setValue(period);
    basedir = sets.bn;
    setFileName(sets.fn);
    groupBox_limit_size->hide();
}

void SetupDataOutputDialog::on_pushButton_Select_clicked()
{
    QString fullname = basedir + QDir::separator() + filename;
    QFileInfo fi(filename);
    if (basedir.isNull())
        basedir = fi.dirPath();

    QString fileName = QFileDialog::getSaveFileName(this, tr("Data output"),
                                                    fullname,
                                                    tr("RAW  files (*.raw *.dat)") + tr(";;") + tr("Any files (*)"));
    if (!fileName.isNull()) {
        QFileInfo fi(fileName);
        filename = fi.fileName();
        basedir = fi.dirPath();
        update_filename();
    }
}

void SetupDataOutputDialog::setMode( int value )
{
    mode = value;
    update_mode();
}

void SetupDataOutputDialog::setRawEnable( bool value )
{
    raw_enable = value;
    update_raw_enable();
}

void SetupDataOutputDialog::update_filename()
{
    QString s;
    s = basedir;
    s += QDir::separator();
    s += filename;
    lineEdit_FileName->setText(s);
}


void SetupDataOutputDialog::setFileName( const QString & text )
{
    filename = text;
    update_filename();
}


void SetupDataOutputDialog::update_mode()
{
    switch(mode) {
    case storage::MODE_CONT :
        radioButton_continuous->setChecked(true);
        periodGroupBox->setEnabled(false);
        break;
    case storage::MODE_RECYCLE :
        radioButton_recycle->setChecked(true);
        periodGroupBox->setEnabled(true);
        break;
    }
}

void SetupDataOutputDialog::update_raw_enable()
{
    groupBox_raw->setChecked(raw_enable);
}

void SetupDataOutputDialog::on_groupBox_raw_toggled( bool value )
{
    raw_enable = value;
}

void SetupDataOutputDialog::on_radioButton_recycle_clicked()
{
    mode = storage::MODE_RECYCLE;
    periodGroupBox->setEnabled(true);
}

void SetupDataOutputDialog::on_radioButton_continuous_clicked()
{
    mode = storage::MODE_CONT;
    periodGroupBox->setEnabled(false);
}

void SetupDataOutputDialog::on_spinBox_period_valueChanged(int val)
{
    period = val;
}

bool SetupDataOutputDialog::updateStorageConfig(StorageConfig &newset) const
{
    if(newset.bn == basedir && newset.fn == filename &&
            newset.mode == mode && newset.period == period && newset.enable == raw_enable)
        return false;
    newset.bn = basedir;
    newset.fn = filename;
    newset.mode = mode;
    newset.period = period;
    newset.enable = raw_enable;

    if (runControlInRunState) {
        int code = QMessageBox::warning ((QWidget*)this, "Confirm...",
                                         "Data taking in progress. Do you want to reset and change data file?",
                                         QMessageBox::Yes|QMessageBox::Default,
                                         QMessageBox::No|QMessageBox::Escape);
        if (code != QMessageBox::Yes)
            return false;
    }

    return true;
}
