
#include <QDebug>
#include <QThread>
#include <QTimer>
#include <QByteArray>
#include <QVector>
#include <QUrl>
#include <QFile>
#include <QDateTime>
#include <QApplication>

#include "adcinstance.h"
#include "adchardw.h"
#include "adcm_raw.h"
#include "adcm_ids.h"
#include "adcm_df.h"
#include "adcm_dws.h"
#include "mlink/mlink.h"
#include "regio/regiobase.h"
#include "regio/regiopacket.h"
#include "mlink/regiomlink.h"
//#include "regioemu.h"
#include "stopwatchtimer.h"


const int hardwarePollIntervalMsec = 30;
const int hitcountsIntervalMsec = 1000;
const int MaxFrameListSize = 1000;

const uint maxQueueEvents = 4;

namespace {
double timestamp_sec()
{
    QDateTime tv = QDateTime::currentDateTime();
    return tv.toTime_t() + 1e-3*tv.time().msec();
}
}

class LinkFactory
{
public:
    static mlink::RegIOBase *create();
};

mlink::RegIOBase *LinkFactory::create()
{
    mlink::RegIOBase *adcmLink = 0;
    // Create hardware interface
    try {
        adcmLink = new mlink::RegIOMLink();
    }
    catch (std::exception &e) {
        qCritical() << "Interface open failed: " << e.what();
    }
    if (!adcmLink) {
        qFatal("Hardware not found");
        //        RegIOEmu *adcmEmuLink = new RegIOEmu();
        //        adcmLink = adcmEmuLink;
        //        QSharedPointer<AdcmDeviceEmu> adcmDeviceEmu(new AdcmDeviceEmu);
        //        adcmEmuLink->insertDeviceEmu(AdcmDeviceEmu::defaultAddress(), adcmDeviceEmu);
    }
//    qDebug() << "Created link";
    return adcmLink;
}

AdcInstance::AdcInstance(QObject *parent)
    : QObject(parent),
      adcmLink(LinkFactory::create()),
      adcmDevice(0),
      pollTimer(new QTimer(this)),
      hitcountsTimer(new QTimer(this)),
      frameFinder(new FrameFinder(this)),
      prev_pos(0),
      decoder_pos(0),
      p_state(UnconnectedState),
      dma_running(false),
      runTimer(new StopWatchTimer(this)),
      lastSentSeq(0),
      lastReceivedSeq(0),
      statsNumBytesReceived(0),
      statsTimestamp(0),
      prevAdcHwCounters(new AdcHwCounters)
{
    setObjectName("AdcInstance thread");
    gain_step_db = 3;
    min_gain_db = 0;
    max_gain_db = 0;
    if (hasGainControl()) {
        min_gain_db = adc_gain_val_to_db(0);
        max_gain_db = adc_gain_val_to_db(15);
    }
    qRegisterMetaType<conf_mem_t>();
    qRegisterMetaType<QVector<double> >("QVector<double>");

    connect(this, SIGNAL(error()), SLOT(on_error()));
    connect(this, SIGNAL(connected()), SLOT(on_connected()));
    connect(this, SIGNAL(disconnected()), SLOT(on_disconnected()));
    connect(pollTimer, SIGNAL(timeout()), SLOT(poll()));
    pollTimer->setInterval(hardwarePollIntervalMsec);
//    pollTimer->setInterval(0);
//    pollTimer->setSingleShot(true);
    pollTimer->start();

    connect(hitcountsTimer, SIGNAL(timeout()), this, SLOT(readHitCounts()));
    hitcountsTimer->setInterval(hitcountsIntervalMsec);
    hitcountsTimer->setSingleShot(true);
    hitcountsTimer->start();

    QTimer *statisticsTimer(new QTimer(this));
    connect(statisticsTimer, SIGNAL(timeout()), SLOT(updateStatistics()));
    statisticsTimer->start(100);
}

AdcInstance::~AdcInstance()
{
    delete prevAdcHwCounters;
    delete adcmDevice;
    delete adcmLink;
}

void AdcInstance::emitError(const QString &str)
{
    setErrorString(str);
    qCritical() << errorString();
    emit error();
}

float AdcInstance::adc_gain_val_to_db(int value) const
{
    return AdcmDevice::adc_gain_val_to_db(value);
}
/*
int AdcInstance::adc_gain_val_to_raw(float db_value) const
{
  return (db_value+9) / 3;
}
*/
float AdcInstance::getMinGain() const
{
    return min_gain_db;
}

float AdcInstance::getMaxGain() const
{
    return max_gain_db;
}

float AdcInstance::getGainStep() const
{
    return gain_step_db;
}

bool AdcInstance::hasGainControl() const
{
    return adcmDevice ? adcmDevice->getConfMem().hasGainControl() : false;
}

bool AdcInstance::hasTriggerControl() const
{
    return adcmDevice ? adcmDevice->getConfMem().hasTriggerControl() : false;
}

quint64 AdcInstance::queueSize() const
{
    return lastSentSeq - lastReceivedSeq;
}

bool AdcInstance::isQueueFull() const
{
    return queueSize() >= maxQueueEvents;
}

void AdcInstance::poll()
{
    if (!adcmDevice) return;
    if (isQueueFull()) {
        pollTimer->start(hardwarePollIntervalMsec);
//        qDebug().nospace() << "AdcInstance::poll thread " << lastSentSeq <<" | "<< lastReceivedSeq
//                           << "; queue " << queueSize() << "/" << maxQueueEvents;
        return;
    } else {
//        qDebug() << "AdcInstance::poll thread polling";
    }
    //    qDebug() << "AdcInstance::poll thread " << QThread::currentThreadId ();
    double start = timestamp_sec();
    try {
        adcmDevice->poll();
    }
    catch (std::exception &e) {
        emitError(QString("Exception in poll(): %1").arg(e.what()));
    }
    double timestampPoll = timestamp_sec();

    FrameList frameList(adcmDevice->popDataPackets());
    frameList.created = QTime::currentTime();

    double timestampFrameList = timestamp_sec();
    const int numFrames = frameList.size();
    if (numFrames > 0) {
        //        qDebug() << "Received" << numFrames << "frames";
            lastSentSeq++;
            frameList.seq = lastSentSeq;
            emit frameReceived(frameList);
    }

    quint64 numBytes = 0;
    Q_FOREACH(QByteArray buf, frameList) {
        numBytes += buf.size();
    }
    numBytesReceived += numBytes;
    double timestampEmitData = timestamp_sec();

    emit timeElapsed(runTimer->elapsed());
    emit byteCountChanged(numBytesReceived);
    statsNumBytesReceived += numBytes;
    double timestampFinish = timestamp_sec();

    if (DEBUG_PRINT_DATA_RATE)
    if (numBytes > 0) {
        qDebug().nospace() << "Readout time (ms): poll " << (timestampPoll - start) * 1e3
                           << ", frameList " << (timestampFrameList - timestampPoll) * 1e3
                           << ", emit data " << (timestampEmitData - timestampFrameList) * 1e3
                           << ", emit stats " << (timestampFinish - timestampEmitData) * 1e3;

        qDebug().nospace() << "Readout throughput: " << (1.0 * numBytes / (timestampPoll - start) / 1024 / 1024) << " MB/s";
    }
//    int nextPollIntervalMs = (numFrames && isQueueFull()) ? 0 : hardwarePollIntervalMsec;
    int nextPollIntervalMs = hardwarePollIntervalMsec;
    pollTimer->start(nextPollIntervalMs);
}

void AdcInstance::updateStatistics()
{
    double now = timestamp_sec();
    double period = now - statsTimestamp;
    double rate = statsNumBytesReceived / period / 1024 / 1024;
    if (DEBUG_PRINT_DATA_RATE) {
        if (dma_running)
            qDebug() << rate << "MB/s";
    }
    statsTimestamp = now;
    statsNumBytesReceived = 0;
}

adcm_counters_t AdcInstance::makeAdcmCounters(const AdcHwCounters &adcHwCounters) const
{
    adcm_counters_t cnt;
    const int n = std::min((int)ADC_NCH, adcHwCounters.hits.size());
    cnt.n = n;
    for (int i=0; i<n; i++)
        cnt.rawhits[i] = adcHwCounters.hits[i];
    for (int i=n; i<(int)ADC_NCH; i++)
        cnt.rawhits[i] = 0;
    cnt.time = adcHwCounters.current_ts;
    return cnt;
}

void AdcInstance::readHitCounts()
{
    try {
        if (adcmDevice) {
            AdcHwCounters cnt = adcmDevice->adc_get_counters();
            emit hitCountsChanged(makeAdcmCounters(cnt - *prevAdcHwCounters));
            *prevAdcHwCounters = cnt;
            QVector<double> v = adcmDevice->readTemperatures();
            Q_EMIT tempChanged(v);
        }
    }
    catch (std::exception &e) {
        emitError(QString("Exception in adc_get_counters(): %1").arg(e.what()));
    }
    hitcountsTimer->start();
}

void AdcInstance::connectToDevice(const QUrl &url)
{
    if (adcmDevice) {
        delete adcmDevice;
        emit disconnected();
    }
    setErrorString("No error");
    p_state = ConnectingState;
    qDebug() << "State: Connecting";
//    if (!url.isValid()) {
//        emitError(url.errorString());
//        return;
//    }
//    const QString &scheme = url.scheme();
//    if (!scheme.isEmpty()) {
//        emitError(QString("Unsupported URL scheme: %1").arg(scheme));
//        return;
//    }
//    const QString &fileName = url.toLocalFile();
//    if (!QFile::exists(fileName)) {
//        emitError(QString("Not a local file: %1").arg(fileName));
//        return;
//    }
    try {
        adcmDevice = new AdcmDevice(*adcmLink);
        adcmDevice->adc_reset();
    }
    catch(std::exception &e) {
        emitError(QString("Exception in connectToDevice(): %1").arg(e.what()));
        return;
    }
    if (adcmDevice) {
        emit confMemModified(adcmDevice->getConfMem());
        emit connected();
    }
}

bool AdcInstance::waitForConnected(int msecs)
{
    const QTime &entryTime = QTime::currentTime();
    while (state() == ConnectingState) {
        const QTime &currentTime = QTime::currentTime();
        int diff = entryTime.msecsTo(currentTime);
        if(diff<0) diff += 86400000;
        if (diff > msecs)
            return false;
        qApp->processEvents();
    }
    return state() == ConnectedState;
}

void AdcInstance::on_connected()
{
    p_state = ConnectedState;
    qDebug() << "State: Connected";
}

void AdcInstance::on_disconnected()
{
    p_state = UnconnectedState;
    qDebug() << "State: Disconnected";
}

void AdcInstance::on_error()
{
    p_state = UnconnectedState;
    qDebug() << "State: Error";
}

void AdcInstance::setup(const AllConfig &cfg)
{
    if (!adcmDevice) return;
    try {
        adcmDevice->setup(cfg);
    }
    catch (std::exception &e) {
        emitError(QString("Exception in setup(): %1").arg(e.what()));
    }
}

void AdcInstance::startReadout()
{
    if (!adcmDevice) return;
    if (dma_running)
        return;
    try {
        if (!adcmDevice->adc_start()) {
            emitError("startReadout failed");
            return;
        }
    }
    catch (std::exception &e) {
        emitError(QString("Exception in adc_start(): %1").arg(e.what()));
        return;
    }
    runTimer->start();
    adcm_counters.clear();
    dma_running = true;
//    qDebug() << "startReadout Ok";
}

void AdcInstance::stopReadout()
{
    if (!adcmDevice) return;
    try {
        adcmDevice->adc_stop();
    }
    catch (std::exception &e) {
        emitError(QString("Exception in adc_stop(): %1").arg(e.what()));
        return;
    }
    runTimer->stop();
    dma_running = false;
//    qDebug() << "stopReadout Ok";
}

double AdcInstance::runtime () const
{
    return runTimer->elapsed();
}

void AdcInstance::reset()
{
    numBytesReceived = 0;
    adcm_counters.clear();
    if (dma_running)
        runTimer->restart();
    else
        runTimer->reset();
}

void AdcInstance::dataProgress(quint64 value)
{
    lastReceivedSeq = value;
//    qDebug() << QString("lastReceivedSeq=%1").arg(lastReceivedSeq);
}

QStringList AdcInstance::getDeviceIdStrList() const
{
    if (!adcmDevice) return QStringList();
    return adcmDevice->getConfMem().getDeviceIdStrList();
}
