//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <cmath>

#include <QDebug>
#include <QTimer>
#include <QLabel>
#include <QPixmap>
#include <QTableWidget>

#include "adcmraw/adcm_mem.h"
#include "config/adcconfig.h"
#include "config/calzeroconfig.h"

#include "calzerodialog.h"

namespace {
const int TABLE_COL_COUNT = 0;
const int TABLE_COL_OFFSET = 1;
const int TABLE_COL_PERCENT = 2;
const int TABLE_COL_RMS = 3;
const int TABLE_N_COLS = 4;
}

CalZeroDialog::CalZeroDialog (QWidget * parent)
    : CalibrationDialog(parent),
      prevCalZero(new CalZeroConfig)
{
    setupUi(this);
    table->setColumnCount(TABLE_N_COLS);
    table->setHorizontalHeaderItem(TABLE_COL_COUNT, new QTableWidgetItem("Count"));
    table->setHorizontalHeaderItem(TABLE_COL_OFFSET, new QTableWidgetItem("Offset"));
    table->setHorizontalHeaderItem(TABLE_COL_PERCENT, new QTableWidgetItem("Offset, %"));
    table->setHorizontalHeaderItem(TABLE_COL_RMS, new QTableWidgetItem("RMS noise"));

    update();
}

CalZeroDialog::~CalZeroDialog()
{
    delete prevCalZero;
}

int CalZeroDialog::adc_baseline_good(int ch) const
{
    double val = calZero->get(ch);
    int halfRange = conf_mem->getAdcHalfRange(ch);
    int mid = conf_mem->getAdcMidValue(ch);
    double tolerance = 0.05; // relative to half-scale
    double max = (double)mid + halfRange * tolerance;
    double min = (double)mid - halfRange * tolerance;
    //  printf("bl_good: %f < %f < %f\n", min, val, max);
    return (min < val && val < max);
}

int CalZeroDialog::adc_noiselevel_good(int ch) const
{
    double val = sqrt(calZero->getcalnoiserms(ch)) * 1000. / conf_mem->getAdcHalfRange(ch);
    //  printf("adc_noiselevel_good=%f\n", val);
    if (val==0) return 0; // failure
    return (0.05 < val && val < 3.0);
}


void CalZeroDialog::update ()
{
    if (calZero == prevCalZero) return;
    QTableWidget *t = table;

    const int nch = calZero->nent.size();
    int oldRowCount = t->rowCount();
    t->setRowCount(nch);
    for(int row=oldRowCount; row < nch; ++row) {
        for(int col=0;col<TABLE_N_COLS; ++col) {
            t->setItem(row, col, new QTableWidgetItem());
        }
        t->setVerticalHeaderItem(row, new QTableWidgetItem(QString::number(row)));
    }

    for(int i=0; i<nch; i++) {
        bool hasHwChannel = (i < conf_mem->nch);
        int adcmid = conf_mem->getAdcMidValue(i);
        int adcHalfRange = conf_mem->getAdcHalfRange(i);
        if (adcHalfRange <= 0)
            adcHalfRange = 1;
        //	qDebug(QString("adcmid[%1]=%2").arg(i).arg(adcmid));
        //	  enable = (cfg.adc.trig[i] & 1);
        bool enable = adcConfig->ch_enabled_offset_calibration(i);
        int cznent = calZero->nent[i];
        double base=0, noise=0;
        if (cznent > 0) {
            base = ((double)calZero->base[i]) / cznent - adcmid;
            noise = (double)calZero->rms[i] / cznent;
        }
//        qDebug() << "CalZeroDialog: " << i << cznent << base << sqrt(noise);
        if (cznent > 0)
            t->item(i, TABLE_COL_COUNT)->setText(QString::number(cznent));
        else
            t->item(i, TABLE_COL_COUNT)->setText(QString());

        if (cznent > 0) {
            t->item(i, TABLE_COL_OFFSET)->setText(QString("%1").arg(base, 0, 'f', 1));
            if (hasHwChannel)
                t->item(i, TABLE_COL_PERCENT)->setText(QString("%1").arg(base/(adcHalfRange)*100., 0, 'f', 2));
            else
                t->item(i, TABLE_COL_PERCENT)->setText(QString());
            t->item(i, TABLE_COL_RMS)->setText(QString("%1").arg(sqrt(noise), 0, 'f', 2));
         } else {
            t->item(i, TABLE_COL_OFFSET)->setText(QString());
            t->item(i, TABLE_COL_PERCENT)->setText(QString());
            t->item(i, TABLE_COL_RMS)->setText(QString());
        }

        for (int j=0;j<TABLE_N_COLS;j++)
            t->item(i, j)->setFlags(enable ? (Qt::ItemIsEnabled | Qt::ItemIsSelectable) : Qt::NoItemFlags);

        if(enable){
            for (int j=0;j<TABLE_N_COLS;j++)
                t->item(i, j)->setBackgroundColor(Qt::transparent);
            QTableWidgetItem *item = t->item(i, TABLE_COL_OFFSET);
            bool ok = !hasHwChannel || cznent == 0 || adc_baseline_good(i);
            if(ok != (item->backgroundColor() == Qt::transparent))
                item->setBackgroundColor(ok ? Qt::transparent : Qt::red);
            item = t->item(i, TABLE_COL_RMS);
            ok = !hasHwChannel || cznent == 0 || adc_baseline_good(i);
            if(ok != (item->backgroundColor() == Qt::transparent))
                item->setBackgroundColor(ok ? Qt::transparent : Qt::red);
        } else {
            for (int j=0;j<TABLE_N_COLS;j++)
                t->item(i, j)->setBackgroundColor(Qt::lightGray);
        }
    }

    t->resizeColumnsToContents();
    *prevCalZero = *calZero;
}
