//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <cmath>
#include <cstdlib>
#include <cstring>
#include <algorithm>

#include <QDebug>
#include <QFileDialog>
#include <QMessageBox>
#include <QComboBox>
#include <QLabel>
#include <QSpinBox>
#include <QSettings>
#include <qwt_scale_engine.h>

//#include "qxw/WaveformView.h"
#include "qxw/WaveformPlot.h"
#include "qxw/WaveformData.h"

#include "config/allconfig.h"
#include "adchardw.h"
#include "dsplib/fir.h"
#include "dsplib/dsplib.h"
#include "dsplib/fitter.h"
#include "dsplib/fftengine.h"
#include "dsplib/Convolver.h"
#include "ch_label.h"
#include "qlog/qlog.h"
//#include "plot.h"

#include "dsplib/fftengine.h"
#include "dcconf.h"
#include "util/TextFormatter.h"

#ifdef ALL_DEBUG
# undef DCCONF_DEBUG
# define DCCONF_DEBUG ALL_DEBUG
#endif

#ifndef DCCONF_DEBUG
# define DCCONF_DEBUG 0         /* dummy */
#endif

#define PREFIX "dcconf: "

double INPFFT_MIN = 1e-1;
double INPFFT_MAX = 1e3;

// local defines for createfir, fir.c use different values
const int FIR_COEFF_BITS = 14;
const int FIR_OPT_MIN_VAL = 5;

const int calibrationColorIndex = 4;
const int referenceColorIndex = 1;
const int filterColorIndex = 6;
const int testColorIndex = 4;

fwinf_t win_type_by_id(int id)
{
    switch (id) {
    case 0:  return WF_RECT;
    case 1:  return WF_BLACKMAN;
    case 2:  return WF_HAMMING;
    case 3:  return WF_HANN;
    case 4:  return WF_BARLETT;
    case 5:  return WF_TUKEY_075;
    case 6:  return WF_TUKEY_050;
    case 7:  return WF_TUKEY_025;
    case 8:  return WF_GAUSS_075;
    case 9:  return WF_GAUSS_050;
    case 10: return WF_GAUSS_025;
    default: return WF_RECT;
    }
}

int win_id_by_type(fwinf_t wt)
{
    switch(wt) {
    case WF_RECT: return 0;
    case WF_BLACKMAN: return 1;
    case WF_HAMMING: return 2;
    case WF_HANN: return 3;
    case WF_BARLETT: return 4;
    case WF_TUKEY_075: return 5;
    case WF_TUKEY_050: return 6;
    case WF_TUKEY_025: return 7;
    case WF_GAUSS_075: return 8;
    case WF_GAUSS_050: return 9;
    case WF_GAUSS_025: return 10;
    default: return 0;
    }
}

dcconf::dcconf (QWidget* parent, const conf_mem_t &i_conf_mem, const AdcConfig &i_adcConfig, const DecoderConfig &i_decoderConfig)
    : QDialog(parent),
      ui(this),
      conf_mem(i_conf_mem),
      adcConfig(i_adcConfig),
      currentDecoderConfig(i_decoderConfig),
      newDecoderConfig(i_decoderConfig),
      fft(new RFFTEngine)
{
    setupUi(this);
    //    extcfg = incfg;
    disable_control_slots = 0;
    restoring = 0;
    ch = 0;

    comboBoxWindowingFunction->insertItem ("Rectangle", 0);
    comboBoxWindowingFunction->insertItem ("Blackman", 1);
    comboBoxWindowingFunction->insertItem ("Hamming", 2);
    comboBoxWindowingFunction->insertItem ("Hanning", 3);
    comboBoxWindowingFunction->insertItem ("Barlett", 4);
    comboBoxWindowingFunction->insertItem ("Tukey 0.75", 5);
    comboBoxWindowingFunction->insertItem ("Tukey 0.5", 6);
    comboBoxWindowingFunction->insertItem ("Tukey 0.25", 7);
    comboBoxWindowingFunction->insertItem ("Gauss 0.25", 8);
    comboBoxWindowingFunction->insertItem ("Gauss 0.5", 9);
    comboBoxWindowingFunction->insertItem ("Gauss 0.75", 10);
    win_type = WF_BLACKMAN;             // default
    comboBoxWindowingFunction->setCurrentItem (win_id_by_type((fwinf_t)win_type));

    sbfirnp->setMaxValue (FIRMAXLEN);
    //  sbfirnp->setValue(FIRMAXLEN);

    Restore();
}

dcconf::~dcconf ()
{
    delete fft;
}

std::vector<double> dcconf::getCalAvgY(int channel) const
{
    std::vector<double> calAvg;
    const size_t n = newDecoderConfig.calavg.data[channel].size();
    if (n == 0) {
// no shape calibration data
    } else {
//    const size_t n = newAllConfig.adc.recwin[channel];
        calAvg.reserve(n);
        for (size_t i = 0; i < n; i++) {
            calAvg.push_back(ca (channel, i));
        }
    }
    return calAvg;
}

double dcconf::getCalAvgAmplitude(int channel) const
{
    const std::vector<double> &calAvg = getCalAvgY(channel);
    if (calAvg.size() == 0) return 0;
    return std::max(1.0, *std::max_element(calAvg.begin(), calAvg.end()));
}

std::vector<double> dcconf::getRefY(int channel, double amplitudeScaler) const
{
    const int lat = data_latency ();
    const size_t n = adcConfig.recwin[channel];
    std::vector<double> ref;
    ref.reserve(n);
    for (size_t i = 0; i < n; i++) {
        ref.push_back(amplitudeScaler * reqfunscaled (int(i) - lat));
    }
    return ref;
}

std::vector<double> dcconf::getRefY(int channel) const
{
    return getRefY(channel, getCalAvgAmplitude(channel));
}

std::vector<double> dcconf::getRefXInUs(int channel) const
{
    const int lat = data_latency ();
    const size_t n = adcConfig.recwin[channel];
    std::vector<double> x;
    x.reserve(n);
    for (size_t i = 0; i < n; i++) {
        x.push_back((int(i) - lat) * getTimeStepInUs (channel));
    }
    return x;
}

std::vector<double> dcconf::getCalAvgXInUs(int channel) const
{
    return getRefXInUs(channel);
}

void dcconf::replotINP ()
{
    WaveformPlot *wp = ui->plotINP; // Input signal

    std::vector<double> vx = getCalAvgXInUs(ch);
    std::vector<double> vy0 = getCalAvgY(ch);
    std::vector<double> vy1 = getRefY(ch);
    size_t n = std::min(vx.size(), std::min(vy0.size(), vy1.size()));
    if (n == 0) {
        wp->clear();
        wp->replot();
        return;
    }
    vx.resize(n);
    vy0.resize(n);
    vy1.resize(n);
    WaveformDataAll wda;
    wda[0] = WaveformData(vx, vy0, tr("Calibration data"), calibrationColorIndex);
    wda[1] = WaveformData(vx, vy1, tr("Reference data"), referenceColorIndex);
    wp->setWaveformData(wda);
    wp->disableLegend();

    QwtText xTitle(tr("Time")+QString::fromUtf8(", μs"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();
}

void dcconf::replotINPFFT ()
{
    WaveformPlot *wp = ui->plotINPFFT; // Input signal FFT


    const std::vector<double> &inp = getCalAvgY(ch);
    const std::vector<double> &req = getRefY(ch);
    const size_t N = std::min(inp.size(), req.size());
    // Find power spectra and peaks
    std::vector<std::complex<double> > inpDFT = fft->DFT(inp);
    inpDFT.resize(N/2);
    std::vector<std::complex<double> > reqDFT = fft->DFT(req);
    reqDFT.resize(N/2);
    const double dftScaler = 1. / sqrt(N);
    const double fullScaleSigned = conf_mem.getAdcHalfRange(ch);
    const double fullScaleDb = val2db(fullScaleSigned);

    const size_t inpDftSize = inpDFT.size();
    std::vector<double> inpDftPy;
    for (size_t i = 0; i < inpDftSize; i++) {
        inpDftPy.push_back(val2db(std::abs(inpDFT[i]) * dftScaler) - fullScaleDb); // normalize DFT
    }
    const size_t reqDftSize = reqDFT.size();
    std::vector<double> dftPx;
    dftPx.reserve(reqDftSize);
    for (size_t i = 0; i < reqDftSize; i++) {
        dftPx.push_back(double(i) / N / getTimeStepInUs (ch));
    }

    std::vector<double> reqDftPy;
    reqDftPy.reserve(reqDftSize);
    for (size_t i = 0; i < reqDftSize; i++) {
        reqDftPy.push_back(val2db(std::abs(reqDFT[i]) * dftScaler) - fullScaleDb); // normalize DFT
    }

    WaveformDataAll wda;
    wda[0] = WaveformData(dftPx, inpDftPy, "Source FFT", calibrationColorIndex);
    wda[1] = WaveformData(dftPx, reqDftPy, "Target FFT", referenceColorIndex);

    // horizontal markers
    if (newDecoderConfig.calavg.nf.size() > ch && newDecoderConfig.pnf[ch] > ch) {
        const double snfdb = newDecoderConfig.calavg.nf[ch] - fullScaleDb;
        const double rnfdb = newDecoderConfig.pnf[ch] - fullScaleDb;
        //  const double snf = db2val (snfdb);
        //  const double rnf = db2val (rnfdb);

        std::vector<double> nfpx(2), snfpy(2, snfdb), rnfpy(2, rnfdb);
        if (reqDftPy.size() >= 2) {
            nfpx[0] = 0;
            //      nfpx[1] = reqDftPy.size() - 1;
            nfpx[1] = dftPx.back();
        } else {
            nfpx[0] = 0;
            nfpx[1] = 1;
        }

        wda[2] = WaveformData(nfpx, snfpy, "Signal Noise Floor", calibrationColorIndex+8);
        wda[3] = WaveformData(nfpx, rnfpy, "Target Noise Floor", referenceColorIndex+8);
    }
    wp->setWaveformData(wda);

    //  wp->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine);
    //  wp->axisScaleEngine(QwtPlot::yLeft)->setAttribute(QwtScaleEngine::Floating);
    //  wp->setAxisOptions (QwtPlot::yLeft, QwtAutoScale::Floating | QwtAutoScale::Logarithmic);
    //  wp->setAxisScale (QwtPlot::yLeft, INPFFT_MIN, INPFFT_MAX);
    //  wp->setxrange (0, H);

    //  printf("%edB  %e   %edB  %e\n", snf, snfdb, rnf, rnfdb);
    //  printf("%e:%e - %e:%e\n", px[0], p1[0], px[1], p1[1]);
    //  printf("%e:%e - %e:%e\n", px[0], p2[0], px[1], p2[1]);
    wp->disableLegend();

    QwtText xTitle(tr("Frequency, MHz"));
    QwtText yTitle(tr("dBFS"));
    xTitle.setFont(font());
    yTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->setAxisTitle(QwtPlot::yLeft, yTitle);
    wp->replot();
}

void dcconf::replotFIR ()
{
    WaveformPlot *wp = ui->plotFIR; // FIR impulse

    if (ch >= newDecoderConfig.fir.size()) {
        wp->clear();
        ui->tlklen->clear();
        ui->tloffs->clear();
        ui->tlsum->clear();
        ui->tlbits->clear();
        return;
    }
    const intfir_t &fir = newDecoderConfig.fir[ch];

    const size_t klen = fir.klen;
    const int firbits = fir.bits;
    const int firdelay = fir.offs;
    std::vector<double> FIRpx, FIRpy, WFpy;
    for (size_t i = 0; i < klen; i++) {
        FIRpx.push_back(i * getTimeStepInUs (ch));
        FIRpy.push_back(fir.coef[i]);
        WFpy.push_back(windowingFunction ((fwinf_t)win_type, i, klen) * BITS2MAXINT (firbits));
    }

    WaveformDataAll wda;
    wda[0] = WaveformData(FIRpx, FIRpy, "Filter Impulse Response", filterColorIndex);
    wda[1] = WaveformData(FIRpx, WFpy, "Windowing", filterColorIndex+1);
    wp->setWaveformData(wda);
    wp->disableLegend();

    QwtText xTitle(tr("Time")+QString::fromUtf8(", μs"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();

    ui->tlklen->setText (QString("Size: %1").arg(klen));
    ui->tloffs->setText (QString("Offs: %1").arg(firdelay));
    double sum = 0;
    for (size_t  i = 0; i < klen; i++)
        sum += newDecoderConfig.fir[ch].coef[i];
    ui->tlsum->setText (QString("Sum: %1").arg(sum));
    ui->tlbits->setText (QString("Bits: %1").arg(firbits));
}

void dcconf::replotFIRFFT ()
{
    WaveformPlot *wp = ui->plotFIRFFT; // FIR impulse FFT

    if (ch >= newDecoderConfig.fir.size()) {
        wp->clear();
        return;
    }
    const intfir_t &fir = newDecoderConfig.fir[ch];

    const size_t N = fir.klen;
    std::vector<double> firCoef;
    firCoef.assign(&fir.coef[0], &fir.coef[N]);

    // Find power spectra and peaks
    std::vector<std::complex<double> > firDft = fft->DFT(firCoef);
    firDft.resize(N/2); // result is symmetric
    const double dftScaler = 1. / sqrt(N);
    const double fullScaleSigned = 1<<(fir.bits-1);
    const double fullScaleDb = val2db(fullScaleSigned);
    std::vector<double> px, FirDftMagnitude, FirDftPhase;
    for (size_t i = 0; i < firDft.size(); i++) {
        px.push_back(double(i) / N / getTimeStepInUs (ch));
        FirDftMagnitude.push_back(val2db(std::abs(firDft[i]) * dftScaler) - fullScaleDb);
        FirDftPhase.push_back(std::arg(firDft[i]));
    }
    WaveformDataAll wda;
    wda[0] = WaveformData(px, FirDftMagnitude, "Filter FFT", filterColorIndex);
    //  wda[1] = WaveformData(px, FirDftPhase, "Filter phase", filterColorIndex+1);
    wp->setWaveformData(wda);
    wp->disableLegend();

    QwtText xTitle(tr("Frequency, MHz"));
    QwtText yTitle(tr("Amplitude, dB"));
    xTitle.setFont(font());
    yTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->setAxisTitle(QwtPlot::yLeft, yTitle);
    wp->replot();
}


void dcconf::replotTEST ()
{
    //TODO
    WaveformPlot *wp = ui->plotTEST; // Test impulse

    if (ch >= newDecoderConfig.fir.size()) {
        wp->clear();
        return;
    }
    const intfir_t &fir = newDecoderConfig.fir[ch];

    const int firlen = fir.klen;
    const int firdelay = fir.offs;
    //    const int firbits = fir.bits;

    const bool has_signal = (newDecoderConfig.calavg.nent.size() > ch) && (newDecoderConfig.calavg.nent[ch] > 0);

    const std::vector<double> &inp = getCalAvgY(ch);
    const size_t inpSize = inp.size();

    const int rlen = inpSize + firlen - 1;
    Q_ASSERT(rlen <= (int)(ADCM_MAX_N_SAMP + FIRMAXLEN));
    std::vector<double> testResult;

    if (firlen > 0) {
        //  printf("convolve: %d, %d, %d\n", firlen, inplen, rlen);
        std::vector<double> firCoef;
        firCoef.assign(&newDecoderConfig.fir[ch].coef[0], &newDecoderConfig.fir[ch].coef[firlen]);
        //      Convolver::convolve (cfg->dec.fir[ch].coef, firlen, &inp[0], inp.size(), &res[0]); // FIXME: use std::vector
        testResult = Convolver::convolve (firCoef, inp);
    }
    const int lat = data_latency () + firdelay;
    const float timeStepInUs = getTimeStepInUs (ch);     // nanoseconds

    double ascale=0;
    //  if (firbits > 1)
    //    ascale = 1. / BITS2MAXINT (firbits);
    //  else
    ascale = 1;

    const size_t N = testResult.size();
    std::vector<double> TSTpx;
    std::vector<double> TSTpy;
    for (size_t i = 0; i < N; i++) {
        TSTpx.push_back((int(i) - lat) * timeStepInUs);
        TSTpy.push_back(testResult[i] * ascale);
    }

    WaveformDataAll wda;
    wda[0] = WaveformData(TSTpx, TSTpy, "Result waveform", testColorIndex);
    wp->setWaveformData(wda);
    wp->disableLegend();

    QwtText xTitle(tr("Time")+QString::fromUtf8(", μs"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();

    //    QwtText yTitle(tr("Amplitude, mV"));
    //    yTitle.setFont(font());
    //    wp->setAxisTitle(QwtPlot::yLeft, yTitle);

    //printf("replot ok\n");

    if (has_signal && (N > 0)) {
        // find peak value
        int peakIndex = 0;
        double peakValue = testResult[0];
        for(size_t i=0; i<testResult.size(); i++) {
            if (testResult[i] > peakValue) {
                peakIndex = i;
                peakValue = testResult[peakIndex];
            }
        }
        double err[3] = { 0, 0, 0 };
        int fs = AllConfig::get_dec_fitside(conf_mem, newDecoderConfig, ch);
        int xl = peakIndex - fs;
        int xr = peakIndex + fs;
        if (xl < 0) xl = 0;
        if (xr >= int(N)) xr = int(N) - 1;
        qDebug() << QString("peakIndex=%1, fit range %2:%3").arg(peakIndex).arg(xl).arg(xr);
        //    fitstatus_t fitstatus;
        //    fitstatus.niter = 100;
        //    FindGausParam (0, TSTpy, rlen, par, &fitstatus);
        std::vector<double> fitData;
        fitData.assign(&TSTpy[xl], &TSTpy[xr]);
        GausPar par = GaussianApproximator::FindGausApproxShifted (fitData);
        //    QuickFitGaus (&TSTpy[xl], xr-xl, par, &fitstatus);
        //    printf("(2) par = %f, %f, %f; xp=%d, fs=%d, newrange=%d-%d\n", par[0], par[1]+xl, par[2], xp, fs, xl, xr);
        //    printf("(3) par err = %f, %f, %f\n", (apar[0]-par[0]), (apar[1]-(par[1]+xl)), (apar[2]-par[2]));

        //	printf ("Expected latency: %f\n", (double)lat);
        //	printf ("Measured latency: %f\n", par[1]+xl);
        const QString plusMinus = QString::fromUtf8("\u00B1");
        ui->tlfamp->setText (QString("Ampl: %1 %2 %3")
                             .arg(par.a, 0, 'f', 1)
                             .arg(plusMinus)
                             .arg(err[0], 0, 'f', 1));
        ui->tlfpos->setText (QString("Pos: %1 %2 %3")
                             .arg((par.x0+xl - lat) * timeStepInUs, 0, 'f', 1)
                             .arg(plusMinus)
                             .arg(err[1] * timeStepInUs, 0, 'f', 2));
        ui->tlfwid->setText (QString("Width: %1 %2 %3")
                             .arg(GausFWHM * par.sigma * timeStepInUs, 0, 'f', 2)
                             .arg(plusMinus)
                             .arg(err[2] * timeStepInUs, 0, 'f', 2));
    } else {
        ui->tlfamp->clear ();
        ui->tlfpos->clear ();
        ui->tlfwid->clear ();
    }
}

void dcconf::Restore ()
{
    //printf("Restore:%d\n", restoring);

    if (restoring)
        return;

    //  printf("Restore (ch=%d)\n", ch);
    restoring = 1;
    disable_control_slots = 1;

    //const char *chname[ADC_NCH] = {"A", "B", "C", "D", "E", "F", "G", "H"};
    QString s;
    s = "[";
    if (adcConfig.trig[ch] & ADCM_CCTRL_BIT_M)
        s += ustr_ch_label_M;
    else
        s += " ";
    if (adcConfig.trig[ch] & ADCM_CCTRL_BIT_S)
        s += ustr_ch_label_S;
    else
        s += " ";
    s += "]";
    tlchannel->setText (s);
    tlchannel->setEnabled (adcConfig.trig[ch] & ADCM_CCTRL_BIT_CMPEN);
    //  tlchannel->setPaletteForegroundColor(QColor(rgbcmap[ch][0],rgbcmap[ch][1],rgbcmap[ch][2]));

    sbpos->setValue ((int) newDecoderConfig.ppos[ch]);
    sbwid->setValue ((int) newDecoderConfig.pwid[ch]);
    sbexp->setValue ((int) newDecoderConfig.expfall[ch]);
    sbblrmd->setValue ((int) (100 * newDecoderConfig.baserc[ch]));
    /*
  CreateFIR();
  calcRES();

*/
    if (newDecoderConfig.nch() > 0) {
//        spinBoxChannel->setMaxValue (newDecoderConfig.nch() - 1);
        spinBoxChannel->setMaxValue (adcConfig.size() - 1);
    } else {
        spinBoxChannel->setMaxValue (0);
        spinBoxChannel->setEnabled(false);
    }

    const double noiseRmsAbsMin = 1. / sqrt (12); // quantization noise
//    double noiseRms = std::max(noiseRmsAbsMin, cfg->getcalnoiserms(ch));
    const double fftGain = sqrt(adcConfig.recwin[ch]);
    INPFFT_MIN = noiseRmsAbsMin / fftGain;  // set to noise level
    INPFFT_MAX = adcConfig.adc_rawdata_maxval(ch) / fftGain;    // set to maximum signal level
    //    ui->plotINPFFT->setAxisScale (QwtPlot::yLeft, INPFFT_MIN, INPFFT_MAX);
    //    QwtScaleEngine *scaleEngine = ui->plotINPFFT->axisScaleEngine(QwtPlot::yLeft);
    //    scaleEngine->setReference(val2db(INPFFT_MIN));
    int snf_db_min = (int) val2db(INPFFT_MIN);
    int snf_db_max = (int) val2db(INPFFT_MAX);
    sbsnf->setRange (snf_db_min, snf_db_max);
    sbrnf->setRange (snf_db_min, snf_db_max);
    if (newDecoderConfig.calavg.nf.size() > ch)
        sbsnf->setValue ((int) (newDecoderConfig.calavg.nf[ch]));
    if (newDecoderConfig.pnf.size() > ch)
        sbrnf->setValue ((int) (newDecoderConfig.pnf[ch]));
    //    const double defaultNoiseFloorDb = val2db(noiseRms / fftGain);
    //    sbsnf->setValue ((int) (defaultNoiseFloorDb));
    //    sbrnf->setValue ((int) (defaultNoiseFloorDb));

    char str[40];
    double nrms = 0;

    int n = (newDecoderConfig.calzero.nent.size() > ch) ? newDecoderConfig.calzero.nent[ch] : 0;
    if (n <= 0)
        sprintf (str, "Baseline: uncal.");
    else
        sprintf (str, "Baseline: %.1f", newDecoderConfig.calzero.base[ch] / n);
    tlbl->setText (str);
    if (n <= 0) {
        sprintf (str, "RMS Noise: uncal.");
        nrms = 1000;
    } else {
        nrms = sqrt (newDecoderConfig.calzero.rms[ch] / n);
        sprintf (str, "RMS Noise: %.1f", nrms);
    }
    tlnrms->setText (str);

    if (n <= 0) {
        tlbl->setPaletteForegroundColor (Qt::red);
        tlnrms->setPaletteForegroundColor (Qt::red);
    } else {
        tlbl->unsetPalette ();
        if (nrms < 10.0)
            tlnrms->unsetPalette ();
        else
            tlnrms->setPaletteForegroundColor (Qt::red);
    }


    if (newDecoderConfig.puthr.size() > ch)
        sbput->setValue ((int) newDecoderConfig.puthr[ch]);
    if (newDecoderConfig.pumindist.size() > ch)
        sbspumd->setValue ((int) newDecoderConfig.pumindist[ch]);

    sbblr1->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    sbblr2->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    if (newDecoderConfig.blr1.size() > ch)
        sbblr1->setValue (newDecoderConfig.blr1[ch]);
    if (newDecoderConfig.blr2.size() > ch)
        sbblr2->setValue (newDecoderConfig.blr2[ch]);
    sbi1->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    sbi2->setRange (-1*ADCM_MAX_N_SAMP, ADCM_MAX_N_SAMP);
    if (newDecoderConfig.int1.size() > ch)
        sbi1->setValue (newDecoderConfig.int1[ch]);
    if (newDecoderConfig.int2.size() > 0)
        sbi2->setValue (newDecoderConfig.int2[ch]);

    if (newDecoderConfig.fir.size() > ch) {
        intfir_t *fir = &newDecoderConfig.fir[ch];
        if (fir->offs > fir->klen)
            fir->offs = fir->klen / 2;
        sbfirnp->setValue (fir->klen);
    }

    if (newDecoderConfig.fwinf.size() > ch) {
        win_type = newDecoderConfig.fwinf[ch];
        int win_id = win_id_by_type((fwinf_t)win_type);
        win_type = win_type_by_id(win_id); // convert twice to make sure value is correct
        comboBoxWindowingFunction->setCurrentItem (win_id);
    }

    disable_control_slots = 0;

    on_pushButtonGenerateFir_clicked ();
    Redraw ();
    restoring = 0;
    //  printf("Restore... Ok\n");

}

void dcconf::Redraw ()
{
    //printf("Redraw\n");

    replotINP ();
    replotINPFFT ();
    replotFIR ();
    replotFIRFFT ();
    replotTEST ();
}

void dcconf::on_pushButtonGenerateFir_clicked ()
{
    if (restoring)
        return;
    //printf("CreateFIR\n");

    {
        int navg = (newDecoderConfig.calavg.nent.size() > ch) ? newDecoderConfig.calavg.nent[ch] : 0;
        //  int nzero = cfg->dec.calzero.nent[ch];
        if (navg <= 0) {
            //    log_notice ("ch %d, No average waveform, please calibrate average first", ch);
            return;
        }
    }
    const std::vector<double> &calAvg = getCalAvgY(ch);
    const std::vector<double> &req = getRefY(ch);
    Q_ASSERT(calAvg.size() == req.size());

    double orig_thr = db2val (newDecoderConfig.calavg.nf[ch]);
    double req_thr = db2val (newDecoderConfig.pnf[ch]);

    dblfir_t df;

    create_deconvolution_dblFIR(&df, calAvg, req, orig_thr, req_thr);

    // FIR truncation. FIR will have fixed number of samples (np)

    const int olen = df.klen;

    int np = sbfirnp->value ();
    if (np > olen)
        np = olen;
    int i1 = (olen - np) / 2;
    int i2 = (olen + np) / 2;
    if (i1 < 0)
        i1 = 0;
    if (i2 > olen)
        i2 = olen;

    for (int i = i1; i < i2; i++) {
        df.coef[i - i1] = df.coef[i];
    }

    df.klen = i2 - i1; // np
    df.offs -= i1;

    // end of FIR truncation

    // Windowing
    for (unsigned int i = 0; i < df.klen; i++) {
        df.coef[i] *= windowingFunction ((fwinf_t)win_type, i, df.klen - 1);
    }

    // After windowing, first and last samples may be near zero
    // TODO

    // Copy new fir to cfg
    newDecoderConfig.fir[ch].klen = df.klen;
    newDecoderConfig.fir[ch].offs = df.offs;
    newDecoderConfig.fir[ch].bits = FIR_COEFF_BITS;
    Make_Discrete (&df.coef[0], &newDecoderConfig.fir[ch].coef[0], df.klen, FIR_COEFF_BITS);

    /*
  if (!restoring)
    Restore ();
*/
    //  Restore(); -- is called from Restore() --
    replotFIR ();
    replotFIRFFT ();
    replotTEST ();
}

double dcconf::reqf (int i) const
{
    double ampl = 0.5 * BITS2MAXUINT (adcConfig.bits[ch]);
    return ampl * reqfunscaled(i);
}

double dcconf::reqfunscaled (int i) const
{
    // i in timebins, not ns!

    const int nl=10;

    double step = (adcConfig.size() > ch) ? adcConfig.step[ch] : 0;
    if (step <= 0)
        return (0);
    double pos = newDecoderConfig.ppos[ch] / step;
    double wid = newDecoderConfig.pwid[ch] / step;
    double value=0;
    for(int l=0;l<nl;l++)
        value += 1. * gauss (i+(1.*l/nl - .5) - pos, wid / GausFWHM);
    value /= nl;
    return value;
}

double dcconf::ca (unsigned int ch, int i) const
{
    int na = newDecoderConfig.calavg.nent[ch];
//    int nz = cfg->dec.calzero.nent[ch];
    if (na <= 0)
        na = 1;
//    if (nz <= 0)
//        nz = 1;

    //  return (1. * (1. * cfg->dec.calavg.data[ch][i] / na));
    double expfall = exp(-10.*i/newDecoderConfig.expfall[ch])/exp(-10.*20/newDecoderConfig.expfall[ch]);
    return (1. *(1.*newDecoderConfig.calavg.data[ch][i] / na)*expfall);

    //  return (1. *(1.*cfg->dec.calavg.data[ch][i] / na -
    //  1.*cfg->dec.calzero.base[ch] / nz )*expfall/* * 10.*exp(-.1*i)*(1+erf(.3*(i-5))*/);

}

double dcconf::getTimeStepInNs (int ch_num) const
{
    double tsNs;

    if (ch_num >= newDecoderConfig.nch()) return 1;

    //  ts = cfg->dec.calavg.step[ch_num];
    //  if (ts <= 0)
    tsNs = (adcConfig.size() > 0) ? adcConfig.step[ch_num] : 0;
    if (tsNs <= 0)
        tsNs = 1;

    //  qDebug() << QString("ts:%1").arg(ts);
    return tsNs;
}

double dcconf::getTimeStepInUs (unsigned int ch_num) const
{
    return getTimeStepInNs(ch_num) / 1000.;
}

//--------- Slot actions -----------------------------------------------------------------

void dcconf::on_comboBoxWindowingFunction_activated (int item)
{
    //  if (restoring) return;
    fwinf_t new_win_type = win_type_by_id(item);
    if (new_win_type == win_type) return;
    win_type = new_win_type;
    on_pushButtonGenerateFir_clicked ();
}

void dcconf::on_sbfirnp_valueChanged (int)
{
    //  if (restoring) return;
    on_pushButtonGenerateFir_clicked ();
}


void dcconf::on_sbsnf_valueChanged (int v)
{
    if (restoring)
        return;
    newDecoderConfig.calavg.nf[ch] = v;
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
    setButtons();
}

void dcconf::on_sbrnf_valueChanged (int v)
{
    if (restoring)
        return;
    newDecoderConfig.pnf[ch] = v;
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
    setButtons();
}



void dcconf::on_sbput_valueChanged (int v)
{
    newDecoderConfig.puthr[ch] = v;
    setButtons();
}

void dcconf::on_sbspumd_valueChanged (int v)
{
    newDecoderConfig.pumindist[ch] = v;
    setButtons();
}

void dcconf::on_sbi1_valueChanged (int v)
{
    int tmp, v1;

    if (disable_control_slots)
        return;

    v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi1->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.int1[ch] = v;
    if (sbi2->value () < v) {
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi2->setValue (v);
        newDecoderConfig.int2[ch] = v;
        disable_control_slots = tmp;
    }
    setButtons();
}

void dcconf::on_sbi2_valueChanged (int v)
{
    int tmp, v1;

    if (disable_control_slots)
        return;

    v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi2->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.int2[ch] = v;
    if (sbi1->value () > v) {
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbi1->setValue (v);
        newDecoderConfig.int1[ch] = v;
        disable_control_slots = tmp;
    }
    setButtons();
}


void dcconf::on_sbblr1_valueChanged (int v)
{
    int tmp, v1;

    if (disable_control_slots)
        return;

    v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr1->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.blr1[ch] = v;
    if (sbblr2->value () < v) {
        tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr2->setValue (v);
        newDecoderConfig.blr2[ch] = v;
        disable_control_slots = tmp;
    }
    setButtons();
}

void dcconf::on_sbblr2_valueChanged (int v)
{
    if (disable_control_slots)
        return;

    int v1 = v;
    if (v < -1*int(ADCM_MAX_N_SAMP)) {
        v1 = -1*int(ADCM_MAX_N_SAMP);
    }
    if (v > int(ADCM_MAX_N_SAMP)) {
        v1 = ADCM_MAX_N_SAMP;
    }

    if (v1 != v) {
        v = v1;
        int tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr2->setValue (v);
        disable_control_slots = tmp;
    }
    newDecoderConfig.blr2[ch] = v;
    if (sbblr1->value () > v) {
        int tmp = disable_control_slots;
        disable_control_slots = 1;
        sbblr1->setValue (v);
        newDecoderConfig.blr1[ch] = v;
        disable_control_slots = tmp;
    }
    setButtons();
}



void dcconf::on_sbpos_valueChanged (int v)
{
    newDecoderConfig.ppos[ch] = v;
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
}

void dcconf::on_sbwid_valueChanged (int v)
{
    newDecoderConfig.pwid[ch] = v;
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
}

void dcconf::on_spinBoxChannel_valueChanged (int inch)
{
    if (inch >= int(currentDecoderConfig.nch())) {
        spinBoxChannel->setValue(newDecoderConfig.nch());
        return;
    }
    ch = inch;
    Restore ();
}

void dcconf::on_sbexp_valueChanged (int v)
{
    newDecoderConfig.expfall[ch] = v;

    on_pushButtonGenerateFir_clicked ();
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    // replotTEST ();
}


void dcconf::on_sbblrmd_valueChanged (int v)
{
    newDecoderConfig.baserc[ch] = 0.01 * v;
}


int dcconf::data_latency () const
{
    return (int) adcConfig.latency;
}


//--------------- Button actions --------------------------------------------------------

void dcconf::on_pushButtonSaveFir_clicked()
{
    QSettings settings;
    QString currentFirFileName = settings.value("currentFirFileName").toString();
    currentFirFileName = QFileDialog::getSaveFileName (this, "Saving FIR file", currentFirFileName,
                                                       "FIR files (*.fir);;All files (*)");
    if (currentFirFileName.isNull())
        return;
    if (!currentFirFileName.endsWith(".fir"))
        currentFirFileName += ".fir";
    settings.setValue("currentFirFileName", currentFirFileName);
    int ret = writefir (currentFirFileName.toLocal8Bit(), &(newDecoderConfig.fir[ch]));
    if (ret)
        QMessageBox::warning (this, "Saving FIR filter", "The file cannot be written", QMessageBox::Cancel, Qt::NoButton);
}

void dcconf::on_pushButtonLoadFir_clicked ()
{
    QSettings settings;
    QString currentFirFileName = settings.value("currentFirFileName").toString();
    currentFirFileName = QFileDialog::getOpenFileName (this, "Loading FIR filter", currentFirFileName,
                                                       "FIR files (*.fir);;All files (*)");

    if (currentFirFileName.isNull())
        return;
    settings.setValue("currentFirFileName", currentFirFileName);

    intfir_t tmp;
    int ret = readfir (currentFirFileName.toLocal8Bit(), &tmp);

    if (ret) {
        QMessageBox::warning (this, "Loading FIR filter", "The file cannot be read", QMessageBox::Cancel, Qt::NoButton);
        return;
    }
    newDecoderConfig.fir[ch] = tmp;
    Restore ();
}

void dcconf::on_pushButtonLoadShape_clicked ()
{
    QSettings settings;
    QString currentShapeFileName = settings.value("currentShapeFileName").toString();
    currentShapeFileName = QFileDialog::getOpenFileName (this, "Loading average shape", currentShapeFileName,
                                                         "Shape files (*.shp);;All files (*)");

    if (currentShapeFileName.isNull())
        return;
    settings.setValue("currentShapeFileName", currentShapeFileName);

    // TODO: read current channel (ch)
    CalAvgConfig tmpCalAvg = newDecoderConfig.calavg;
    QSettings s(currentShapeFileName, QSettings::IniFormat);
    tmpCalAvg.readFromIni(&s);
    newDecoderConfig.calavg = tmpCalAvg;
    Restore ();
}

void dcconf::on_pushButtonSaveShape_clicked ()
{
    QSettings settings;
    QString currentShapeFileName = settings.value("currentShapeFileName").toString();
    currentShapeFileName = QFileDialog::getSaveFileName (this, "Saving average shape", currentShapeFileName,
                                                         "Shape files (*.shp);;All files (*)");

    if (currentShapeFileName.isNull())
        return;

    if (!currentShapeFileName.endsWith(".shp"))
        currentShapeFileName += ".shp";

    // TODO: write current channel (ch)
    QSettings s(currentShapeFileName, QSettings::IniFormat);
    newDecoderConfig.calavg.writeToIni(&s);
    settings.setValue("currentShapeFileName", currentShapeFileName);
}


bool dcconf::modified() const
{
    return currentDecoderConfig != newDecoderConfig;
}

void dcconf::apply()
{
    emit decoderConfigModified(newDecoderConfig);
    currentDecoderConfig = newDecoderConfig;
    setButtons();
}

void dcconf::setButtons ()
{
    buttonBox->button(QDialogButtonBox::Apply)->setEnabled(modified());
    buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(newDecoderConfig != AllConfig().dec);
}

void dcconf::restoreDefaults()
{
    newDecoderConfig = AllConfig().dec;
    setButtons();
    Restore();
}

void dcconf::on_buttonBox_clicked(QAbstractButton* button)
{
    switch(buttonBox->standardButton(button)) {
    case QDialogButtonBox::Ok :
    case QDialogButtonBox::Apply :
        apply();
        break;
    case QDialogButtonBox::RestoreDefaults :
        restoreDefaults();
        break;
    default:;
    }
}
