
#include <cmath>

#include <QTimer>
#include <QLineEdit>
#include <QPushButton>
#include <QSpinBox>
#include <QCheckBox>

#include "dsplib/dsplib.h"
#include "histplot.h"
#include "histdialog.h"
#include "th1.h"

const int pxf_m = 4;

HistDialog::HistDialog (QWidget* parent, const th1Ptr histPtr)
    : QDialog(parent),
      h(histPtr)
{
    setupUi(this);
    alloc_nbins = 0;
    modseq = -1;

    updatetimer = new QTimer (this);
    connect (updatetimer, SIGNAL (timeout ()), SLOT (update_view ()));
    updatetimer->start (500, FALSE);

    disable_slots = 0;
    adjustSize();
}
/*
double hist::scaled_xmin() const
{
  return (xmin-base)*scaler;
}

double hist::scaled_xmax() const
{
  return (xmax-base)*scaler;
}

void update_scaled_xrange() {
  sxmin = scaled_xmin();
  sxmax = scaled_xmax();
}
*/

//void hist::set_h( const HistPtr histPtr )
//{
//  h = histPtr;
//}
//
const th1Ptr HistDialog::get_h () const
{
    return h;
}

void HistDialog::update_view ()
{
    if (!isShown()) return;

    update_controls();
    update();
}

void HistDialog::update_controls()
{
    bool e = (h != NULL);

    hp->setEnabled(e);
    le1->setEnabled(e);
    le2->setEnabled(e);
    le3->setEnabled(e);
    pushButtonSave->setEnabled(e && (h->get_nent() > 0));

    //  update();

    if (h) {
        le1->setText (QString::number(h->get_nent()));
        le2->setText (QString::number(h->get_over ()));
        le3->setText (QString::number(h->get_under ()));
        le10->setText (QString::number(h->get_sum (), 'e', 3));
    } else {
        le1->clear ();
        le2->clear ();
        le3->clear ();
        le10->clear ();
    }
    if (fit_ok) {
        //    sprintf(str, "%d", value);
        //    le6->setText(str);
        le4->setText (QString::number(mean, 'e', 3));
        le5->setText (QString::number(sigma * GausFWHM, 'e', 2));
        le9->setText (QString::number(sigma, 'e', 2));
        le6->setText (QString("%1 %").arg((mean > 0) ? sigma * GausFWHM / mean * 100. : 0, 0, 'f', 2));
    } else {
        le4->clear ();
        le5->clear ();
        le6->clear ();
        le9->clear ();
    }
    bool fit_enable = h?(h->get_nent() > 0 && h->fit_enabled()):false;
    le4->setEnabled(fit_enable);
    le5->setEnabled(fit_enable);
    le6->setEnabled(fit_enable);
    le9->setEnabled(fit_enable);
}

void HistDialog::update ()
{
    if (!h)
        return;
    int now_modseq = h->get_modseq();
    if (modseq == now_modseq) return;

    setCaption(getName ());

    const double xmin = h->get_xmin ();
    const double xmax = h->get_xmax ();

    const unsigned int nbins = h->get_nbins ();

    if ((nbins == 0) || (xmin == xmax))
        return;

    if (alloc_nbins != nbins) {
        // reallocate
        px.resize(nbins);
        py.resize(nbins);
        pxf.resize(nbins*pxf_m);
        pf.resize(nbins*pxf_m);
        alloc_nbins = nbins;
    }

    // Reset histogram arrays
    const double step = 1. / nbins * (xmax-xmin);
    const double stepFit = 1. / nbins / pxf_m *(xmax-xmin);
    for (unsigned int i = 0; i < nbins; i++) {
        px[i] =  xmin + (double) (i+1) * step;
    }
    py = h->get_all_bins();

    // Fit
    const bool fit_enable = h->fit_enabled();
    const bool need_fit = fit_enable && (h->get_nent () > 10);
    fit_ok = false;
    value=0;mean=0;sigma=0;norm=0;
    if (need_fit) {
        int ret = h->fitgaus (&value, &mean, &sigma, &norm);
        if (ret==0 && norm < 2 && sigma != 0) fit_ok = true;
    }

    // Reset fit arrays
    for (unsigned int i = 0; i < pxf_m*nbins; i++) {
        double x = xmin + (double) (i+.5) * stepFit;
        pxf[i] = x;
        if (fit_ok)
            pf[i] = value*gauss(x-mean, sigma);
        else
            pf[i] = 0;
    }

    // Reset grafs in histPlot
    if (hp) {
        if(fit_ok){
            hp->setpoints (px, py);
            hp->set_fit_points (pxf, pf);
        } else {
            hp->set_fit_points (pxf, pf);
            hp->setpoints (px, py);
        }
    }
    update_bin_range();
    modseq = now_modseq;

    if (hp) hp->replot ();
}

HistDialog::~HistDialog ()
{
    delete updatetimer;
}

int HistDialog::getid () const
{
    return (myid);
}

void HistDialog::setid (int id)
{
    myid = id;
}

/*
void hist::newbase ()
{
  if (disable_slots)
    return;
  char nptr[32], *endptr;
  strncpy (nptr, le7->text (), 32);
  double d = strtod (nptr, &endptr);
  if (endptr > nptr)
    base = d;
  Restore ();
  update ();
}

void hist::newscaler ()
{
  if (disable_slots)
    return;
  char nptr[32], *endptr;
  strncpy (nptr, le8->text (), 32);
  double d = strtod (nptr, &endptr);
  if (endptr > nptr)
    scaler = d;
  Restore ();
  update ();
}
*/
void HistDialog::Restore ()
{
//    disable_slots = 1;
//    disable_slots = 0;
}

QString HistDialog::getName () const
{
    if (h)
        return h->getname();
    else
        return "unknown";
}

HistPlot * HistDialog::getplot() const
{
    return hp;
}

void HistDialog::show() {
    update_controls();
    update();
    QDialog::show();
}

void HistDialog::get_fit_param(bool *ffitok, double *fvalue, double *fmean, double *fsigma, double *fnorm) const
{
    if (fmean)
        *fmean = mean;
    if (fvalue)
        *fvalue = value;
    if (fsigma)
        *fsigma = sigma;
    if (fnorm)
        *fnorm = norm;
    if (ffitok)
        *ffitok = fit_ok;
}

void HistDialog::update_bin_range()
{
    if (!hp || !h) return;
    double ar_min=0, ar_max=0;
    h->update_autorange();
    h->getrange(&ar_min, &ar_max);
    hp->set_x_range(ar_min, ar_max);
}

void HistDialog::on_pushButtonSave_clicked()
{
    if (h)
        h->save_to_file (this);}
