#include <vector>
#include <cstdio>
#include <cmath>

#include <QDebug>

#include "th1.h"
#include "histmanager.h"
#include "AdcmChbuf.h"
#include "ch_label.h"
#include "adcmlib/util.h"
#include "adcmlib/adcm_ev.h"
#include "adcmlib/adcm_dws.h"
#include "adcmraw/adcm_mem.h"


#define PREFIX "histmgr: "
#define HISTMGR_DEBUG_LEVEL 1
#define HMDL HISTMGR_DEBUG_LEVEL /* shortcut */



bool cfg_channel_may_have_data(const AdcConfig &adcConfig, int ch)
{
    Q_ASSERT(ch < adcConfig.size());
    //  bool en = false;
    int trig = adcConfig.trig[ch];
    //  int trigmode = cfg.adc.trigbits;
    if (adcConfig.trigbits & (ADCM_TRIGBIT_INT_SIMPLE|ADCM_TRIGBIT_EXT))
        return trig & ADCM_CCTRL_BIT_CMPEN;

    QVector<bool> en_m(adcConfig.size()), en_s(adcConfig.size());
    bool en_any_m = false, en_any_s = false;
    int mask_m = (ADCM_CCTRL_BIT_M | ADCM_CCTRL_BIT_CMPEN);
    int mask_s = (ADCM_CCTRL_BIT_S | ADCM_CCTRL_BIT_CMPEN);
    for(int i=0; i<adcConfig.size(); i++) {
        en_m[i] = ((trig & mask_m) == mask_m);
        en_s[i] = ((trig & mask_s) == mask_s);
        if (en_m[i]) en_any_m = true;
        if (en_s[i]) en_any_s = true;
    }
    if (en_any_m && en_any_s)
        return (trig & mask_m) || (trig & mask_s);
    return false;
}

HistManager::HistManager (QObject *parent)
    : QObject(parent),
      hist_nch(0),
      conf_mem(new conf_mem_t),
      mutex(new QMutex(QMutex::Recursive))
{
    h_dT_index = -1; // Time Ts-Tm fine
    h_dTmax_index = -1; // Time Ts-Tm coarse
    h_Tmr_index = -1; // Rec. quality Tm
    h_Tsr_index = -1; // Rec. quality Ts
    h_Ts_index = -1; //Time Ts
    h_wm_index = -1; // Filt. pulse width Wm
    h_ws_index = -1; // Filt. pulse width Ws
    h_slideev_index = -1; // Raw pulses / event
    h_pulsev_index = -1; // Fitted pulses / event
    h_niter_index = -1; // Fitter iterations / event
    h_evdist_index = -1; // Channel time, rel. Spill
    h_decodetime_index = -1; // Event decode time
    h_T_index = -1; // Ch Time rel. Master
    h_Int_index = -1; // Integral
    h_bl_index = -1; // BaseLine
    h_dT_Ev_index = -1; // Time Ev(i)-Ev(i-1)
    h_n_total = 0;
    //    print();
}

HistManager::~HistManager ()
{
    hl.clear();
    delete conf_mem;
}

int HistManager::nch() const
{
    return conf_mem->nch;
}

void HistManager::print() const
{
    qDebug() << QString("hl n:%1").arg(hl.count());
}

void HistManager::create_histograms()
{
    hist_nch = nch();

    unsigned int i=0;
    h_dT_index = i; i++;
    h_dTmax_index = i; i++;
    h_Ts_index = i; i++;
    h_Tmr_index = i; i++;
    h_Tsr_index = i; i++;
    h_decodetime_index = i; i++;
    h_slideev_index = i; i++;
    h_pulsev_index = i; i++;
    h_niter_index = i; i++;
    h_evdist_index = i; i += nch();
    h_wm_index = i; i++;
    h_ws_index = i; i++;
    h_T_index = i; i+= nch();
    h_Int_index = i; i+= nch();
    h_bl_index = i; i+= nch();
    h_n_total = i;

    QString s;
    const QString &str_M = ustr_ch_label_M;
    const QString &str_S = ustr_ch_label_S;
    QString str_ns = QString(", ns");

    s = QString("Time T") + str_S + " - T" + str_M + QString(" (fine)") + str_ns;
    int dT_x_scaler = 1;
    if (conf_mem->getMasterAdcBits() > 10)
        dT_x_scaler = 4;
    th1Ptr p_dT(new th1 (s, dT_x_scaler*8000, -100, 100, th1::FIT));
    hl.insert(h_dT_index, p_dT);

    s = QString("Time T") + str_S + " - T" + str_M + QString(" (coarse)") + str_ns;
    th1Ptr p_dTmax(new th1 (s, 8000, -1000, 1000, th1::FIT));
    hl.insert(h_dTmax_index, p_dTmax);

    s = QString("Rec. quality T") + str_M + str_ns;
    th1Ptr p_Tmr(new th1 (s, 200, 0, 1));
    hl.insert(h_Tmr_index, p_Tmr);

    s = QString("Rec. quality T") + str_S + str_ns;
    th1Ptr p_Tsr(new th1 (s, 200, 0, 1));
    hl.insert(h_Tsr_index, p_Tsr);

    s = QString("Time T") + str_S + str_ns;
    th1Ptr p_Ts(new th1 (s, 1000, -300, 500));
    hl.insert(h_Ts_index, p_Ts);

    th1Ptr p_dT_Ev(new th1 ("Time b/w Events, ns", 200, 0, 100000));
    hl.insert(h_dT_Ev_index, p_dT_Ev);

    //  hl.insert(h_decodetime_index, ); new th1 ("Event decode time, us", 2000, 0, 200);
    th1Ptr p_decodetime(new th1 ("Event decode time, us/us", 2000, 0, 200));
    hl.insert(h_decodetime_index, p_decodetime);

    th1Ptr p_slideev(new th1 ("Raw pulses / event", conf_mem->nch, 0, conf_mem->nch));
    hl.insert(h_slideev_index, p_slideev);

    th1Ptr p_pulsev(new th1 ("Fitted pulses / event", conf_mem->nch, 0, conf_mem->nch));
    hl.insert(h_pulsev_index, p_pulsev);

    th1Ptr p_iter(new th1 ("Fitter iterations / event", 500, 0, 500));
    hl.insert(h_niter_index, p_iter);

    s = QString("Filt. pulse width W") + str_M + str_ns;
    th1Ptr p_wm(new th1 (s, 1000, 0, 100));
    hl.insert(h_wm_index, p_wm);

    s = QString("Filt. pulse width W") + str_S + str_ns;
    th1Ptr p_ws(new th1 (s, 1000, 0, 100));
    hl.insert(h_ws_index, p_ws);

    for (int i = 0; i < nch(); i++) {
        char str[50];
        int minraw = conf_mem->getAdcMinValue(i);
        //        int midraw = conf_mem->getAdcMidValue(cfg, i);
        int maxraw = conf_mem->getAdcMaxValue(i);
        sprintf (str, "Ch %d Time, ms rel. Spill", i);
        th1Ptr p_evdist(new th1 (str, 2000, 0, 20, 0));
        hl.insert(h_evdist_index + i, p_evdist);

        sprintf (str, "Ch %d Time, ns rel. Master", i);
        th1Ptr p_T(new th1 (str, 8000, -1000, 1000, th1::FIT));
        hl.insert(h_T_index + i, p_T);

        maxraw = (maxraw+1)/2*2;
        sprintf (str, "Integral %d", i);
        th1Ptr p_Int(new th1 (str, maxraw*2, -maxraw, maxraw, th1::FIT|th1::FIT_RESOL));
        hl.insert(h_Int_index + i, p_Int);

        sprintf (str, "Baseline %d", i);
        th1Ptr p_bl(new th1 (str, maxraw, minraw, maxraw, th1::FIT|th1::FIT_RESOL));
        hl.insert(h_bl_index + i, p_bl);
    }

    //  hl.n = n;
    //  printf("init_histograms: %p\n", hl.insert(0].h);
    emit histListModified(hl);
}

void HistManager::delete_histograms()
{
    hist_nch = 0;
    hl.clear();
}

int HistManager::fill_status (const adcm_decstat_t &st)
{
    //  if (!hl.items || (hl.n < 1))
    //    return (0);
    HistList::Iterator ih = hl.find(h_decodetime_index);
    if (ih == hl.end())
        return 0;
    th1Ptr h = ih.data();
    //  th1 &h1 = hl[h_decodetime_index];
    double t = st.prof.tl[decst_n_all];
    //  if (t<0) printf("t=%f\n", t);
    //  stime_t tnow = stime();
    //  double t = tnow-tprev;
    //  tprev=tnow;
    if (t>0)
        h->filln(t*1e6, (int)(t*1e6));
    return 0;
}

int HistManager::fill (const AdcmEvent &ev)
{
    QMutexLocker locker(mutex);
    //  if (!hl) return (0);
    //  if (!hl.items || (hl.n < 1))
    //    return (0);
    const QVector<double> step_ns(conf_mem->nch, conf_mem->getMasterAdcStepNs());

    HistList::Iterator ihdt = hl.find(h_dT_index);
    HistList::Iterator ihdtmax = hl.find(h_dTmax_index);
    HistList::Iterator ihtmr = hl.find(h_Tmr_index);
    HistList::Iterator ihtsr = hl.find(h_Tsr_index);
    HistList::Iterator ihts = hl.find(h_Ts_index);
    HistList::Iterator ihslideev = hl.find(h_slideev_index);
    HistList::Iterator ihpulsev = hl.find(h_pulsev_index);
    HistList::Iterator ihniter = hl.find(h_niter_index);
//    HistList::Iterator ihevdist = hl.find(h_evdist_index);
    HistList::Iterator ihwm = hl.find(h_wm_index);
    HistList::Iterator ihws = hl.find(h_ws_index);
    HistList::Iterator iend = hl.end();


    int cm[ADCM_N_CH], cs[ADCM_N_CH];
    double tm[ADCM_N_CH], ts[ADCM_N_CH];
    double tmr[ADCM_N_CH], tsr[ADCM_N_CH];
    double wm[ADCM_N_CH], ws[ADCM_N_CH];
    unsigned int mNumbers = 0, sNumbers = 0;
    unsigned int pulsesNumbers = 0;
    unsigned int npuls_fitok = 0;
    for (int c = 0; c < std::min(nch(), (int)ADCM_N_CH); c++) {
        const struct adcm_chbuf_t *cb = ev.get_ev_chbuf_const(c);
        if (!cb) continue;
        unsigned int trig = adcConfig.trig[c];

        if (cb->flags & ADCM_CH_HAS_DATA) {
            pulsesNumbers++;
            //	  printf("ev %d, ch %d, pulsesNumbers=%d\n", adcm_get_ev_evnum(ev), c, pulsesNumbers);
            HistList::Iterator ihevdist = hl.find(h_evdist_index + c);
            if (ihevdist != iend)
                ihevdist.data()->fill(cb->ts * step_ns[c] / 1e6); // convert ns to ms
        }
        unsigned int inte_required_flags = ADCM_CH_EVAL_BL | ADCM_CH_EVAL_INT;
        if ((cb->flags & inte_required_flags) == inte_required_flags)
        {
            HistList::Iterator ihi = hl.find(h_Int_index + c);
            HistList::Iterator ihb = hl.find(h_bl_index + c);
            if (ihb != iend)
                ihb.data()->fill (cb->b);
            if (ihi != iend)
                ihi.data()->fill (cb->i - cb->b);
        }

        if (cb->flags & (ADCM_CH_EVAL_TFAST | ADCM_CH_EVAL_TFINE))
            if (!(cb->flags & ADCM_CH_DT_CUT)) {
                npuls_fitok++;
                //  	  if (hw) hw->fill(cb->pw);
                if (ihniter != iend) ihniter.data()->fill(cb->pniter);

                if (trig & ADCM_CCTRL_BIT_M) {
                    double step = 1.;
                    double rt = cb->pt;
                    rt = rt - trunc(rt/step)*step;
                    if (rt<0) rt+=step;
                    cm[mNumbers] = c;
                    tmr[mNumbers] = rt;
                    tm[mNumbers] = cb->pt;
                    wm[mNumbers] = cb->pw;
                    mNumbers++;
                }
                if (trig & ADCM_CCTRL_BIT_S) {
                    double step = 1.;
                    double rt = cb->pt;
                    rt = rt - trunc(cb->pt/step)*step;
                    if (rt<0) rt+=step;
                    cs[sNumbers] = c;
                    tsr[sNumbers] = rt;
                    ts[sNumbers] = cb->pt;
                    ws[sNumbers] = cb->pw;
                    sNumbers++;
                }
            }
    }
    // fill from cache
    {
        //      if (htm) {
        //        for (unsigned int im = 0; im < mNumbers; im++)
        //          htm->fill (tm[im] * step_ns);
        //      }
        if (ihtmr != iend) {
            for (unsigned int im = 0; im < mNumbers; im++)
                ihtmr.data()->fill (tmr[im]);
        }
        if (ihwm != iend) {
            for (unsigned int im = 0; im < mNumbers; im++)
                ihwm.data()->fill (wm[im] * step_ns[cm[im]]);
        }

        if (ihtsr != iend) {
            for (unsigned int is = 0; is < sNumbers; is++)
                ihtsr.data()->fill (tsr[is]);
        }
        if (ihts != iend) {
            for (unsigned int is = 0; is < sNumbers; is++)
                ihts.data()->fill (ts[is] * step_ns[cs[is]]);
        }
        if (ihws != iend) {
            for (unsigned int is = 0; is < sNumbers; is++)
                ihws.data()->fill (ws[is] * step_ns[cs[is]]);
        }

        for (unsigned int im = 0; im < mNumbers; im++) {
            for (unsigned int is = 0; is < sNumbers; is++) {
                double time_m = tm[im]*step_ns[cm[im]];
                double time_s = ts[is]*step_ns[cs[is]];
                double dt = time_s-time_m;
                //          printf("cm[%d]=%d, cs[%d]=%d, time_m=%e, time_s=%e, dt=%e\n", im, cm[im], is, cs[is], time_m, time_s, dt);
                if (ihdt != iend) ihdt.data()->fill (dt);
                if (ihdtmax != iend) ihdtmax.data()->fill (dt);
                HistList::Iterator iht = hl.find(h_T_index + cs[is]);
                if (iht != iend)
                    iht.data()->fill (dt);
            }
        }

    } // end fill from cache

    if (ihpulsev != iend) ihpulsev.data()->fill(npuls_fitok);
    if (ihslideev != iend) ihslideev.data()->fill(pulsesNumbers);
    return (0);
}

int HistManager::fillEvDt (const AdcmEvent &ev, quint64 prevEvTs)
{
    QMutexLocker locker(mutex);

    HistList::Iterator ih_dT_Ev = hl.find(h_dT_Ev_index);
    double dT = ev.getEventTimestamp() - prevEvTs; // in steps
    dT *= conf_mem->getMasterAdcStepNs(); // in ns
//    qDebug() << ev.getEventNumber() << dT;
//    if(dT > 10000)
//        qWarning() << "too big dT=" << dT;

    if(ih_dT_Ev != hl.end())
        ih_dT_Ev.data()->fill(dT);
    return (0);
}


void HistManager::clear (int index)
{
    QMutexLocker locker(mutex);
    if(index==-1){
        Q_FOREACH(th1Ptr h, hl) {
            h->clear();
        }
    } else {
        HistList::iterator th1It = hl.find(index);
        Q_ASSERT(th1It!=hl.end());
        th1It->data()->clear();
    }
}


bool HistManager::get_hist_enable(int index) const
{
    bool en_dt = false;
    std::vector<bool> en_ch, en_m, en_s;
    en_ch.resize(ADC_NCH);
    en_m.resize(ADC_NCH);
    en_s.resize(ADC_NCH);
    bool en_any_m= false, en_any_s = false, en_any;

    for(int i=0;i<nch();i++) {
        int trig = adcConfig.trig[i];
        int mask_m = (ADCM_CCTRL_BIT_M | ADCM_CCTRL_BIT_CMPEN);
        int mask_s = (ADCM_CCTRL_BIT_S | ADCM_CCTRL_BIT_CMPEN);
        en_m[i] = ((trig & mask_m) == mask_m);
        en_s[i] = ((trig & mask_s) == mask_s);
        // bit ADCM_CCTRL_BIT_TEST is ignored
        if (en_m[i]) en_any_m = true;
        if (en_s[i]) en_any_s = true;
        en_ch[i] = (en_m[i]||en_s[i]);
        // enable unconditionally in simple or external trigger mode
        en_ch[i] = cfg_channel_may_have_data(adcConfig, i);
    }
    en_any = (en_any_m || en_any_s);
    en_dt = (en_any_m && en_any_s);
    if (index == h_dT_index) return en_dt;
    if (index == h_dTmax_index) return en_dt;
    if (index == h_Tmr_index) return en_any_m;
    if (index == h_Tsr_index) return en_any_s;
    if (index == h_Ts_index) return en_any_s;
    if (index == h_decodetime_index) return (DECODER_PROFILING != 0);
    if (index == h_slideev_index) return en_any;
    if (index == h_pulsev_index) return en_any;
    if (index == h_niter_index) return en_any;
    if (index == h_wm_index) return en_any_m;
    if (index == h_ws_index) return en_any_s;
    if ((index >= h_evdist_index) && (index < h_evdist_index+nch())) return en_ch[index-h_evdist_index];
    if ((index >= h_T_index) && (index < h_T_index+nch())) return en_any_m && en_s[index-h_T_index];
    if ((index >= h_Int_index) && (index < h_Int_index+nch())) return en_ch[index-h_Int_index];
    if ((index >= h_bl_index) && (index < h_bl_index+nch())) return en_ch[index-h_bl_index];
    if (index == h_dT_Ev_index) return true;
    return true;
}

void HistManager::setAdcConfig(const AdcConfig &r)
{
    QMutexLocker locker(mutex);
    adcConfig = r;
}

void HistManager::setAdcConfMem(const conf_mem_t &confMem)
{
    QMutexLocker locker(mutex);
    bool modified = (*conf_mem != confMem);
    *conf_mem = confMem;
    if (modified) {
        delete_histograms();
        create_histograms();
    }
}

void HistManager::addHist(HistManager *hmgr)
{
    Q_ASSERT(hmgr->getHistList().size() == hl.size());
    QMapIterator<int, th1Ptr> sumMapIt(hl);
    QMapIterator<int, th1Ptr> addMapIt(hmgr->getHistList());
    while(sumMapIt.hasNext()){
        sumMapIt.next();
        addMapIt.next();
        Q_ASSERT(hmgr->getHistList().contains(sumMapIt.key()));

//        const th1Ptr add = hmgr->getHistList().value(sumMapIt.key());
//        sumMapIt.value()->add(add.data());
        sumMapIt.value()->add(addMapIt.value().data());
    }
}

void HistManager::lock()
{
    if(!mutex->tryLock()){
        qDebug()<<"fail to lock";
        mutex->lock();
    }
    QMapIterator<int, th1Ptr> mapIt(hl);
    while(mapIt.hasNext()){
        mapIt.next();
        mapIt.value()->lock();
    }
}

void HistManager::unlock()
{
    QMapIterator<int, th1Ptr> mapIt(hl);
    while(mapIt.hasNext()){
        mapIt.next();
        mapIt.value()->unlock();
    }
    mutex->unlock();
}
