
#include <cmath>
#include <vector>
//#include <algorithm>

#include <QDateTime>
#include <QDebug>
#include <QDir>
#include <QFile>
#include <QFileDialog>
#include <QMutex>
#include <QMutexLocker>
#include <QRegExp>
#include <QSettings>
#include <QTextStream>

#include "dsplib/fitter.h"
#include "th1.h"

th1::th1 (const QString &name, int fnbins, double fmin, double fmax, int foptions):
    fn(name),
    nbins(fnbins),
    explorerManagerMutex(new QMutex(QMutex::Recursive))
{
    options = foptions;
    if (0==(options & NO_AUTORANGE))
        options |= AUTORANGE;
    mod=0;
    py.resize(nbins);
    xmin = fmin;
    xmax = fmax;
    setrange (fmin, fmax);
    clear ();
}

th1::~th1 ()
{
}

void th1::add(const th1 *summand)
{
    QMutexLocker lock(explorerManagerMutex);
    Q_ASSERT(summand->get_nbins()==get_nbins());
//    std::transform(py.begin(), py.end(), summand->get_all_bins().begin(),
//                   py.begin(), std::plus<double>());
    std::vector<double>::iterator itAdd = summand->get_all_bins().begin();
    for(std::vector<double>::iterator itSum=py.begin(); itSum!=py.end();++itSum){
        double add = *itAdd;
        double sum = *itSum;
        sum +=add;
        *itSum = sum;
//        *itSum += *itAdd;
        ++itAdd;
    }

    bool isModified=false;
    if(summand->get_under()){
        isModified=true;
        under += summand->get_under();
    }
    if(summand->get_over()){
        isModified=true;
        over += summand->get_over();
    }
    if(summand->get_nent()){
        isModified=true;
        nent += summand->get_nent();
    }

    if(isModified)
        mod++;
}

void th1::lock()
{
    explorerManagerMutex->lock();
}

void th1::unlock()
{
    explorerManagerMutex->unlock();
}

void th1::copyEntry(const th1 *copy)
{
    Q_ASSERT(copy->get_nbins()==get_nbins());
    if(get_nbins())
        memcpy(&py[0], &(copy->get_all_bins()[0]), sizeof(double)*get_nbins());
    under=copy->get_under();
    over=copy->get_over();
    nent=copy->get_nent();
}

bool th1::autorange_enabled() const
{
    return ((options & AUTORANGE) == AUTORANGE);
}

bool th1::fit_enabled() const
{
    bool state = ((options & FIT) == FIT);
    return state;
}

void th1::set_autorange_enable(bool state) const
{
    if (state == autorange_enabled()) return;
    if (state)
        options |= AUTORANGE;
    else
        options &= ~AUTORANGE;
    update_autorange();
}

void th1::set_fit_enable(bool state) const
{
    if (state)
        options |= FIT;
    else
        options &= ~FIT;
}

double th1::bincenter (int n) const
{
    double binwidth = (xmax - xmin) / nbins;
    double result = xmin + (0.5 + n) * binwidth;
    return result;
}

int th1::findbin (double x) const
{
    double hw = xmax - xmin;
    if (hw == 0)
        hw = 1;

    int bin = (int) ((x - xmin) / hw * nbins);
    return bin;
}

void th1::setrange (double inmin, double inmax)
{
    if (inmin == xmin && inmax == xmax) return;
    if (inmin < inmax) {
        xmin = inmin;
        xmax = inmax;
    } else {
        xmin = 0;
        xmax = 1;
    }
    ar_min = xmin;
    ar_max = xmax;
    mod++;
}

void th1::getrange (double *outmin, double *outmax) const
{
    double min = ar_min;
    double max = ar_max;
    if (!autorange_enabled() || (ar_min>=ar_max)) { // no autorange
        min = xmin;
        max = xmax;
    }

    if (outmin)
        *outmin = min;
    if (outmax)
        *outmax = max;
}

void th1::fill (double x)
{
    filln (x, 1);
}

void th1::filln (double x, double count)
{
    mod++;
    int bin = findbin (x);
    if (bin < 0) {
        //    under+=(int)count;
        under++;
        return;
    }
    if ((unsigned int) bin >= nbins) {
        //    over+=(int)count;
        over++;
        return;
    }
    //  nent += (int)count;
    nent ++;
    py[bin] += count;
}

void th1::clear ()
{
    QMutexLocker lock(explorerManagerMutex);
    under = 0;
    over = 0;
    nent = 0;
    if(nbins)
        memset(&py[0], 0, sizeof(double)*nbins);
//    py.clear();
//    py.resize(nbins, 0.);
    mod++;
}

int th1::fitgaus (double *fvalue, double *fmean, double *fsigma, double *fnorm) const
{
    QMutexLocker lock(explorerManagerMutex);
    double value, mean, sigma, norm;
    int ret=0;

    unsigned int fitstart = findbin(ar_min);
    unsigned int fitend = findbin(ar_max);
    if (fitstart >= fitend || fitend > nbins) { // bad range
        fitstart=0;
        fitend=nbins;
    }
    unsigned int fitnbins = fitend-fitstart;

    //  int *inp = new int[fitnbins];
    double par[3];
    //  double err[3], chi2;
    //  for (i = 0; i < fitnbins; i++)
    //    inp[i] = (int)py[fitstart+i];
    // select fit method: simple or with MIGRAD

    //  if (nbins > 100)
    //	FindGausApprox(inp, nbins, par);
    //  else

    fitstatus_t fs;
    fs.niter = 20;
    QuickFitGaus(&py[fitstart], fitnbins, par, &fs);

    //  delete (inp);
    value = par[0];


    double binwidth = (xmax - xmin) / nbins;
    mean = xmin + (fitstart+0.5 + par[1]) * binwidth;
    sigma = par[2] * binwidth;
    norm = (value>0)?(fs.fnorm/value):1000;

    if (fvalue)
        *fvalue = value;
    if (fmean)
        *fmean = mean;
    if (fsigma)
        *fsigma = sigma;
    if (fnorm)
        *fnorm = norm;

    return (ret);
}


QString th1::getname () const
{
    return(fn);
}

bool th1::save_to_file (QWidget *parent, const QString &dir) const
{
    QString subdir;
    if(dir.isNull()){
        subdir = QDateTime::currentDateTime().toString("yyyy-MM-dd_hh-mm-ss");
    } else {
        subdir = dir;
    }

    QSettings settings;
    settings.beginGroup("Histograms");
    QString allHistDirPath = settings.value("dir").toString();
    if(allHistDirPath.isEmpty()){
        allHistDirPath = QFileDialog::getExistingDirectory(parent, "Open Histograms Directory","",
                                      QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
        if(allHistDirPath.isEmpty())
            return false;
        else
            settings.setValue("dir", allHistDirPath);
    }

    QDir histDir;
    if(!subdir.isEmpty()){
        histDir = QDir(allHistDirPath+QDir::separator()+subdir);
    } else {
        histDir = QDir(allHistDirPath);
    }

    if(!histDir.exists() && !histDir.mkpath(".")){
        qWarning()<<"Can't create directory "<<histDir.absolutePath();
        return false;
    }

    QMutexLocker lock(explorerManagerMutex);
    QString filename(fn+".txt");
    QRegExp rx("[/]");
    filename.replace(rx, "_");
    filename = histDir.absolutePath()+QDir::separator()+filename;

    QFile f(filename);
    if (!f.open(QIODevice::WriteOnly)) {
        qWarning(QString("Can't write histogram '%1' to file '%2'").arg(fn).arg(filename).utf8());
        return false;
    }
    QTextStream ts(&f);
    ts.setEncoding(QTextStream::UnicodeUTF8);
    for (unsigned int i = 0; i < nbins; i++)
        ts << bincenter(i) << "\t" << py[i] << "\n";

    f.close();
    return true;
}

unsigned int th1::get_nbins () const
{
    return (nbins);
}

std::vector<double> th1::get_all_bins () const
{
    QMutexLocker lock(explorerManagerMutex);
    return py;
}

double th1::get_bin (unsigned int index) const
{
    QMutexLocker lock(explorerManagerMutex);
    if (index >= nbins)
        return (0);
    return (py.at(index));
}

unsigned int th1::get_nent () const
{
    QMutexLocker lock(explorerManagerMutex);
    return (nent);
}

unsigned int th1::get_over () const
{
    QMutexLocker lock(explorerManagerMutex);
    return (over);
}

unsigned int th1::get_under () const
{
    QMutexLocker lock(explorerManagerMutex);
    return (under);
}

double th1::get_sum () const
{
    QMutexLocker lock(explorerManagerMutex);
    double s=0;
    for(unsigned int i=0;i<py.size();i++)
        s += py[i];
    return s;
}

int th1::get_modseq () const
{
    QMutexLocker lock(explorerManagerMutex);
    return (mod);
}

double th1::get_xmin () const
{
    return (xmin);
}

double th1::get_xmax () const
{
    return (xmax);
}

void th1::update_autorange() const
{
    if (!autorange_enabled()) {
        ar_min = xmin;
        ar_max = xmax;
        return;
    }

    QMutexLocker lock(explorerManagerMutex);
    double maxvalue = 0;
    for (unsigned int i = 0; i < nbins; i++) {
        if (py[i] > maxvalue) maxvalue = py[i];
    }

    double threshold = maxvalue * 0.025;
    if (threshold < 2) threshold = 2;

    double pxmin = bincenter(0), pxmax = bincenter(nbins-1);
    int i;
    for (i = 1; i < (int)nbins; i++)
        if (py[i] > threshold) {
            pxmin = bincenter(i-1);
            break;
        }
    for (int j = (int)nbins - 3; j >= i; --j)
        if (py[j] > threshold) {
            pxmax = bincenter(j+2);
            break;
        }
    if (pxmin<pxmax) {
        ar_min = pxmin;
        ar_max = pxmax;
    }
}
