/*
 *  adchardw - ADC low-level interface library
 *	(C) 2003-2013 Ilja Slepnev <Ilja.Slepnev@jinr.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#
#include <unistd.h>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <cstdlib>
#include <cmath>

#include <iostream>
#include <sstream>
#include <exception>
#include <stdexcept>

#include <QDebug>
#include <QDateTime>

#include "mlink/mlink.h"
#include "mlink/frame_hw.h"           /* M-Link hardware interface */
#include "mlink/mlink_frame_header.h"
#include "util/gettimeofday.h"

#include "adcm16.h"
#include "adcm_df.h"

#include "adchardw.h"
#include "adcm_mem.h"
#include "bitop.h"
#include "adcm_raw_types.h"

#include "adcmdevice.h"
#include "adcm_ids.h"
#include "config/allconfig.h"

//unsigned long DEFAULT_DMA_BUFFER_SIZE = 16384 * 4096; // 64M

const int DEFAULT_ADCM_MLINK_ADDRESS = 2;

conf_runtime_t::conf_runtime_t(size_t numberOfChannels)
{
    runmode = ADCM_RUNMODE_RUN;
    run_enable = false;
    trigbits = 0;
    timestamp_is_utc = 1;
    ch_ctrl.resize(numberOfChannels);
    testpass.all = 0;
    testfail.all = 0;
}

void conf_runtime_t::clear()
{
    runmode = ADCM_RUNMODE_RUN;
    run_enable = false;
    trigbits = 0;
    for(int i=0; i<ch_ctrl.size(); i++)
        ch_ctrl[i].all = 0;
    testpass.all = 0;
    testfail.all = 0;
}

AdcmDevice::AdcmDevice(mlink::RegIOBase &adcmLink)
    : CustomMlinkDevice(adcmLink, DEFAULT_ADCM_MLINK_ADDRESS),
      conf_runtime(ADCM_N_CH)
{
    ast = ADC_OFF;
    nrun = 0;         // run number
    ts_write_ts = 0;

    unsigned long driverVersion = link.getDriverVersion();
    QVector<quint32> supportedDriverVersions;
    supportedDriverVersions.push_back(0x00010020);
    supportedDriverVersions.push_back(0x00010021);
    if (!supportedDriverVersions.contains(driverVersion)) {
        QString s;
        s = QString("Unsupported driver version %1.%2")
                .arg(driverVersion>>16).arg(driverVersion&0xFFFF);
        throw std::runtime_error(s.toStdString());
    }
}

#define CHECK_CH(ch) Q_ASSERT(((unsigned int)(ch) < ADC_NCH))

int AdcmDevice::nch() const
{
    return conf_mem.nch;
}

const char *AdcmDevice::adc_state_str (int st)
{
    switch (st) {
    case ADC_OFF:
        return ("OFF");
    case ADC_READY:
        return ("READY");
    case ADC_RUN:
        return ("RUN");
    case ADC_ERROR:
        return ("ERROR");
    default:
        return ("unknown");
    }
}

unsigned int AdcmDevice::adc_dma_cycle () const
{
    return (nrun);
}


void AdcmDevice::clearcounts ()
{
    nrun = 0;
}

int AdcmDevice::state_online() const
{
    if ((ast == ADC_READY) || (ast == ADC_RUN))
        return 1;
    qWarning() << QString("state_online: not ready, state is %1").arg(adc_state_str (ast));
    return 0;
}

QString AdcmDevice::getDeviceIdStr() const
{
    QStringList sl;
    Q_FOREACH(conf_mem_board_t b, conf_mem.boards()) {
        sl.push_back(b.descriptionStr() + b.firmwareStr());
    }
    return sl.join(", ");
}

QVector<double> AdcmDevice::readTemperatures()
{
    QVector<double> v;
    // read Temperatures
    for(int i=0; i<conf_mem.boards().size(); i++) {
        quint16 rawTemp = regRead(ADCM16_TEMPERATURE_REG + i * ADCM_BOARD_NREG);
        int signedTemp;
        if (rawTemp & 0x800)
            signedTemp = -((int)rawTemp & 0x7FF);
        else
            signedTemp = rawTemp;
        double realTemp = 0.0625 * rawTemp;
        v.push_back(realTemp);
    }
//    conf_mem.temperature = v;
    return v;
}

void AdcmDevice::setClock()
{
    const int TS_COUNT = 4;
    struct timeval tv0, tv;
    gettimeofday(&tv0, NULL);
    for(int i=0; i<TS_COUNT; i++) {
        gettimeofday(&tv, NULL);
        qint64 ts = (qint64)ADC_TS_FREQ() * tv.tv_sec + (qint64)ADC_TS_FREQ()/1e6 * tv.tv_usec;
        writeClock(ts);
    }
    double dt = (tv.tv_sec-tv0.tv_sec) + 1e-6*(tv.tv_usec-tv0.tv_usec);
    double ts_roundtrip = dt/TS_COUNT;
    qDebug() << QString("Timestamp write delay %1 s").arg(ts_roundtrip);
}

void AdcmDevice::writeClock(quint64 value)
{
    regWrite64(ADCM16_TS_SET_REG64, value);
    ts_write_ts = value;
}

quint64 AdcmDevice::readClock()
{
    const int read_count = 4;
    std::vector<qint64> host_ts_v, adc_ts_v;
    for(int i=0; i<read_count; i++) {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        host_ts_v.push_back((qint64)ADC_TS_FREQ() * tv.tv_sec + (qint64)ADC_TS_FREQ()/1e6 * tv.tv_usec);
        adc_ts_v.push_back(regRead64(ADCM16_TS_READ_REG64));
    }
    qint64 min_delta = adc_ts_v[0] - host_ts_v[0];
    int min_index = 0;
    for(int i=0; i<read_count; i++) {
        qint64 delta = adc_ts_v[i] - host_ts_v[i];
        if (delta < min_delta) {
            min_delta = delta;
            min_index = i;
        }
    }
    qint64 host_ts = host_ts_v[min_index];
    qint64 adc_ts  = adc_ts_v[min_index];
    if (adc_ts == 0 || adc_ts == -1) {
        qWarning() << QString("ADC Timestamp not available");
        return 0;
    }
    double ts_step = 1.0/ADC_TS_FREQ();
//    logger.info(QString("Timestamp period %1, frequency %2").arg(ts_step).arg(ADC_TS_FREQ()));
//    logger.info(QString("Local Timestamp: %1 s").arg(ts_step*(host_ts), 0, 'f', 6));
//    logger.info(QString("ADC   Timestamp: %1 s").arg(ts_step*(adc_ts), 0, 'f', 6));
    double delta_ts = ts_step * (adc_ts - host_ts);
    double ts_runtime = ts_step*(host_ts - ts_write_ts);
//    double delta_freq = delta_ts / ts_runtime;
    if (ts_runtime > 1.0) {
        if (std::fabs(delta_ts) > 10) {
//            qWarning() << QString("Timestamp delta %1 s, period %2 s, delta_freq %3").arg(delta_ts).arg(ts_runtime).arg(delta_freq);
        } else {
//            qDebug() << QString("Timestamp delta %1 s, period %2 s, delta_freq %3").arg(delta_ts).arg(ts_runtime).arg(delta_freq);
        }
    }
    return adc_ts;
}

int decodeSlaveCount(quint16 masterBoardType, quint16 slaveBits)
{
    int slave_count = 0;
    for(int i=0; i<3; i++) {
        if (slaveBits & (1<<i)) {
            slave_count++;
        } else {
            break;
        }
    }
    return std::min(slave_count, maxSlaveCount(masterBoardType));
}

conf_mem_t AdcmDevice::discover()
{
    conf_mem_t conf_mem;
    QVector<adcm_status_reg_t> status_regs;
    // read ID from master board
    try {
        adcm_status_reg_t master_sreg;
        master_sreg.all = regRead(adc_reg_status(0));
        status_regs.push_back(master_sreg);
    }
    catch (std::exception &e) {
        QString s = QObject::tr("Can't read device id register (%1)")
                .arg(e.what());
        throw std::runtime_error(s.toStdString());
    }

    quint16 masterBoardType = status_regs.at(0).bit.id;

    if (!isDeviceTypeSupported(masterBoardType)) {
        QString s = QObject::tr("Unsupported device (0x%1)").arg(masterBoardType, 2, 16, QChar('0'));
        throw std::runtime_error(s.toStdString());
    }

    int expectedSlaveCount = decodeSlaveCount(masterBoardType, status_regs.at(0).bit.serial);

    // Detect slave boards
    for(int i=1; i < 1+expectedSlaveCount; i++) {
        adcm_status_reg_t slave_sreg;
        try {
            slave_sreg.all = regRead(adc_reg_status(i));
        }
        catch (std::exception &e) {
            QString s = QObject::tr("Can't read device id register on slave board %1 (%2)")
                    .arg(i)
                    .arg(e.what());
            qCritical() << s;
            break;
        }
        status_regs.push_back(slave_sreg);
    }
    const int boardCount = status_regs.size();

    // recreate board descriptors
    conf_mem.clear();
    // read properties
    for(int i=0; i<boardCount; i++) {
        const quint16 reg_base = i * ADCM_BOARD_NREG;
        try {
            AdcmHardwareConfig hw;
            hw.fw_ver       = regRead(reg_base + ADCM16_FW_VER_REG);
            hw.fw_rev       = regRead(reg_base + ADCM16_FW_REV_REG);
            hw.serial_id    = regRead(reg_base + ADCM16_SERIAL_ID_REG);
            if (hw.fw_rev == 0 || (hw.fw_rev >= 15645 && hw.fw_rev < 0xFFFF)) {
                hw.serial_id |= ((quint32)regRead(reg_base + ADCM16_SERIAL_ID_HI_REG) << 16);
            }
            hw.prop_reg.all = regRead(reg_base + ADCM16_PROPERTIES_REG);
            hw.ch_dpm_ks    = regRead(reg_base + ADCM16_CH_DPM_KS_REG);
            hw.flen_max     = regRead(reg_base + ADCM16_FRAGMMAX_REG);
            hw.fcnt_max     = regRead(reg_base + ADCM16_MAX_NFRAG_REG);
            if (!hw.validateAny(i)) break;
            if (i >= 1) {
                if (!hw.validateSlave(i, conf_mem.boards().at(0).hw)) break;
            }
            conf_mem_board_t b(status_regs.at(i).bit.id, hw);
            conf_mem.append(b);
        }
        catch (std::exception &e) {
            QString s = QObject::tr("Can't read hardware configuration registers on board %1 (%2)")
                    .arg(i)
                    .arg(e.what());
            qCritical() << s;
            break;
        }

    }
    return conf_mem;
}

void AdcmDevice::adc_reset ()
{
    conf_runtime.clear();
    conf_mem.clear();
    clearcounts ();

    ast = ADC_OFF;
    link.reset (); // Reset may succeed while device is offline...
    setBufferSize();

    conf_mem = discover(); // throws exception on fatal error

    // set defaults
    regWrite(ADCM16_CTRL_REG, ADCM16_CTRL_DEFAULT);
    // TODO: soft reset

//    if (!test_regs())
//        throw std::runtime_error("Register test failed");

    ast = ADC_READY;

//    readTemperatures();

    // clock requires board clock set in conf_mem
    setClock();
    readClock();
}

int AdcmDevice::adc_get_testpass(adcm_testpass_t *fpass, adcm_testpass_t *ffail) const
{
    if (fpass)
        *fpass = conf_runtime.testpass;
    if (ffail)
        *ffail = conf_runtime.testfail;
    return 0;
}

void AdcmDevice::adc_setreg_inrun(int reg, int data)
{
    regWrite(reg, data);
}

void AdcmDevice::adc_write_ctrl_reg ()
{
    adcm_ctrl_reg_t ctrl;

    ctrl.all = ADCM16_CTRL_DEFAULT;
    ctrl.bit.run = conf_runtime.run_enable && (conf_runtime.runmode == ADCM_RUNMODE_RUN);
    ctrl.bit.test = conf_runtime.run_enable && (conf_runtime.runmode == ADCM_RUNMODE_BASECAL);
    ctrl.bit.trigbits = conf_runtime.trigbits & 0xF;
    ctrl.bit.timestamp_is_utc = conf_runtime.timestamp_is_utc;

//    qDebug() << QString("adc_write_ctrl_reg: CTRL=0x%1").arg(ctrl.all, 8, 16);
    regWrite(ADCM16_CTRL_REG, ctrl.all);
}

bool AdcmDevice::test_regs()
{
    // TODO: test slave regs
    if (conf_mem.boards().size() == 0) return false;

    bool testOk = true;
    try {
        const int nloops = 10;
        for(int i=0; i<nloops; i++) {
            QMap<quint16, quint16> m;
            m[ADCM16_WIN_REG] = rand();
            for(int n=0; n<conf_mem.boards().at(0).nch; n++) {
                m[adc_reg_ch_thr(n)] = rand();
            }
            Q_FOREACH(quint16 addr, m.keys())
                regWrite(addr, m[addr]);
            Q_FOREACH(quint16 addr, m.keys()) {
                quint16 rd = regRead(addr);
                if (rd != m[addr]) {
                    testOk = false;
                    qWarning() << QString("Register test failed at address 0x%1: wrote 0x%2, read 0x%3")
                                  .arg(addr, 4, 16, QChar('0'))
                                  .arg(m[addr], 4, 16, QChar('0'))
                                  .arg(rd, 4, 16, QChar('0'));
                    Q_FOREACH(quint16 srch_addr, m.keys()) {
                        if (m[srch_addr] == rd) {
                            qDebug() << QString("%1 was wrote to %2 addr")
                                        .arg(rd, 4, 16, QChar('0'))
                                        .arg(srch_addr, 4, 16, QChar('0'));
                            break;
                        }
                    }
                }
            }
        }
    }
    catch (std::exception &e) {
        qWarning() << QString("Register test failed: %1").arg(e.what());
        testOk = false;
    }
    conf_runtime.testfail.bit.adcregtest_m = !testOk;
    conf_runtime.testpass.bit.adcregtest_m = 1;
    conf_runtime.testfail.bit.adcregtest_s1 = !testOk;
    conf_runtime.testpass.bit.adcregtest_s1 = 1;
    if (testOk)
        qDebug() << QString("Register test Ok");
    return testOk;
}

bool AdcmDevice::isRunning() const
{
    return ast == ADC_READY || ast == ADC_RUN;
}

bool AdcmDevice::adc_start ()
{
    conf_runtime.run_enable = true;

    if (ast != ADC_READY) {
        qWarning() << QString("adc_start: not ready, state is %1").arg(adc_state_str (ast));
        return false;
    }

    try {
        adc_write_ctrl_reg ();
    }
    catch (std::exception &e) {
        ast = ADC_ERROR;
        qWarning() << QString("adc_start failed: %1").arg(e.what());
        throw;
    }
    nrun++;
    ast = ADC_RUN;
//    qDebug() << QString("adc_start ok %1%2")
//                .arg((conf_runtime.runmode & ADCM_RUNMODE_RUN) ? "RUN " : "")
//                .arg((conf_runtime.runmode & ADCM_RUNMODE_RUN) ? "TEST " : "");
    return true;
}

bool AdcmDevice::adc_stop ()
{
    conf_runtime.run_enable = false;
    adcm_ctrl_reg_t ctrl;

    ctrl.all = ADCM16_CTRL_DEFAULT;
    try {
        regWrite(ADCM16_CTRL_REG, ctrl.all);
        usleep (1000);                // let DMA complete transfer
    }
    catch (std::exception &e) {
        ast = ADC_ERROR;
        qWarning() << QString("adc_stop failed: %1").arg(e.what());
        throw;
    }

    ast = ADC_READY;
    return true;
}

void AdcmDevice::adc_setthr (int ch, int thr)
{
    CHECK_CH(ch);

    if (!state_online()) return;

    if (conf_runtime.runmode == ADCM_RUNMODE_BASECAL) {
        thr = 0xFFFF;
    } else {
        thr += conf_mem.getAdcHalfRange(ch);
    }
    //  thr &= 0x3FF;
    int reg = adc_reg_ch_thr (ch);
    adc_setreg_inrun (reg, thr);
//    qDebug() << QString("thr[%1]: %2").arg(ch, 2).arg(thr);
}

// set hysteresis (low threshold)
void AdcmDevice::adc_sethyst (int ch, int hyst)
{
    if (!conf_mem.hasHyst(ch)) return;
    CHECK_CH(ch);

    if (!state_online()) return;

    if (conf_runtime.runmode == ADCM_RUNMODE_BASECAL) {
        hyst = 0xFFFF;
    } else {
        hyst += conf_mem.getAdcHalfRange(ch);
    }
    //  hyst &= 0x3FF;
    int reg = adc_reg_ch_hyst (ch);
    adc_setreg_inrun (reg, hyst);
    qDebug() << QString("hyst[%1]: %2").arg(ch, 2).arg(hyst);
}

void AdcmDevice::adc_settrig (int ch, int trig)
{
    CHECK_CH(ch);

    if (conf_runtime.trigbits & ADCM_TRIGBIT_EXT) {
        // external trigger mode, master is special channel N+1
        trig &= ~ADCM_CCTRL_BIT_M;
        trig |= ADCM_CCTRL_BIT_S;
    }

//    // fixup trig bits - BUG in hardware
//    if (trig & ADCM_CCTRL_BIT_TEST) {
//        trig |= ADCM_CCTRL_BIT_M;
//        trig |= ADCM_CCTRL_BIT_S;
//    }

    // disable test pulse in normal run mode
    if (conf_runtime.runmode == ADCM_RUNMODE_RUN) {
        if (! (trig & ADCM_CCTRL_BIT_CMPEN)) {
            trig &= ~(ADCM_CCTRL_BIT_M|ADCM_CCTRL_BIT_S);
        }
        trig &= ~ADCM_CCTRL_BIT_TEST;
    } else  if (conf_runtime.runmode == ADCM_RUNMODE_BASECAL) {
        trig &= ~ADCM_CCTRL_BIT_CMPEN;
//        trig |= ADCM_CCTRL_BIT_TEST;
        trig |= (ADCM_CCTRL_BIT_M|ADCM_CCTRL_BIT_S);
    }

    Q_ASSERT(ch < conf_runtime.ch_ctrl.size());
//    set_bit_value(trig, ADCM_CCTRL_BIT_CMPEN|ADCM_CCTRL_BIT_TEST, trig&ADCM_CCTRL_BIT_CMPEN);
    conf_runtime.ch_ctrl[ch].bit.trig = (trig & 0x000F);
    conf_runtime.ch_ctrl[ch].bit.invert = (trig & ADCM_CCTRL_BIT_INV)?1:0;
    conf_runtime.ch_ctrl[ch].bit.zero = (trig & ADCM_CCTRL_BIT_ZERO)?0:1; // inverted
    int reg = adc_reg_ch_ctrl (ch);
    int reg_s = adc_reg_ch_ctrl_s (ch);

    adc_setreg_inrun (reg, conf_runtime.ch_ctrl[ch].all);
    if (reg != reg_s) {
        adc_setreg_inrun (reg_s, conf_runtime.ch_ctrl[ch].all);
    }
//    if (conf_runtime.ch_ctrl[ch].all)
//    qDebug() << QString("adc_settrig: trig[%1]: 0x%2 %3%4%5%6%7%8").arg(ch, 2)
//                .arg(conf_runtime.ch_ctrl[ch].bit.trig, 4, 16, QChar('0'))
//                .arg((conf_runtime.ch_ctrl[ch].bit.trig&ADCM_CCTRL_BIT_CMPEN)?" EN":"")
//                .arg((conf_runtime.ch_ctrl[ch].bit.trig&ADCM_CCTRL_BIT_M)?" M":"")
//                .arg((conf_runtime.ch_ctrl[ch].bit.trig&ADCM_CCTRL_BIT_S)?" S":"")
//                .arg((conf_runtime.ch_ctrl[ch].bit.trig&ADCM_CCTRL_BIT_TEST)?" TEST":"")
//                .arg(conf_runtime.ch_ctrl[ch].bit.zero?" ZERO":"")
//                .arg(conf_runtime.ch_ctrl[ch].bit.invert?" INV":"")
//                ;
}

void AdcmDevice::set_timestamp_is_utc(bool value)
{
    conf_runtime.timestamp_is_utc = value;
}

//void AdcmDevice::adc_setrunmode (adcm_runmode_t runmode)
//{
//    conf_runtime.runmode = runmode;
//    if (running) {
//        adc_write_ctrl_reg ();
//    }
//}

void AdcmDevice::adc_settrigbits (unsigned int trigbits)
{

    conf_runtime.trigbits = trigbits;
//    if (running) {
        adc_write_ctrl_reg ();
//    }
}

float AdcmDevice::adc_gain_val_to_db(int value)
{
    return value * 3 - 9;
}

int AdcmDevice::adc_gain_val_to_raw(float db_value)
{
    return (int)((db_value+9) / 3);
}


void AdcmDevice::adc_setgain (int ch, float fgain)
{
    CHECK_CH(ch);
    if (!state_online()) return;

    int gain = adc_gain_val_to_raw(fgain);
    // limit gain range
    if (gain < 0) gain = 0;
    if (gain > 0x00FF) gain = 0x00FF;
    conf_runtime.ch_ctrl[ch].bit.gain = gain;
    int reg = adc_reg_ch_ctrl (ch);

    adc_setreg_inrun (reg, conf_runtime.ch_ctrl[ch].all);
//    qDebug() << QString("gain[%1]: 0x%2").arg(ch, 2)
//                .arg(conf_runtime.ch_ctrl[ch].bit.gain, 4, 16);
}

void AdcmDevice::adc_setwindow (int searchback, int searchwin, int latency)
{
    if (!state_online()) return;

    if (conf_mem.hasExtMatchWin()) {
        if (searchback < 0) {
            //	  latency += (-searchback);
            searchback = (-searchback) | 0x8000;
        }
        adc_setreg_inrun (ADCM16_EXT_LAT_REG, latency);
        adc_setreg_inrun (ADCM16_EXT_MATCHWIN_REG, searchwin);
        adc_setreg_inrun (ADCM16_EXT_PREWIN_REG, searchback);
        //	logm(L_EXP, LOG_INFO, "set adc registers: [%02X] = %04X", ADCM16_EXT_LAT_REG, latency);
        //	logm(L_EXP, LOG_INFO, "set adc registers: [%02X] = %04X", ADCM16_EXT_PREWIN_REG, searchback);
        //	logm(L_EXP, LOG_INFO, "set adc registers: [%02X] = %04X", ADCM16_EXT_MATCHWIN_REG, searchwin);
    } else {
        int reg = ADCM16_WIN_REG;
        // bits 0-4: search back
        // 5-10: search width
        // 11-15: data latency
        latency += 1;
        if (latency > 0x1F) latency = 0x1F;
        searchback &= 0x1F;
        searchwin &= 0x3F;
        latency &= 0x1F;

        adcm_win_reg_t data;
        //  data = searchback | (searchwin << 5) | (latency << 11);
        data.bit.searchback = searchback;
        data.bit.searchwin = searchwin;
        data.bit.latency = latency;

        adc_setreg_inrun (reg, data.all);
    }
}

void AdcmDevice::adc_set_trig_mult (unsigned int mult_min, unsigned int mult_max)
{
    if (!conf_mem.hasTrigMultiplicity()) return;
    if (mult_min > mult_max) {
        unsigned int tmp;
        tmp = mult_min;
        mult_min = mult_max;
        mult_max = tmp;
    }
    int d = ((mult_max & 0xFF)<<8) | (mult_min & 0xFF);
    adc_setreg_inrun (ADCM16_TRIGMULT_REG, d);
//    qDebug() << QString("mult_min=%1, mult_max=%2").arg(mult_min).arg(mult_max);
}

void AdcmDevice::adc_set_fragm_len (unsigned int fragm_len)
{
    if (!conf_mem.hasMultiFragment()) return;

    if (fragm_len < ADCM_MIN_FRAGM_LEN) fragm_len = ADCM_MIN_FRAGM_LEN;
    fragm_len = (int)(fragm_len/2)*2;
    int d = (int)fragm_len-2;
    for (int i=0; i < conf_mem.boards().size(); i++) {
        adc_setreg_inrun (i * ADCM_BOARD_NREG + ADCM16_FRAGMLEN_REG, d);
        // if (i == 1) break; // ADC64 BUG?
    }
}

void AdcmDevice::adc_set_fragm_count (int ch, unsigned int fragm_count)
{
    if (!conf_mem.hasMultiFragment()) return;
    CHECK_CH(ch);
    {
        int reg = adc_reg_ch_fragm(ch);
        int d = fragm_count-1;
//        qDebug() << QString("adc_set_fragm_count(%1, %2)").arg(ch).arg(fragm_count);
        adc_setreg_inrun (reg, d);
    }
}
int AdcmDevice::adc_get_fragm_len_max (unsigned int *fragm_len_max) const
{
    if (!conf_mem.hasMultiFragment()) {*fragm_len_max = 128; return 0;}
    if (!fragm_len_max) return -EINVAL;
//    if (running) return -EBUSY;
    *fragm_len_max = conf_mem.boards().at(0).hw.flen_max;
    return 0;
}
int AdcmDevice::adc_get_fragm_count_max (unsigned int *fragm_count_max) const
{
    if (!conf_mem.hasMultiFragment()) {*fragm_count_max = 1; return 0; }
    if (!fragm_count_max) return -EINVAL;
//    if (running) return -EBUSY;
    *fragm_count_max = conf_mem.boards().at(0).hw.fcnt_max;
    return 0;
}

void AdcmDevice::setBufferSize()
{
    // 100 MB/s, 200 bytes/packet
    int averagePacketSize = 200;
    int bytesThroughput = 100*1024*1024;
    double maxProgramLatencyMs = 100; // 10 Hz
    int averageBufferOccupancy = (double)bytesThroughput *  maxProgramLatencyMs / 1000;
    int safetyFactor = 3;
    int bufSizeLimit = averageBufferOccupancy * safetyFactor;
    int bufPacketsLimit = bufSizeLimit / averagePacketSize;
    qDebug() << "Receive buffer demultiplexer extimation:" << bytesThroughput/1024/1024 << "MB/s, max program latency" << maxProgramLatencyMs << "ms";
    qDebug() << "Receive buffer size:"
             << bufPacketsLimit
             << "packets of size" << averagePacketSize
             << "total size" << bufSizeLimit/1024/1024 << "MB, safety factor" << safetyFactor;
    link.setBufSizeLimit(bufPacketsLimit);
}

void AdcmDevice::poll ()
{
    if (!isRunning())
        return;
    try {
        link.receive();
    }
    catch (std::exception &e) {
        qWarning() << QString("Exception in receive(): %1").arg(e.what());
    }
}

mlink::RawPacketList AdcmDevice::popDataPackets()
{
    return link.popDataPackets();
}

//qint64 AdcmDevice::read(char * data, qint64 maxBytes)
//{
//    return link.read(data, maxBytes);
//}

quint64 convRegData16To64(const QVector<quint16> &v)
{
    Q_ASSERT(v.size() == 4);
    if (v.size() != 4) return 0;
    quint64 b0 = v[0];
    quint64 b1 = v[1];
    quint64 b2 = v[2];
    quint64 b3 = v[3];
    return (b3<<48) | (b2<<32) | (b1<<16) | b0;
}

namespace {
double timestamp_sec()
{
    QDateTime tv = QDateTime::currentDateTime();
    return tv.toTime_t() + 1e-3*tv.time().msec();
}
}

AdcHwCounters AdcmDevice::adc_get_counters ()
{
    readClock();

    AdcHwCounters adcHwCounters;

    unsigned int nch = conf_mem.nch;
    Q_ASSERT(nch < ADC_NCH);
    std::vector<int> registers;

    const int ts_regs_offset = registers.size();
    for (int i=0; i<4; i++) registers.push_back(ADCM16_TS_READ_REG64+i);
    for (int i=0; i<4; i++) registers.push_back(ADCM16_TS_A_REG64+i);
    for (int i=0; i<4; i++) registers.push_back(ADCM16_TS_B_REG64+i);

    const int cnt_regs_offset = registers.size();
    for (unsigned int ch = 0; ch < std::min(ADCM_BOARD_NCH, nch); ch++) {
        int regNumber;
        if (conf_mem.boards()[0].hw.fw_ver == 0)
            regNumber = adc_reg_ch_cnt_compat (ch);
        else
            regNumber = adc_reg_ch_cnt (ch);
        registers.push_back(regNumber);
        registers.push_back(regNumber+1);
    }
    // read master board registers atomically by single request
    try {
        QVector<quint16> result = QVector<quint16>::fromStdVector(regReadBlk(registers));
        Q_ASSERT(result.size() == (int)registers.size());

        // BUG: slave board can't handle multiple requests in single frame
        for (unsigned int i=ADCM_BOARD_NCH; i<nch; i++) {
            int regNumber = adc_reg_ch_cnt(i);
            result.push_back(regRead(regNumber));
            result.push_back(regRead(regNumber+1));
        }

        for (unsigned int i = 0; i < nch; i++) {
            int dhi = result[cnt_regs_offset + 2*i];
            int dlo = result[cnt_regs_offset + 2*i+1];
            int d = (dhi << 16) | dlo;
            adcHwCounters.hits.push_back(d);
        }

        // read adcm_time registers
        // counters are reset by run_start
        // can't use ts_start_stop while run in progress
        quint64 ts_current = convRegData16To64(result.mid(ts_regs_offset, 4));
        quint64 ts_run_start = convRegData16To64(result.mid(ts_regs_offset+4, 4));
    //    quint64 ts_run_stop = convRegData16To64(result.mid(ts_regs_offset+8, 4));
        if (((conf_mem.boards().size() > 0) && conf_mem.boards().at(0).hw.fw_ver >= 0x0100) && (conf_mem.boards().at(0)).hw.fw_rev >= 10820) {
    //        adcm_counters.time = (double)(ts_run_stop - ts_run_start) / ADC_TS_FREQ();
            adcHwCounters.startrun_ts = (double)ts_run_start / ADC_TS_FREQ();
            adcHwCounters.current_ts  = (double)ts_current / ADC_TS_FREQ();
        } else {
            adcHwCounters.startrun_ts = 0; // FIXME
            adcHwCounters.current_ts = timestamp_sec(); // use host time
        }
    //    qDebug() << "adcm_counters.time =" << adcm_counters.time;
        return adcHwCounters;
    }
    catch (std::exception &e) {
        qWarning() << QString("Can't read hit counters: %1").arg(e.what());
        return AdcHwCounters();
    }
}

void AdcmDevice::setup(const AllConfig &cfg)
{
//    qDebug() << "AdcmDevice::setup";
    if (!state_online()) {
//        qDebug() << "AdcmDevice::setup - OFFLINE";
        return;
    }
    conf_runtime.runmode = cfg.adc.runmode; // set this first

    bool saveRunEnableState = conf_runtime.run_enable;
//    conf_runtime.run_enable = false;
//    adc_write_ctrl_reg ();

    int prewin = cfg.adc.prewin;
    int matchwin = cfg.adc.matchwin;
    if (cfg.adc.is_simple_trigger()) {
        prewin = 0;
        matchwin = 0;
    }
    adc_setwindow (prewin, matchwin, cfg.adc.latency);
    adc_settrigbits (cfg.adc.trigbits); // Either internal or external trigger

    int mult_min_now = cfg.adc.mult_min;
    int mult_max_now = cfg.adc.mult_max;
    if (cfg.adc.is_testmode()) {
        mult_min_now = 1;
        mult_max_now = nch();
    }
    if(mult_max_now == 0) mult_max_now = nch();
    adc_set_trig_mult (mult_min_now, mult_max_now);

    set_timestamp_is_utc(cfg.adc.timestamp_is_utc);

    // apply trigger bits
    for (int c = 0; c < std::min(nch(), cfg.adc.size()); c++) {
        adc_settrig (c, cfg.adc.trig[c]);
    }

    // fill with defaults
    QVector<quint16> fragmentLength(nch(), cfg.adc.fragm_len_max);
    QVector<quint16> fragmentCount(nch(), cfg.adc.fragm_count_max);
    // config may have different nunber of channels than actual hardware
    for (int c = 0; c < std::min(nch(), cfg.adc.size()); c++) {
        unsigned int flen=0, fcnt=0;
        unsigned int recwin = cfg.adc.recwin[c];
        const unsigned int max_recwin = cfg.adc.adc_cfg_get_recwin_max();
        if (cfg.adc.is_testmode()) {
            if (recwin > max_recwin)
                recwin = max_recwin;
        }
        decomp_recwin(recwin, &flen, &fcnt, cfg.adc.fragm_len_max, cfg.adc.fragm_count_max);
        fragmentLength[c] = flen;
        fragmentCount[c] = fcnt;
        // FIXME: implement per-channel recwin setting
    }
    for (int c = 0; c < nch(); c++) {
        adc_set_fragm_len (fragmentLength[c]); // hardware limitation: no per channel fragm_len setting
        adc_set_fragm_count (c, fragmentCount[c]);
    }

    if(conf_mem.hasGainControl())
        for (int c = 0; c < std::min(nch(), cfg.adc.size()); c++)
            adc_setgain (c, cfg.adc.gain[c]);

    for (int c = 0; c < std::min(nch(), cfg.adc.size()); c++) {
        // set threshold
        const int thr = cfg.thr_mv_to_value (c, cfg.adc.thr[c]);
        adc_setthr (c, thr);

        // set hysteresis
        const double hyst_factor = 0.50;
        const unsigned int hyst = cfg.thr_mv_to_value (c, cfg.adc.thr[c] * hyst_factor);
        adc_sethyst (c, hyst);
    }
    conf_runtime.run_enable = saveRunEnableState;
    adc_write_ctrl_reg ();
}

double AdcmDevice::ADC_TS_FREQ() const
{
    double step_ns = conf_mem.getMasterAdcStepNs();
    if (step_ns > 0)
        return 1.0/(step_ns*1e-9);
    else
        return 1.0;
}

bool AdcmDevice::decomp_recwin(unsigned int recwin, unsigned int *flen, unsigned int *fcnt, unsigned int flen_max, unsigned int fcnt_max)
{
    //  qDebug() << QString("decomp_recwin: recwin=%1 flen_max=%2 fcnt_max=%3").arg(recwin).arg(flen_max).arg(fcnt_max);
    if (recwin <= flen_max) {
        *flen = recwin;
        *fcnt = 1;
        //        qDebug() << (1, QString("decomp_recwin: (1) flen=%1 fcnt=%2").arg(*flen).arg(*fcnt));
        return true;
    }

    int d = recwin;
    unsigned int dl=flen_max, dc=fcnt_max;
    for (unsigned int c=2; c<=fcnt_max; c++)
        for (unsigned int l=2; l<=flen_max; l+=2) {
            unsigned int n = c*l;
            if (n==recwin) {
                *flen = l;
                *fcnt = c;
                //        qDebug() << (1, QString("decomp_recwin: (2) flen=%1 fcnt=%2").arg(*flen).arg(*fcnt));
                return true;
            }
            int dd = std::abs((int)recwin - (int)n);
            if (dd<d) {
                d = dd;
                dc = c;
                dl = l;
            }
        }
    *flen = dl;
    *fcnt = dc;
    //  qDebug() << (1, QString("decomp_recwin: (3) flen=%1 fcnt=%2, actual recwin=%3").arg(*flen).arg(*fcnt).arg((*flen) * (*fcnt)));
    return false;
}
