
#include "adcm_mem.h"

#include <QDebug>

#include "adchardw.h"
#include "adcm_ids.h"

AdcmHardwareConfig::AdcmHardwareConfig() :
    fw_ver(0),
    fw_rev(0),
    serial_id(0),
    ch_dpm_ks(0),
    flen_max(0),
    fcnt_max(0)
{
    prop_reg.all = 0;
}

bool AdcmHardwareConfig::operator==(const AdcmHardwareConfig &rhs) const
{
    return true
            && fw_ver == rhs.fw_ver
            && fw_rev == rhs.fw_rev
            && serial_id == rhs.serial_id
            && ch_dpm_ks == rhs.ch_dpm_ks
            && flen_max == rhs.flen_max
            && fcnt_max == rhs.fcnt_max
            && prop_reg.all == rhs.prop_reg.all
            ;
}

bool AdcmHardwareConfig::validateAny(int i)
{
    if (ch_dpm_ks==0 || ch_dpm_ks == 0xFFFFu) {
        qWarning() << QString("Board %1: bad ch_dpm_ks=%2, set to 1").arg(i).arg(ch_dpm_ks);
        ch_dpm_ks = 1;
    }
    if (flen_max < 8 || flen_max == 0xFFFFu) {
        qWarning() << QString("Board %1: bad flen_max=%2, set to 120").arg(i).arg(flen_max);
        flen_max = 120;
    } else {
        flen_max -= 8; // FIXME: 8 samples - data block header/trailer overhead
    }
    if (fcnt_max == 0 || fcnt_max == 0xFFFFu) {
        qWarning() << QString("Board %1: bad fcnt_max=%2, set to 1").arg(i).arg(fcnt_max);
        fcnt_max = 1;
    } else {
        fcnt_max += 1;
    }
    if (prop_reg.bit.fixed != 0x1) {
        qWarning() << QString("Board %1: bad prop_reg=%2, set to 0").arg(i).arg(prop_reg.all);
        prop_reg.all = 0;
    }
    return true;
}

bool AdcmHardwareConfig::validateSlave(int i, const AdcmHardwareConfig &master)
{
    bool result = true;
    if (master.prop_reg.bit.ext_rwind && !prop_reg.bit.ext_rwind) {
        qCritical() << QString("Board %1: inconsistent ext_rwind %2, master has %3")
                       .arg(i)
                       .arg(prop_reg.bit.ext_rwind)
                       .arg(master.prop_reg.bit.ext_rwind);
        result = false;
    }
    if (master.prop_reg.bit.trig_mult && !prop_reg.bit.trig_mult) {
        qCritical() << QString("Board %1: inconsistent trig_mult %2, master has %3")
                       .arg(i)
                       .arg(prop_reg.bit.trig_mult)
                       .arg(master.prop_reg.bit.trig_mult);
        result = false;
    }
    // TODO: allow different master/slave memory size combinations
    if (master.ch_dpm_ks > ch_dpm_ks) {
        qCritical() << QString("Board %1: insufficinet ch_dpm_ks %2, master has %3")
                       .arg(i)
                       .arg(ch_dpm_ks)
                       .arg(master.ch_dpm_ks);
        result = false;
    }
    if (master.flen_max > flen_max) {
        qCritical() << QString("Board %1: insufficinet flen_max %2, master has %3")
                       .arg(i)
                       .arg(flen_max)
                       .arg(master.flen_max);
        result = false;
    }
    if (master.fcnt_max > fcnt_max) {
        qCritical() << QString("Board %1: insufficinet fcnt_max %2, master has %3")
                       .arg(i)
                       .arg(fcnt_max)
                       .arg(master.fcnt_max);
        result = false;
    }

    return result;
}

conf_mem_board_t::conf_mem_board_t() :
    board_type(0),
    nch(0),
    adcbits(0),
    adcstep_ns(0)
{
}

conf_mem_board_t::conf_mem_board_t(quint16 typeId, const AdcmHardwareConfig &adcmHardwareConfig) :
    board_type(typeId),
    nch(boardChannelCount(typeId)),
    adcbits(boardAdcBits(typeId)),
    adcstep_ns(boardAdcStepNs(typeId)),
    hw(adcmHardwareConfig)
{
    qDebug() << descriptionStr();
}

bool conf_mem_board_t::operator==(const conf_mem_board_t &rhs) const
{
    return true
            && board_type == rhs.board_type
            && nch == rhs.nch
            && adcbits == rhs.adcbits
            && adcstep_ns == rhs.adcstep_ns
            && hw == rhs.hw
            ;
}

QString conf_mem_board_t::descriptionStr() const
{
    return QString("%1 %2")
            .arg(adc_board_name_str(board_type))
            .arg(firmwareStr());
}

QString conf_mem_board_t::firmwareStr() const
{
    QString fw_tag;
    if (hw.fw_ver != 0 && hw.fw_ver != 0xFFFF) fw_tag = QString("ver %1.%2").arg(hw.fw_ver>>8).arg(hw.fw_ver&0xFF);
    if (hw.fw_rev != 0 && hw.fw_rev != 0xFFFF) fw_tag += QString(".%1").arg(hw.fw_rev);
    if (hw.serial_id != 0 && hw.serial_id != 0xFFFF) fw_tag +=
            QString(", S/N %1-%2")
            .arg(hw.serial_id >> 16, 4, 16, QChar('0')).arg(hw.serial_id & 0xFFFF, 4, 16, QChar('0')).toUpper();
    if (!fw_tag.isEmpty()) {
        fw_tag = QString("%1").arg(fw_tag);
    }
    return fw_tag;
}

//void conf_mem_board_t::setFwProp(quint16 value) {
//    adcm_prop_reg.all = value;
//    qDebug() << fwPropStr();
//}

QString conf_mem_board_t::fwPropStr() const
{
    if (hw.prop_reg.bit.fixed == 1) {
        return QString("ADCM firmware properties: 0x%1, %2%3%4%5%6%7").arg(hw.prop_reg.all, 4, 16)
                .arg(hw.prop_reg.bit.invert?"inv ":"")
                .arg(hw.prop_reg.bit.hyst?"hyst ":"")
                .arg(hw.prop_reg.bit.trig_mult?"tr_mult ":"")
                .arg(hw.prop_reg.bit.trig_mult?"var_sample ":"")
                .arg(hw.prop_reg.bit.ext_rwind?"ext_rwind ":"")
                .arg(hw.prop_reg.bit.framer_logic?"framer_logic ":"")
                ;
    }
    return QString();
}

int conf_mem_board_t::getAdcHalfRange() const
{
    return 1<<(boardAdcBits(board_type)-1);
}

int conf_mem_board_t::getAdcMaxValue() const
{
    return (1<<(boardAdcBits(board_type)-1))-1;
}

int conf_mem_board_t::getAdcMidValue() const
{
    return 0;
}

int conf_mem_board_t::getAdcMinValue() const
{
    return -(1<<(boardAdcBits(board_type)-1));
}

conf_mem_t::conf_mem_t(const QVector<conf_mem_board_t> &boards) :
    b(boards),
    nch(getNch())
{
}

conf_mem_t::conf_mem_t()
    :nch(0)
{
}

conf_mem_t::conf_mem_t(const conf_mem_t &ref)
{
    *this = ref;
}

conf_mem_t &conf_mem_t::operator=(const conf_mem_t &rhs)
{
    if (this == &rhs) return *this;
    b = rhs.b;
    nch = rhs.nch;
    return *this;
}

bool conf_mem_t::operator==(const conf_mem_t &rhs) const
{
    return b == rhs.b;
}

void conf_mem_t::clear()
{
    *this = conf_mem_t();
}

quint32 conf_mem_t::master_board_id() const
{
    if (b.isEmpty()) return 0;
    else return b[0].board_type;
}

adcm_prop_reg_t conf_mem_t::master_firmware_prop() const
{
    return b.at(0).hw.prop_reg;
}

bool conf_mem_t::map_ch(int ch, int *bn, int *bc) const
{
    int c=0;
    for (int i=0; i<b.size(); i++) {
        int i_nch = b[i].nch;
        if (ch >= c && ch < c + i_nch) {
            *bn = i;
            *bc = ch-c;
            return true;
        }
        c += i_nch;
    }
    return false;
}

int conf_mem_t::getNch() const
{
    int c=0;
    for (int i=0; i<b.size(); i++) {
        c += b[i].nch;
    }
    return c;
}

bool conf_mem_t::hasHyst(int ch) const
{
    if (b.size() == 0) return false;
    int bn, bc;
    if (!map_ch(ch, &bn, &bc)) return false;
    return b[bn].hw.prop_reg.bit.hyst;
}

int conf_mem_t::getAdcHalfRange(int ch) const
{
    if (b.size() == 0) return 0;
    int bn, bc;
    if (!map_ch(ch, &bn, &bc)) return 0;
    return b[bn].getAdcHalfRange();
}

int conf_mem_t::getAdcMaxValue(int ch) const
{
    if (b.size() == 0) return 0;
    int bn, bc;
    if (!map_ch(ch, &bn, &bc)) return 0;
    return b[bn].getAdcMaxValue();
}

int conf_mem_t::getAdcMidValue(int ch) const
{
    if (b.size() == 0) return 0;
    int bn, bc;
    if (!map_ch(ch, &bn, &bc)) return 0;
    return b[bn].getAdcMidValue();
}

int conf_mem_t::getAdcMinValue(int ch) const
{
    if (b.size() == 0) return 0;
    int bn, bc;
    if (!map_ch(ch, &bn, &bc)) return 0;
    return b[bn].getAdcMinValue();
}

bool conf_mem_t::hasExtMatchWin() const
{
    if (b.size() == 0) return false;
    return b[0].hw.prop_reg.bit.ext_rwind;
}

bool conf_mem_t::hasMultiFragment() const
{
    if (b.size() == 0) return false;
    return b[0].hw.prop_reg.bit.sample_len;
}

bool conf_mem_t::hasTrigMultiplicity() const
{
    if (b.size() == 0) return false;
    return b[0].hw.prop_reg.bit.trig_mult;
}

bool conf_mem_t::hasGainControl() const
{
    if (b.size() == 0) return false;
    return master_board_id() == ADCMG_BOARD_ID;
}


bool conf_mem_t::hasTriggerControl() const
{
    if (b.size() == 0) return false;
    if (master_board_id() == ADCMG_BOARD_ID) return true;
    if (master_board_id() == ADCMLTC_BOARD_ID) return true;
    if (master_board_id() == ADC64_BOARD_ID) return true;
    if (master_board_id() == ADC64S2_BOARD_ID) return true;
    return false;
}

int conf_mem_t::adc_get_prewin_min() const
{
    if (hasExtMatchWin()) return -63;
    //  if (firmware_prop.bit.ext_rwind) return -31;
    else return 0;
}
int conf_mem_t::adc_get_prewin_max() const
{
    if (hasExtMatchWin()) return 63;
    return 31;
}
int conf_mem_t::adc_get_matchwin_min() const
{
    return 0;
}
int conf_mem_t::adc_get_matchwin_max() const
{
    if (hasExtMatchWin()) return 1023;
    return 63;
}
int conf_mem_t::adc_get_latency_min() const
{
    return 1;
}
int conf_mem_t::adc_get_latency_max() const //-V524
{
    if (hasExtMatchWin()) return 63;
    return 31;
}

double conf_mem_t::getMinGainDb() const
{
    if (master_board_id() == ADCMG_BOARD_ID) return -9.0;
    return 0.0;
}

double conf_mem_t::getMaxGainDb() const
{
    if (master_board_id() == ADCMG_BOARD_ID) return 36.0;
    return 0.0;
}

double conf_mem_t::getGainStepDb() const
{
    if (master_board_id() == ADCMG_BOARD_ID) return 3.0;
    return 1.0;
}

QStringList conf_mem_t::getDeviceIdStrList() const
{
    QStringList sl;
    for (int i=0; i<b.size(); i++)
        sl.push_back(b[i].descriptionStr());
    return sl;
}

int conf_mem_t::getMasterAdcBits() const
{
    if(b.isEmpty()) return 0;
    return b.at(0).adcbits;
}

double conf_mem_t::getMasterAdcStepNs() const
{
    if(b.isEmpty()) return 0;
    return b.at(0).adcstep_ns;
}

int conf_mem_t::getMasterMaxFlen() const
{
    if(b.isEmpty()) return 0;
    return b.at(0).hw.flen_max;
}

int conf_mem_t::getMasterMaxFcnt() const
{
    if(b.isEmpty()) return 0;
    return b.at(0).hw.fcnt_max;
}

int conf_mem_t::getMasterChannelMemorySizeKs() const
{
    if(b.isEmpty()) return 0;
    return b.at(0).hw.ch_dpm_ks;
}
