#include <cstdio>
#include <cstdarg>
#include <cstring>
#include <cerrno>
#include <cassert>

#include "mlink/mlink.h"
#include "mlink/frame_hw.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_types.h"
#include "regio/frame_util.h"
#include "colorlog/log.h"

#include "adcm16.h"
#include "adcm_el.h"
#include "adcm_ev.h"
#include "AdcmChbuf.h" // access to adcm_chbuf_t structure
#include "adcm_raw.h"
#include "adcm_raw_types.h"
#include "adcm_dws.h"

#include "adcm_mem.h"
#include "DecodeManager.h"

#include <zlib.h>

//using colorlog;

const bool DO_CHECK_FRAME_CRC = true;

const unsigned int ADCM_MIN_SPW = 2;
const unsigned int ADCM_MAX_SPW = 2;
const unsigned int ADCM_MIN_BIT = 8;
const unsigned int ADCM_MAX_BIT = 14;
const unsigned int ADCM_MIN_NS = ADCM_MIN_N_SAMP;
const unsigned int ADCM_MAX_NS = ADCM_MAX_N_SAMP+1; // LSB is special

//int PRINT_N_FRAME_ERR = 20;
unsigned int PRINT_N_DATA_ERR = 20;

framelist_t::framelist_t()
    : count(0)
{
    for (unsigned int i=0; i<FRAMELIST_CAP; i++) {
        start[i] = 0;
        end[i] = 0;
    }
}

quint32 calculate_crc32(const void *buf, size_t size)
{
    uLong crc = crc32(0L, Z_NULL, 0);
    crc = crc32 (crc, (Bytef *)buf, size);
    return crc;
}

void print_decoder_frame_err (adcm_decstat_t * decst, const char *fmt, ...)
{
    if (!DEBUG_PRINT_DECODER_FRAME_ERRORS) return;
    char buf[1024] = { '\0' };
    size_t len;
    va_list msg;

//    if (decst) {
        if (decst->frame_errors_suppressed)
            return;
        /*
    if (decstat_n_frame_err (decst) > PRINT_N_FRAME_ERR) {
      printf ("frame error: more errors, suppressed...\n");
      decst->frame_errors_suppressed = 1;
      return;
    }
*/
//    }

    va_start (msg, fmt);
    vsnprintf (buf, sizeof (buf), fmt, msg);
    va_end (msg);

    buf[sizeof (buf) - 1] = '\0';
    len = strlen (buf);
    if ((len > 0) && (buf[len - 1] != '\n'))
        strncat (buf, "\n", sizeof (buf)-strlen(buf));

    fprintf (stderr, "frame error: %s", buf);

    //  if (!adcm_dec_msgbuf) return;
    //  strncat(adcm_dec_msgbuf, buf, ADCM_DEC_MSGBUF_SIZE);
}

void print_decoder_data_err (adcm_decstat_t * decst, const char *fmt, ...)
{
    if (!DEBUG_PRINT_DECODER_DATA_ERRORS) return;
    char buf[1024] = { '\0' };
    size_t len;
    va_list msg;

//    if (decst) {
        if (decst->data_errors_suppressed)
            return;
        if (decstat_n_data_err (*decst) > PRINT_N_DATA_ERR) {
            fprintf (stderr, "data error: more errors, suppressed...\n");
            decst->data_errors_suppressed = 1;
            return;
        }
//    }

    va_start (msg, fmt);
    vsnprintf (buf, sizeof (buf), fmt, msg);
    va_end (msg);

    buf[sizeof (buf) - 1] = '\0';
    len = strlen (buf);
    if ((len > 0) && (buf[len - 1] != '\n'))
        strncat (buf, "\n", sizeof (buf)-strlen(buf));

    fprintf (stderr, "data error: %s", buf);
}
/*****************************************************************************/
/*  Frame print functions                                                    */
/*****************************************************************************/
void adcm_print_raw_frame (const unsigned int *f, size_t len)
{
//    int ferr = ml_frame_validate (f, len);
    /*
  for (i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
    unsigned int d = f[i];
    printf ("[%4d]:%08x  ", i, d);
//    printf ("[%4d]: %08x", i, d);
//    printf ("\n");
  }
  printf ("\n");
*/
    fprintf (stderr, "--------------- Raw frame dump --------[]-\n");

    for (size_t i = 0; i < len; i++) {
        unsigned int d = f[i];

        int c = i % 4;
        if (c == 0)
            fprintf (stderr, "[%04zX]:  ", i);
        fprintf (stderr, "%08x  ", d);
        if (c == 3)
            fprintf (stderr, "\n");
    }
    fprintf (stderr, "-----------------------------------------------\n");
}

void adcm_print_ctrlreq_frame (const unsigned int *f, size_t len)
{

    for (size_t i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
        unsigned int d = f[i];

        printf ("[%4zu]: %08x", i, d);
        printf (" -- ctrl req -- %s addr %04X, data %04X", (d & 0x80000000) ? "read" : "write", (d >> 16) & 0x7FFF, d & 0xFFFF);
        printf ("\n");
    }
}

void adcm_print_ctrlack_frame (const unsigned int *f, size_t len)
{

    for (size_t i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
        unsigned int d = f[i];

        printf ("[%4zu]: %08x", i, d);
        printf (" -- ctrl ack -- ");
        printf ("\n");
    }
}

void adcm_print_status_frame (const unsigned int *f, size_t len)
{

    for (size_t i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
        unsigned int d = f[i];

        printf ("[%4zu]: %08x", i, d);
        printf (" -- status -- ");
        printf ("\n");
    }
}

// data block length in words
size_t adcm_block_len (const adcm_data_header_t * h)
{
    switch (h->id) {
    case 0 :
    case 2 :
    {
        adcm_data_header_0_t *ah = (adcm_data_header_0_t *) h;

        int hdrlen = (sizeof (adcm_data_header_0_t) + 3) / 4;
        int trllen = (sizeof (adcm_data_trailer_0_t) + 3) / 4;
        int datalen = 0;
        if (ah->spw > 0)
            datalen = ((ah->ns/2*2)+ ah->spw - 1) / ah->spw;
        return (hdrlen + datalen + trllen); // hdr, data, trailer
    }
    case 1:                      // stat
    {
        adcm_data_header_1_t *ah = (adcm_data_header_1_t *) h;

        int hdrlen = (sizeof (adcm_data_header_1_t) + 3) / 4;
        int trllen = (sizeof (adcm_data_trailer_t) + 3) / 4;
        int datalen = (ah->wpc * ah->nc);
        return (hdrlen + datalen + trllen); // hdr, data, trailer
    }
    default:                     // generic
        return (sizeof (adcm_data_header_t) + sizeof (adcm_data_trailer_t));
    }
}

// Fast unpacking code
void unpack_samples(quint32 sampmask, const quint32 *iptr, quint32 datablock_ns, AdcmCbufSample *optr)
{
    const int n = datablock_ns/2;
    const quint32 *iptrEnd = &iptr[n];
    while(iptr < iptrEnd) {
        // "offset binary"
        *optr++ = (*iptr >> 16) & sampmask;
        *optr++ = (*iptr)     & sampmask;
        // "2's complement"
        //		  *optr++ = (((*iptr >> 16) & sampmask) + offbin) & sampmask;
        //		  *optr++ = (((*iptr)       & sampmask) + offbin) & sampmask;
        iptr++;
    };
}

//
// Decode frame data. Input is frame data block without header and trailer.
//

int AdcmDataBlockDecoder::adcm_decode_data_blocks (const unsigned int *f, size_t len, struct adcm_el_t * el, adcm_decstat_t * st)
{
    assert(f);
    assert(el);
    assert(st);
    if (!f || !el || !st)
        return (-EINVAL);
    if (len == 0)
        return 0;                 // empty

    int err = 0;
    unsigned int n_blk_ok = 0;
    const adcm_data_header_t *h = NULL;
    const adcm_data_trailer_t *t = NULL;

    const size_t hdrlen = (sizeof (adcm_data_header_t) + 3) / 4;
    const size_t trllen = (sizeof (adcm_data_trailer_t) + 3) / 4;

    const bool var_sample_len = true; //(conf_mem.firmware_prop.bit.sample_len != 0);

    const int end = len;
    int i = 0;
    while (i < end) {
        // begin data block processing
        st->nblk_all++;

        const size_t remain = end - i;
        if (remain < hdrlen + trllen) {
            // if (st) st->nblk_len++;
            print_decoder_data_err (st, "short data block length:%zd, minimum is %zd words\n", remain, hdrlen + trllen);
            goto badblk;
        }
        // set block header, check block type
        h = (adcm_data_header_t *) & f[i];
        switch (h->id) {
        case 0:
        case 1:
        case 2:
            break;
        default:
            print_decoder_data_err (st, "unknown block type: %d\n", h->id);
            goto badblk;
        }

        // calculate block length
        size_t blen = 0;
        const size_t tmp = adcm_block_len (h);
        if (tmp < (hdrlen + trllen)) {
            print_decoder_data_err (st, "bad block len:%d\n", tmp);
            goto badblk;
        }
        if (tmp > remain) {
            print_decoder_data_err (st, "block len:%d > remaining size:%zd words, truncated\n", tmp, remain);
            blen = remain;
            //            goto badblk;
        } else {
            blen = tmp;
        }

        // length is ok. get trailer.
        t = (adcm_data_trailer_t *) & f[i + blen - trllen];

        // additional checks by individual block type
        if (h->id == 0 || h->id == 2) {
            const adcm_data_header_0_t * const ah = (adcm_data_header_0_t *) h;

            if ((ah->spw < ADCM_MIN_SPW) || (ah->spw > ADCM_MAX_SPW)) { // spw hardcoded -- data is corrupt
                print_decoder_data_err (st, "dadc, bad spw value:%d, range %d-%d\n", ah->spw, ADCM_MIN_SPW, ADCM_MAX_SPW);
                goto badblk;
            }
            if ((ah->ns < ADCM_MIN_NS) || (ah->ns > ADCM_MAX_NS)) {  // ns hardcoded -- data is corrupt
                print_decoder_data_err (st, "bad ns value:%d, range %d-%d\n", ah->ns, ADCM_MIN_NS, ADCM_MAX_NS);
                goto badblk;
            }
            if ((ah->bits < ADCM_MIN_BIT) || (ah->bits > ADCM_MAX_BIT)) {  // bits hardcoded -- data is corrupt
                print_decoder_data_err (st, "bad bit value:%d, range %d-%d\n", ah->bits, ADCM_MIN_BIT, ADCM_MAX_BIT);
                goto badblk;
            }
            if (ah->ch >= ADCM_N_CH) {
                print_decoder_data_err (st, "bad ch:%d, range %d-%d\n", ah->ch, 0, ADCM_N_CH);
                goto badblk;
            }
        } else if (h->id == 1) {
            const adcm_data_header_1_t * const sh = (adcm_data_header_1_t *) h;

            if (sh->wpc != 2)
                goto badblk;
            if (sh->nc != 16)
                goto badblk;
        } else {                    // unknown block type
            goto badblk;
        }

        // data block correct. count it.
        switch (h->id) {
        case 0:
            st->nblk_adc_0++;
            break;
        case 1:
            st->nblk_adc_1++;
            //            st->nblk_stat++;
            break;
        case 2:
            st->nblk_adc_2++;
            break;
        }

        // adc data block
        if (h->id == 0 || h->id == 2) {
            int data_good = 1;

            const adcm_data_header_0_t * const ah = (adcm_data_header_0_t *) h;
            const size_t ahdrlen = (sizeof (adcm_data_header_0_t) + 3) / 4;
            const size_t atrllen = (sizeof (adcm_data_trailer_0_t) + 3) / 4;
            const adcm_data_trailer_0_t * const at = (adcm_data_trailer_0_t *) & f[i + blen - atrllen];

            const unsigned int ch = ah->ch;
            //            const unsigned int spw = ah->spw;
            //            const unsigned int nwords = (ah->ns + spw - 1) / spw;
            // TODO: check data block consistency
            if (data_good)
                st->nblk_adc_ok++;

            // Store event
            AdcmEvent *ev = el->getEvent(ah->ec);
            assert(ev);
            // set event header
            const quint64 ts = (quint64)(at->tslo) + (useLowerTimeStampBits ? 0 : ((quint64)(at->tshi) << 32));

            unsigned int tmpEventNumber = ah->ec;
            // workaround for 16-bit event counter
            {
                if (ah->ec < prev_ah_ec) {
                    ec_offset += 65536;
                }
                tmpEventNumber += ec_offset;
                prev_ah_ec = ah->ec;
            }
            ev->setEventNumber(tmpEventNumber);
            if (h->id == 0)
                ev->setBurstNumber(ah->bc);
            else
                ev->setBurstNumber(0);
            ev->setEventTimestamp(ts);            //FIXME
            if (prev_ec != tmpEventNumber) {
                if (prev_ec+1 != tmpEventNumber && prev_ec!=0) {
                    st->nblk_adc_evnum++;
                    print_decoder_data_err (st,
                                            "event number %u changed to %u (delta=%u)",
                                            prev_ec, tmpEventNumber,
                                            tmpEventNumber - prev_ec);
                }
                prev_ec = tmpEventNumber;
            }

            const int datablock_ns = ah->ns & ~1;
            const bool last_fragm = !var_sample_len || ((ah->ns & 1) != 0);

            //      printf("raw: ev %d, ts %16llX, ch %d, len %d\n", ev->ec, ev->ts, ch, datablock_ns);
            struct adcm_chbuf_t *evc = ev->new_ev_chbuf(ch); // may return NULL
            if (evc) {
                // TODO: optimize

                /*
        if (evc->flags & ADCM_CH_HAS_DATA) {
          if (st)
            st->nblk_adc_dup++;
          print_decoder_data_err (st, "duplicate data for channel %d\n", ch);
          err = 1;
          ev->flags |= ADCM_EV_EDUPCH;
          data_good = 0;
//    break;
        }
*/
                AdcmCbufWaveform &w = evc->w;
                if ((evc->flags & ADCM_CH_HAS_DATA) == 0) {
                    evc->flags |= ADCM_CH_HAS_DATA;
                    evc->flags &= ~ADCM_CH_FINISHED;
                    evc->ch = ch;
                    w.clear();
                    evc->ts = ts;
                    evc->w_bits = ah->bits;
                }
                if (evc->flags & ADCM_CH_FINISHED) {
                    print_decoder_data_err (st, "duplicate waveform for channel %d\n", ch);
                    st->nblk_adc_dup++;
                    //          ev->flags |= ADCM_EV_EDUPCH;
                }
                if (last_fragm)
                    evc->flags |= ADCM_CH_FINISHED;
                const int start_nsamp = evc->len();
                const size_t newSize = evc->len() + datablock_ns;
                if (newSize <= ADCM_MAX_N_SAMP) {
                    //	      evc->len += ah->ns&~1; // LSB is last-fragment-flag
                    w.reserve(newSize);
                    w.resize(newSize);
                } else {
                    print_decoder_data_err (st, "too many samples for channel %d: %d, limit %d\n", ch, newSize, ADCM_MAX_N_SAMP);
                    err = 1;
                    //          ev->flags |= ADCM_EV_EDUPCH;
                    data_good = 0;
                    break;
                }
                //        printf("ev->ec=%d, ch=%d, evc->len=%d, datablock_ns=%d, start_nsamp=%d\n", ev->ec, ch, evc->len, datablock_ns, start_nsamp);
                //   print_decoder_data_err(st, "ec:%04X, bc:%04X, ts %16llX\n", h->ec, h->bc, ts);
                //   print_decoder_data_err(st, "ADC data, bits:%d, spw:%d, ns:%d, ch:%d\n", h->bits, h->spw, h->ns, h->ch);

                // get samples - Fast unpacking code
                {
                    const quint32 sampmask = (1 << ah->bits) - 1;
                    const quint32 *iptr = &f[i + ahdrlen]; // unsigned???
                    AdcmCbufSample *optr = &w[start_nsamp];
                    unpack_samples(sampmask, iptr, datablock_ns, optr);
                }
            } else {                  // (!evc)
                //   print_decoder_data_err(st, "eventlist problem: can't store event #%u\n", h->ec);
            }

        } else if (h->id == 1) {    /* end if (h->id == 0) */
            /*
      adcm_data_header_1_t *ah = (adcm_data_header_1_t *)h;
      adcm_data_trailer_1_t *at;
      size_t ahdrlen = (sizeof(adcm_data_header_1_t)+3)/4;
      size_t atrllen = (sizeof(adcm_data_trailer_1_t)+3)/4;
      at = (adcm_data_trailer_1_t *) & f[i + blen - atrllen];
// stat data block

*/
        } else {
            //            logger.warn(QString("Unknown data block id %1").arg(h->id));
            // unknown data block - do nothing
        }

        // good block
        i += blen;
        // i+=trllen;
        n_blk_ok++;
        continue;

    }
    if (err)
        return (-1);
    else
        return n_blk_ok;

badblk:
    st->nblk_bad++;
    st->pblk_err = i;
    {
        char str[256], tmpstr[256];

        str[255]=0; tmpstr[255]=0;
        //  printf ("Bad data block at %04X", i);
        snprintf (str, 255, "Bad data block at %04X", i);
        if (h) {
            unsigned int bblen = adcm_block_len (h);

            snprintf (tmpstr, 255, ", id:%u, len:%u", h->id, bblen);
            strncat (str, tmpstr, 255-strlen(str));
            /*
    if (h->id == 0) {
          adcm_data_header_0_t *ah = (adcm_data_header_0_t *)h;
      snprintf (tmpstr, 255, ", spw:%d, ns:%d", ah->spw, ah->ns);
      strncat (str, tmpstr, 255);
    }
*/
//            if ((bblen >= trllen) && (i + bblen - trllen <= len)) {
//                unsigned int crc;

//                t = (adcm_data_trailer_t *) & f[i + bblen - trllen];
//                crc = t->crc;
//                snprintf (tmpstr, 255, ", crc:%08X %s", crc, crc == ADCM_BLK_CRC_HACK ? "ok" : "bad");
//                strncat (str, tmpstr, 255-strlen(str));
//            }

        }
        sprintf (tmpstr, "\n");
        strncat (str, tmpstr, 255-strlen(str));
        print_decoder_data_err (st, str);
        err = 1;
        //  ev->flags |= ADCM_EV_BADBLK;
        return err;
    }
}

/*****************************************************************************/
/*  Frame search                                                             */
/*****************************************************************************/

size_t tmp_p_start=0, tmp_p_end=0;

frame_search_t FrameFinder::find_frame_forw (const unsigned int *buf, size_t pos, size_t end, size_t * fnewpos,
                                             size_t * ffstart, size_t * ffend, adcm_decstat_t * decst)
{
    assert(buf);
    size_t p, newpos, fstart = pos, fend = pos;
    int found = 0, resync = 0;
//    int resync_start = pos;
//    int nomsg = decst?(decst->frame_errors_suppressed):0;
    //  printf("searching %08X-%08X\n", pos, end);
    if (tmp_p_start>pos || tmp_p_end>pos) {tmp_p_start=0; tmp_p_end=0;}
    p = pos;
    while (p + sizeof (frame_hdr_t)+1 < end) { // hdr+crc
        const frame_hdr_t *fh = (frame_hdr_t *) & buf[p];
        //	int bad_sync = 0;
        //	int bad_len = 0;
        int frame_ok = 1;

        // check sync
        if (fh->sync != ML_FRAME_SYNC) {
            //	  bad_sync = 1;
            frame_ok=0;
            if (!resync) {
                print_decoder_frame_err (decst, "search: [%08zX]: bad frame sync %04X\n", p, fh->sync);
//                if (!decst || !decst->frame_errors_suppressed) {
//                    int pre = p;
//                    if (p>5) pre = p-5; else pre=p;
//                    adcm_print_raw_frame(&buf[pre], 10);
//                }
            }
        }

        // check len
        if (frame_ok && (fh->len < ML_MIN_RAW_FRAME_WORDS || fh->len > ML_MAX_RAW_FRAME_WORDS)) {
            //	  bad_len = 1;
            frame_ok=0;
            if (!resync) {
                print_decoder_frame_err (decst, "search: [%08zX]: bad frame len %04X\n", p, fh->len);
//                if (!decst || !decst->frame_errors_suppressed) {
//                    int pre = p;
//                    if (p>5) pre = p-5; else pre=p;
//                    adcm_print_raw_frame(&buf[pre], 10);
//                }
            }
        }

        // check if frame is truncated
        if (frame_ok && p+fh->len > end) { // FIXME
            frame_ok=0;
            //	   p = end;
            break; // incomplete frame
        }


        // check crc
        if (frame_ok && DO_CHECK_FRAME_CRC) {
            const quint32 crc = buf[p+fh->len- 1];
            if (crc != 0) {
                const quint32 mycrc = calculate_crc32 (&buf[p], (fh->len-1)*4);
                if (crc != mycrc) {
                    if (decst)
                        decst->nf_crc++;
                    print_decoder_frame_err (decst,
                                             "search: [%08zX]: bad crc %08X (calculated %08X), frame type:%04X, len:%u, %04X -> %04X, seq:%u\n",
                                             p/*fend - 1*/, crc, mycrc, fh->type, fh->len, fh->src, fh->dst, fh->seq);
                    frame_ok = 0;
                }
            }
        }
        // end of checks

        if (!frame_ok) {
            if (!resync) {
                resync = 1;
                if (decst) decst->n_resync++;
//                resync_start=p;
                print_decoder_frame_err(decst, "resync at %08X", p);
            }
            p++;
            continue;
        }

        if (fh->type == FRAME_TYPE_ADC_DATA) {
            // check sequence number (should increment for each source/dest address pair)
            unsigned short exp_seq = (prev_seq+1) & 0xFFFF;
            if (seq_initialized && (exp_seq != fh->seq)) {
                if (decst)
                    decst->nf_seq++;
                //	  frame_ok=0;
                if (!resync) {
                    print_decoder_frame_err (decst, "search: [%08zX]: bad frame seq %04X, expected %04X\n", p, fh->seq, exp_seq);
//                    if (!decst || !decst->frame_errors_suppressed) {
//                        int pre = p;
//                        if (p>5) pre = p-5; else pre=p;
//                        adcm_print_raw_frame(&buf[pre], 10);
//                    }
                }
            }
            prev_seq = fh->seq;
            seq_initialized = true;
        }


        fstart = p;
        fend = fstart + fh->len;

        // frame ok.
        // printf("found frame, seq %d\n", fh->seq);

        /*
// AUCHTUNG !!! unreachable code! look above: if (resync) continue;
    if (p<end && resync && decst) {
      if (decst) decst->n_resync++;
      print_decoder_frame_err (decst, "search: [%08zX]: resync, %zu words skipped\n", fstart, fstart - pos);

if (debug_level_ge(1)) {
//      if (debug_level_ge(1))
          if (!nomsg)
          {
            printf("Resync, pos=%d, prev=%d\n", p, tmp_p_start);
            if (tmp_p_start<resync_start) {
                  int i1= tmp_p_start;
                  int i2=tmp_p_end;
                  int err = ml_frame_validate(&buf[i1], i2-i1);
              printf("Previous frame was %s:\n", err?"BAD":"Ok");
          adcm_print_raw_frame(&buf[i1], i2-i1);
                }
        adcm_print_raw_frame(&buf[resync_start], p-resync_start);
                {
                  int i1 = p;
                  int i2=p+ML_MAX_RAW_FRAME_WORDS;
                  if (i1>end) i1=end;
                  if (i2>end) i2=end;
                  int i=i1+1;
                  while(i<i2) {
            const frame_hdr_t *fhl = (frame_hdr_t *) & buf[i];
                        if (fhl->sync == ML_FRAME_SYNC) break;
                        i++;
                  }
                  i2=i;
                  if (i1<i2) {
                    int err = ml_frame_validate(&buf[i1], i2-i1);
                        if (err) {
                  printf("Next frame BAD:\n");
                          adcm_print_raw_frame(&buf[i1], i2-i1);
                        } else {
                  printf("Next frame Ok\n");
                        }
                  }
                }
                printf("\n");
          }
} // end if (debug_level_ge(1))
    }
*/
        resync = 0;
        found = 1;
        p = fend;
        if (decst) {
            decst->nf++;
            switch (fh->type) {
            case FRAME_TYPE_ECHO:
                decst->nf_echo++;
                break;
            case FRAME_TYPE_ADC_DATA:
                decst->nf_dadc++;
                break;
            case FRAME_TYPE_CTRL_REQ:
                decst->nf_creq++;
                break;
            case FRAME_TYPE_CTRL_ACK:
                decst->nf_cack++;
                break;
            case FRAME_TYPE_STAT:
                decst->nf_stat++;
                break;
            default:
                decst->nf_unkn++;
            }
        }
        break;
    }


    newpos = p;
    tmp_p_start=fstart;
    tmp_p_end=fend;

    if (decst) {
        decst->nw += (fend-fstart);
        decst->lastpos = newpos;
    }

    if (fnewpos)
        *fnewpos = newpos;
    if (found) {
        if (ffstart)
            *ffstart = fstart;
        if (ffend)
            *ffend = fend;
    }

    frame_search_t ret = 0;
    if (found)
        ret |= FRAME_SEARCH_BIT_FOUND;
    if (resync)
        ret |= FRAME_SEARCH_BIT_RESYNC;
    return (ret);
}

frame_search_t find_frame_forw_m (const unsigned int *buf, size_t pos, size_t end, size_t * fnewpos,
                                  struct framelist_t *fl, adcm_decstat_t * decst)
{
    assert(buf);
    size_t p, newpos, fstart = pos, fend = pos;
    int found = 0, resync = 0;
//    size_t resync_start = pos;
    size_t flc = 0; // number of frames in framelist

    if (fl)
        fl->count=0;

    //  printf("searching...\n");
    if (tmp_p_start>pos || tmp_p_end>pos) {tmp_p_start=0; tmp_p_end=0;}
    p = pos;
    while ((p + sizeof (frame_hdr_t) < end) && (flc < FRAMELIST_CAP)) {
        const frame_hdr_t *fh = (frame_hdr_t *) & buf[p];
        //	int bad_sync = 0;
        //	int bad_len = 0;
        int frame_ok = 1;

        // check sync
        if (fh->sync != ML_FRAME_SYNC) {
            //	  bad_sync = 1;
            frame_ok=0;
            if (!resync) {
                print_decoder_frame_err (decst, "search: [%08zX]: bad frame sync %04X\n", p, fh->sync);
//                if (!decst->frame_errors_suppressed) {
//                    int pre = p;
//                    if (p>5) pre = p-5; else pre=p;
//                    adcm_print_raw_frame(&buf[pre], 10);
//                }
            }
        }

        // check len
        if (frame_ok && (fh->len < ML_MIN_RAW_FRAME_WORDS || fh->len > ML_MAX_RAW_FRAME_WORDS)) {
            //	  bad_len = 1;
            frame_ok=0;
            if (!resync) {
                print_decoder_frame_err (decst, "search: [%08zX]: bad frame len %u\n", p, fh->len);
//                if (!decst->frame_errors_suppressed) {
//                    int pre = p;
//                    if (p>5) pre = p-5; else pre=p;
//                    adcm_print_raw_frame(&buf[pre], 10);
//                }
            }
        }

        // check if frame is truncated
        if (frame_ok && p+fh->len > end) {
            frame_ok=0; p = end;
        }

        // check crc
        if (frame_ok && DO_CHECK_FRAME_CRC) {
            const quint32 crc = buf[p+fh->len- 1];
            const quint32 mycrc = calculate_crc32 (&buf[p], (fh->len-1)*4);
            if (crc != mycrc) {
                if (decst)
                    decst->nf_crc++;
                print_decoder_frame_err (decst,
                                         "search: [%08zX]: bad crc %08X (calculated %08X), frame type:%04X, len:%u, %04X -> %04X, seq:%u\n",
                                         fend - 1, crc, mycrc, fh->type, fh->len, fh->src, fh->dst, fh->seq);
                frame_ok = 0;
            }
        }
        if (!frame_ok) {
            if (!resync) {
                resync = 1;
//                resync_start=p;
            }
            p++;
            continue;
        }

        fstart = p;
        fend = fstart + fh->len;

        // frame ok.
        // printf("found frame, seq %d\n", fh->seq);
        if (p<end && resync && decst) {
            decst->n_resync++;
            print_decoder_frame_err (decst, "search: [%08zX]: resync, %zu words skipped\n", fstart, fstart - pos);
#ifdef DEBUG
            //      if (debug_level_ge(1))
            if (!decst->frame_errors_suppressed) {
                printf("pos=%zu, prev=%zu\n", p, tmp_p_start);
//                if (tmp_p_start<resync_start) {
//                    printf("\nPrevious frame was:\n");
//                    adcm_print_raw_frame(&buf[tmp_p_start], tmp_p_end-tmp_p_start);
//                }
//                adcm_print_raw_frame(&buf[resync_start], p-resync_start);
//                printf("\n");
            }
#endif
        }
        resync = 0;
        found = 1;
        p = fend;
        if (decst) {
            decst->nf++;
            switch (fh->type) {
            case FRAME_TYPE_ECHO:
                decst->nf_echo++;
                break;
            case FRAME_TYPE_ADC_DATA:
                decst->nf_dadc++;
                break;
            case FRAME_TYPE_CTRL_REQ:
                decst->nf_creq++;
                break;
            case FRAME_TYPE_CTRL_ACK:
                decst->nf_cack++;
                break;
            case FRAME_TYPE_STAT:
                decst->nf_stat++;
                break;
            default:
                decst->nf_unkn++;
            }
        }

        // Count statistics
        if (decst) {
            decst->nw += (fend-fstart);
        }

        // Add framelist entry
        if (fl) {
            if (flc < FRAMELIST_CAP) {
                fl->start[flc] = fstart;
                fl->end[flc] = fend;
                flc++;
                fl->count = flc;
            } else
                break;
        }
        //    break;
    }

    newpos = p;
    if (decst) {
        decst->lastpos = p;
    }

    tmp_p_start=fstart;
    tmp_p_end=fend;

    if (fnewpos)
        *fnewpos = newpos;

    frame_search_t ret = 0;
    if (found)
        ret |= FRAME_SEARCH_BIT_FOUND;
    if (resync)
        ret |= FRAME_SEARCH_BIT_RESYNC;
    return (ret);
}

void FrameFinder::reset()
{
    prev_seq = 0;
    seq_initialized = false;
}

void AdcmDataBlockDecoder::reset()
{
    prev_ah_ec=decoderIndex;
    ec_offset=0;
    prev_ec = decoderIndex;
}
