//
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "allconfig.h"

#include <cmath>
#include <QDir>
#include <QDebug>

#include "adcm16.h"
#include "dsplib/fir.h"
#include "dsplib/fitter.h"
#include "adcmraw/adcm_mem.h"

const int CFG_VERSION = 0x00000086;

template<class T>
bool INRANGE(const T &x, const T &min, const T &max)
{
    return !(x < min || x > max);
}

int clear_cfgcalzero (AllConfig * cfg)
{
    if (!cfg)
        return (-1);
    for (unsigned int c = 0; c < ADC_NCH; c++)
        if (cfg->adc.trig[c] & ADCM_CCTRL_BIT_CMPEN) {
        cfg->dec.calzero.nent[c] = 0;
        cfg->dec.calzero.base[c] = 0;
        cfg->dec.calzero.rms[c] = 0;
        cfg->dec.calzero.asym[c] = 0;
    }
    return (0);
}

AllConfig::AllConfig()
{
    version = CFG_VERSION;

    stor_raw.bn = QString(QDir::separator()) + "data";
    stor_raw.fn = "adcm_raw.dat";

    stor_dev.enable = 1;
    stor_dev.bn = QString(QDir::separator()) + "data";
    stor_dev.fn = "adcm.dat";

    // Resolve cross-dependencies
    for (int c = 0; c < std::min(dec.int2.size(), adc.size()); c++) {
        dec.int2[c] = adc.recwin[c]-adc.latency-(adc.matchwin-adc.prewin);
    }
}

bool AllConfig::operator==(const AllConfig &other) const
{
    if(DEBUG_PRINT_CONFIG_DIFFERENCE){
        if(version != other.version)
            qDebug() << "versions differs";
        if(adc != other.adc)
            qDebug() << "adc differs";
        if(stor_raw != other.stor_raw)
            qDebug() << "stor_raw differs";
        if(stor_dev != other.stor_dev)
            qDebug() << "stor_dev differs";
        if(dec != other.dec)
            qDebug() << "dec differs";
        if(hist != other.hist)
            qDebug() << "hist differs";
        if(run != other.run)
            qDebug() << "run differs";
    }
    return true
            && version == other.version
            && adc == other.adc
            && stor_raw == other.stor_raw
            && stor_dev == other.stor_dev
            && dec == other.dec
            && hist == other.hist
            && run == other.run
            ;
}

void AllConfig::print() const
{
    fprintf (stderr, "version: %d\n", version);

    adc.print();
    dec.print();
    fprintf (stderr, "Calibrations:\n");
    for (int c = 0; c < dec.calzero.nch(); c++) {
        fprintf (stderr, "calzero[%u] = %g\n", c, dec.calzero.get(c));
    }
}

bool AllConfig::isValid() const
{
    bool ok = true;
    //  double ts;

    if (version != CFG_VERSION) {
        fprintf (stderr, "bad version %d, not %d\n", version, CFG_VERSION);
        return false;
    }
    for (int c = 0; c < adc.size(); c++) {
        if ((adc.bits[c] < 0) || (adc.bits[c] > 32)) {
            fprintf (stderr, "bad adc.bits[%u]: %d\n", c, adc.bits[c]);
            ok = false;
        }
    }
    for (int c = 0; c < adc.step.size(); c++) {
        if ((adc.step[c] <= 0) || (adc.step[c] > 1e9)) {
            fprintf (stderr, "bad adc.step[%u]: %g\n", c, adc.step[c]);
            ok = false;
        }
    }
    for (int c = 0; c < adc.size(); c++) {
        if (adc.recwin[c] >= ADCM_MAX_N_SAMP) {
            fprintf (stderr, "bad adc.recwin[%u]: %d\n", c, adc.recwin[c]);
            ok = false;
        }
    }
    //  if (!INRANGE (adc.prewin, ADC_PREWIN_MIN, ADC_PREWIN_MAX)) {
    //    fprintf (stderr, "bad adc.prewin: %d\n", adc.prewin);
    //    ok = false;
    //  }
    //  if (!INRANGE (adc.matchwin, ADC_MATCHWIN_MIN, ADC_MATCHWIN_MAX)) {
    //    fprintf (stderr, "bad adc.matchwin: %d\n", adc.matchwin);
    //    ok = false;
    //  }
    //  if (!INRANGE (adc.latency, ADC_LATENCY_MIN, ADC_LATENCY_MAX)) {
    //    fprintf (stderr, "bad adc.latency: %d\n", adc.latency);
    //    ok = false;
    //  }

    if ((dec.fprof < 0) || (dec.fprof > 100)) {
        fprintf (stderr, "bad dec.fprof: %d\n", dec.fprof);
        ok = false;
    }
    if ((dec.fititer < 0) || (dec.fititer > 10000)) {
        fprintf (stderr, "bad dec.fititer: %d\n", dec.fititer);
        ok = false;
    }
    if ((dec.fitside < 0) || (dec.fitside > 63)) {
        fprintf (stderr, "bad dec.fitside: %d\n", dec.fitside);
        ok = false;
    }
    for (unsigned int c = 0; c < ADC_NCH; c++) {
        //    ts = adc.step[c];
        //    if ((dec.search_start[c]<0) || (dec.search_start[c]>RECWIN)) {printf("bad dec.search_start[%d]: %d\n", c, dec.search_start[c]); ok=0;}
        //    if ((dec.search_end[c]<0) || (dec.search_end[c]>RECWIN)) {printf("bad dec.search_end[%d]: %d\n", c, dec.search_end[c]); ok=0;}
        if ((dec.baserc[c] < 0) || (dec.baserc[c] > 100)) {
            fprintf (stderr, "bad dec.baserc[%u]: %g ns\n", c, dec.baserc[c]);
            ok = false;
        }
        if ((dec.tadj[c] < -200) || (dec.tadj[c] > 200)) {
            fprintf (stderr, "bad dec.tadj[%u]: %g ns\n", c, dec.tadj[c]);
            ok = false;
        }
        if ((dec.pnf[c] < -100) || (dec.pnf[c] > 100)) {
            fprintf (stderr, "bad dec.pnf[%u]: %g\n", c, dec.pnf[c]);
            ok = false;
        }
        if (!INRANGE (dec.blr1[c], -1*(int)ADCM_MAX_N_SAMP, (int)ADCM_MAX_N_SAMP)) {
            fprintf (stderr, "bad dec.blr1[%u]: %d\n", c, dec.blr1[c]);
            ok = false;
        }
        if (!INRANGE (dec.blr2[c], dec.blr1[c], (int)ADCM_MAX_N_SAMP)) {
            fprintf (stderr, "bad dec.blr2[%u]: %d\n", c, dec.blr2[c]);
            ok = false;
        }
        if (!INRANGE (dec.int1[c], -1*(int)ADCM_MAX_N_SAMP, (int)ADCM_MAX_N_SAMP)) {
            fprintf (stderr, "bad dec.int1[%u]: %d\n", c, dec.int1[c]);
            ok = false;
        }
        if (!INRANGE (dec.int2[c], dec.int1[c], (int)ADCM_MAX_N_SAMP)) {
            fprintf (stderr, "bad dec.int2[%u]: %d\n", c, dec.int2[c]);
            ok = false;
        }
        if ((dec.fir.size() > (int)c) && !fir_valid (&dec.fir[c])) {
            fprintf (stderr, "bad FIR[%u]\n", c);
            ok = false;
        }
    }
    if ((hist.enable < 0) || (hist.enable > 1)) {
        fprintf (stderr, "bad hist.enable: %d\n", hist.enable);
        ok = false;
    }

    if (ok)
        return true;
    else {
        //    cfg_dump(cfg);
        return false;
    }
}

bool AllConfig::enabled_ch_has_calzero() const
{
    bool ok = false;

    for (int c = 0; c < dec.calzero.nch(); c++) {
        if (adc.ch_enabled_offset_calibration (c) && (dec.calzero.nent[c] > 0))
            ok = true;
    }
    return ok;
}

bool AllConfig::enabled_ch_has_calavg() const
{
    bool ok = false;

    for (int c = 0; c < dec.calavg.nent.size(); c++) {
        if (adc.ch_enabled_shape_calibration (c) && (dec.calavg.nent[c] > 0))
            ok = true;
    }
    return ok;
}

bool AllConfig::enabled_ch_has_fir () const
{
    bool ok = true;

    for (int c = 0; c < dec.fir.size(); c++) {
        if (adc.ch_enabled_reconstruction (c))
            if (!fir_valid (&dec.fir[c]) || (dec.fir[c].klen <= 0))
                ok = false;
    }
    return ok;
}

double AllConfig::calavg_integrate (int ch, int start, int end) const
{

    int i1, i2;
    double val;

    if (dec.calavg.nent[ch] <= 0)
        return (0);

    //  st = 1;
    val = 0;
    i1 = start;
    i2 = end;
    if (i1 < 0) {
        //    i1 = 0;
    }
    if (i1 >= (int)ADCM_MAX_N_SAMP) {
        i1 = (int)ADCM_MAX_N_SAMP - 1;
        //    st = 0;
    }
    if (i2 < 0) {
        i1 = 0;
        //    st = 0;
    }
    if (i2 >= (int)ADCM_MAX_N_SAMP) {
        i2 = (int)ADCM_MAX_N_SAMP - 1;
        //    st = 0;
    }
    for (int i = i1; i < i2; i++)
        val += dec.calavg.data[ch][i];
    if (i2 > i1)
        val /= (double) (i2 - i1);
    val /= dec.calavg.nent[ch];
    //    val -= getcalzero(cfg, c);

    return val;
}

double AllConfig::calavg_baseline (unsigned int ch) const
{
    return calavg_integrate (ch, adc.prewin + dec.blr1[ch], adc.prewin + dec.blr2[ch]);
}

double AllConfig::calavg_integral (unsigned int ch) const
{
    return (calavg_integrate (ch, adc.prewin + dec.int1[ch], adc.prewin + dec.int2[ch]));
}

double AllConfig::calavg_preintegral (unsigned int ch) const
{
    return (calavg_integrate (ch, 0, adc.prewin + dec.int1[ch]));
}

double AllConfig::calavg_postintegral (unsigned int ch) const
{
    return (calavg_integrate (ch, adc.prewin + dec.int2[ch], adc.recwin[ch]));
}

//double AllConfig::getcalzero (unsigned int ch) const
//{
//    double cz;

//    if (ch >= ADC_NCH)
//        return 0;
//    if (dec.calzero.nent[ch] > 0)
//        cz = (double) dec.calzero.base[ch] / dec.calzero.nent[ch];
//    else
//        //    cz = adc.offs[ch];
//        cz = 1<<(adc.bits[ch]-1);
//    //	cz=0;
//    return cz;
//}

//int AllConfig::getcalzero_all (unsigned int nch, float *result) const
//{
//    if (nch > ADC_NCH)
//        return (-1);

//    for(unsigned int ch=0; ch<nch; ch++) {
//        if (dec.calzero.nent[ch] > 0)
//            result[ch] = dec.calzero.base[ch] / dec.calzero.nent[ch];
//        else
//            //      result[ch] = adc.offs[ch];
//            result[ch] = 1<<(adc.bits[ch]-1);
//        //	result[ch]=0;
//    }
//    return 0;
//}

int AllConfig::thr_mv_to_value (const CalZeroConfig &calZero, const AdcConfig &adc, unsigned int ch, double mv)
{
//    double val = mv * AdcConfig::mv_to_value_scaler () + calZero.get(ch);
    double val = mv + calZero.get(ch);

    if (val < adc.adc_rawdata_minval(ch))
        val = adc.adc_rawdata_minval(ch);
    if (val > adc.adc_rawdata_maxval(ch))
        val = adc.adc_rawdata_maxval(ch);
    //  fprintf (stderr, "mv_to_value: %f, %u\n", mv, (unsigned int) val);
    return val;
}

int AllConfig::thr_mv_to_value (unsigned int ch, double mv) const
{
    return thr_mv_to_value(dec.calzero, adc, ch, mv);
}

//double thr_value_to_mv (const CalZeroConfig &calZero, unsigned int ch, unsigned int val)
//{
//    double mv = val / AdcConfig::mv_to_value_scaler () - calZero.get(ch);
//    //  fprintf (stderr, "value_to_mv: %d, %f\n", val, mv);
//    return (int) mv;
//}

//double AllConfig::thr_value_to_mv (unsigned int ch, unsigned int val) const
//{
//    return ::thr_value_to_mv(dec.calzero, ch, val);
//}

int AllConfig::get_dec_fitside(const conf_mem_t &conf_mem, const DecoderConfig &dec, int ch)
{
    int default_value = 6;

    Q_ASSERT(ch < dec.pwid.size());
//    Q_ASSERT(ch < adc.step.size());
    double step_ns = conf_mem.getMasterAdcStepNs();

    int fside = 0;
    double k = 1.0, b = 0;
    if (dec.fprof == FIT_PROFILE_SUPERFAST) {k = 0.7; }
    if (dec.fprof == FIT_PROFILE_FAST) {k = 1.0; }
    if (dec.fprof == FIT_PROFILE_OPTIMAL) { k=1.2; b = 1; }
    if (dec.fprof == FIT_PROFILE_PRECISION) {k = 1.6; b = 1; }
    if (dec.fprof == FIT_PROFILE_CUSTOM) {fside = dec.fitside;}

    if (fside <= 0) // auto
        fside = (int)((double)dec.pwid.at(ch) / step_ns * k + b);
    if (fside<=0 || fside > 63)
        fside = default_value;
    if (fside < 2)
        fside = 2;
    // safe check
    int maxfside = (MAX_FITGAUS_POINTS-1)/2;
    if (fside > maxfside)
        fside = maxfside;

    //  printf("ch %2d: pwid=%f, k=%.1f, b=%.0f  fside=%d\n", ch, dec.pwid[ch], k, b, fside);
    return fside;
}

//int AllConfig::dt_cut_good(double dt) const
//{
//    if (!dec.dt_cut_en)
//        return 1;
//    return (dec.dt_cut_l <= dt && dt <= dec.dt_cut_r);
//}
