//
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "calzeroconfig.h"

#include <cstdio>

#include "adcm16.h"
#include "adcconfig.h"

CalZeroConfig::CalZeroConfig(int numChannels)
{
    resize(numChannels);
}

void CalZeroConfig::resize(int n)
{
    initialized_nch = n;
    nent.resize(n);
    base.resize(n);
    rms.resize(n);
    asym.resize(n);
}

bool CalZeroConfig::operator==(const CalZeroConfig &other) const
{
    return true
            && nent == other.nent
            && base == other.base
            && rms == other.rms
            && asym == other.asym
            ;
}

int CalZeroConfig::nch() const
{
    return initialized_nch;
}

void CalZeroConfig::print() const
{
    for (int c = 0; c < nch(); c++) {
        //	fprintf(stderr, "calzero[%d]=%g\n", getcalzero(calzero, c));
        if (nent[c] <= 0)
            continue;
        fprintf (stderr, "calzero[%u]: nent=%d, base=%g, rms=%g, asym=%g\n",
                 c, nent[c], base[c], rms[c], asym[c]);
    }
}

void CalZeroConfig::clear_calzero_ch(int ch)
{
    Q_ASSERT(ch < nch());
    if (ch >=nch())
        return;
    nent[ch] = 0;
    base[ch] = 0;
    rms[ch] = 0;
    asym[ch] = 0;
}

CalZeroConfig &CalZeroConfig::operator+=(const CalZeroConfig &rhs)
{
    if (nch() < rhs.nch())
        resize(rhs.nch());
    for(int c=0; c<rhs.nch(); c++) {
        nent[c] += rhs.nent[c];
        base[c] += rhs.base[c];
        rms[c] += rhs.rms[c];
        asym[c] += rhs.asym[c];
    }
    return *this;
}

const CalZeroConfig &CalZeroConfig::operator+(const CalZeroConfig &other) const
{
    return CalZeroConfig(*this) += other;
}

void CalZeroConfig::clear()
{
    initialized_nch = 0;
    nent.clear();
    base.clear();
    rms.clear();
    asym.clear();
}

double CalZeroConfig::get(int ch) const
{
    if (ch >= nch())
        return 0;
    Q_ASSERT(ch < std::min(base.size(), nent.size()));
    if (nent[ch] > 0)
        return (double) base[ch] / nent[ch];
    return 0;
}

double CalZeroConfig::getcalnoiserms (int ch) const
{
    if (ch >= std::min(rms.size(), nent.size()))
        return 0;
    double cz;

    if (ch >= nch())
        return (0);
    if (nent[ch] > 0)
        cz = (double) rms[ch] / nent[ch];
    else
        cz = 1.0;
    return cz;
}


void CalZeroConfig::clear_enabled(const AdcConfig &adc)
{
    for (int c = 0; c < nch(); c++) {
        if (!adc.ch_enabled_offset_calibration (c))
            continue;
        clear_calzero_ch(c);
    }
}
