//
//    Copyright 2013 Ivan Filippov
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QString>
#include <QSettings>
#include <QDebug>

#include "allconfig.h"

template<class T>
void writeQVectorToIni(QSettings *settings, QString name, const QVector<T> &values)
{
    settings->beginWriteArray(name);
    for(int i=0; i< values.size(); ++i){
        settings->setArrayIndex(i);
        settings->setValue("val", values[i]);
    }
    settings->endArray();
}

void writeQVectorToIni(QSettings *settings, QString name, const QVector<int> &values)
{
    QStringList sl;
    Q_FOREACH(int item, values) {
        sl.push_back(QString::number(item));
    }
    settings->setValue(name, sl);
}

void writeQVectorToIni(QSettings *settings, QString name, const QVector<unsigned int> &values)
{
    QStringList sl;
    Q_FOREACH(unsigned int item, values) {
        sl.push_back(QString::number(item));
    }
    settings->setValue(name, sl);
}

void writeQVectorToIni(QSettings *settings, QString name, const QVector<double> &values)
{
    QStringList sl;
    Q_FOREACH(double item, values) {
        sl.push_back(QString::number(item, 'e', 16));
    }
    settings->setValue(name, sl);
}

void writeQVectorToIni(QSettings *settings, QString name, const QVector<fwinf_t> &values)
{
    QStringList sl;
    Q_FOREACH(fwinf_t item, values) {
        sl.push_back(QString::number(item));
    }
    settings->setValue(name, sl);
}

void writeQVectorToIni(QSettings *settings, QString name, const QVector<intfir_t> &values)
{
    settings->beginWriteArray(name);
    for(int i=0; i<values.size(); ++i) {
        settings->setArrayIndex(i);
        values[i].writeToIni(settings, "unused");
    }
    settings->endArray();
}


template<typename T>
void readQVectorFromIni(QSettings *settings, const QString &name, QVector<T> &value);

void readQVectorFromIni(QSettings *settings, const QString &name, QVector<int> &value)
{
    if (!settings->contains(name)) return;
    value.clear();
    QStringList sl = settings->value(name).toStringList();
    Q_FOREACH(QString s, sl) {
        value.push_back(s.toInt());
    }
    settings->setValue(name, sl);
}

void readQVectorFromIni(QSettings *settings, const QString &name, QVector<unsigned int> &value)
{
    if (!settings->contains(name)) return;
    value.clear();
    QStringList sl = settings->value(name).toStringList();
    Q_FOREACH(QString s, sl) {
        value.push_back(s.toUInt());
    }
    settings->setValue(name, sl);
}

void readQVectorFromIni(QSettings *settings, const QString &name, QVector<double> &value)
{
    if (!settings->contains(name)) return;
    value.clear();
    QStringList sl = settings->value(name).toStringList();
    Q_FOREACH(QString s, sl) {
        value.push_back(s.toDouble());
    }
    settings->setValue(name, sl);
}

void readQVectorFromIni(QSettings *settings, const QString &name, QVector<fwinf_t> &value)
{
    if (!settings->contains(name)) return;
    value.clear();
    QStringList sl = settings->value(name).toStringList();
    Q_FOREACH(QString s, sl) {
        value.push_back((fwinf_t)s.toInt());
    }
    settings->setValue(name, sl);
}

void readQVectorFromIni(QSettings *settings, const QString &name, QVector<intfir_t> &value)
{
    int size = settings->beginReadArray(name);
    if(size){
        value.clear();
        for(int i=0; i< size; ++i){
            settings->setArrayIndex(i);
            intfir_t f;
            f.readFromIni(settings, "unused");
            value.push_back(f);
        }
    }
    settings->endArray();
}

void AdcConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    settings->setValue("nch", nch);
    writeQVectorToIni(settings, "bits", bits);
    writeQVectorToIni(settings, "step", step);
    writeQVectorToIni(settings, "gain", gain);
    writeQVectorToIni(settings, "thr", thr);
    writeQVectorToIni(settings, "trig", trig);
    writeQVectorToIni(settings, "recwin", recwin);


    settings->setValue("prewin", prewin);
    settings->setValue("matchwin", matchwin);
    settings->setValue("latency", latency);
    settings->setValue("runmode", (uint)runmode);
    settings->setValue("trigbits", trigbits);
    settings->setValue("mult_min", mult_min);
    settings->setValue("mult_max", mult_max);
    settings->setValue("fragm_len_max", fragm_len_max);
    settings->setValue("fragm_count_max", fragm_count_max);
    settings->setValue("ch_dpm_ks", ch_dpm_ks);
    settings->setValue("timestamp_is_utc", timestamp_is_utc);

    settings->endGroup();
}

void AdcConfig::readFromIni(QSettings *settings, const QString &section)
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    readQVectorFromIni(settings, "bits", bits);
    readQVectorFromIni(settings, "step", step);
    readQVectorFromIni(settings, "gain", gain);
    readQVectorFromIni(settings, "thr", thr);
    readQVectorFromIni(settings, "trig", trig);
    readQVectorFromIni(settings, "recwin", recwin);

    setValueIfPresent(settings, "prewin", prewin);
    setValueIfPresent(settings, "matchwin", matchwin);
    setValueIfPresent(settings, "latency", latency);
    if(settings->value("runmode").toUInt() == ADCM_RUNMODE_BASECAL)
        runmode = ADCM_RUNMODE_BASECAL;
    else if(settings->value("runmode").toUInt() == ADCM_RUNMODE_RUN)
        runmode = ADCM_RUNMODE_RUN;
    setValueIfPresent(settings, "trigbits", trigbits);
    setValueIfPresent(settings, "mult_min", mult_min);
    setValueIfPresent(settings, "mult_max", mult_max);
    setValueIfPresent(settings, "fragm_len_max", fragm_len_max);
    setValueIfPresent(settings, "fragm_count_max", fragm_count_max);
    setValueIfPresent(settings, "ch_dpm_ks", ch_dpm_ks);
    setValueIfPresent(settings, "timestamp_is_utc", timestamp_is_utc);

    resize(nch); // enshure that all fields has at least nch values
    settings->endGroup();
}

void StorageConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    settings->setValue("enable", enable);
    settings->setValue("mode", mode);
    settings->setValue("period", period);
    settings->setValue("bn", bn);
    settings->setValue("fn", fn);
    settings->setValue("szlimit", szlimit);

    settings->endGroup();
}

void StorageConfig::readFromIni(QSettings *settings, const QString &section)
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    setValueIfPresent(settings, "enable", enable);
    setValueIfPresent(settings, "mode", mode);
    setValueIfPresent(settings, "period", period);
    setValueIfPresent(settings, "bn", bn);
    setValueIfPresent(settings, "fn", fn);
    setValueIfPresent(settings, "szlimit", szlimit);

    settings->endGroup();
}

void CalZeroConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    settings->setValue("initialized_nch", initialized_nch);
    writeQVectorToIni(settings,"nent", nent);
    writeQVectorToIni(settings,"base", base);
    writeQVectorToIni(settings,"rms", rms);
    writeQVectorToIni(settings,"asym", asym);

    settings->endGroup();
}

void CalZeroConfig::readFromIni(QSettings *settings, const QString &section)
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    setValueIfPresent(settings,"initialized_nch", initialized_nch);
    readQVectorFromIni(settings, "nent", nent);
    readQVectorFromIni(settings, "base", base);
    readQVectorFromIni(settings, "rms", rms);
    readQVectorFromIni(settings, "asym", asym);

    settings->endGroup();
}

void CalAvgConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    writeQVectorToIni(settings,"nent", nent);
    writeQVectorToIni(settings,"nf", nf);

    settings->beginWriteArray("data");
    for(int i=0; i<data.size(); ++i) {
        settings->setArrayIndex(i);
        int ch = data.keys()[i];
        settings->setValue("ch", ch);
        writeQVectorToIni(settings, "val", data[ch]);
    }
    settings->endArray();

    writeQVectorToIni(settings,"step", step);

    settings->endGroup();
}

void CalAvgConfig::readFromIni(QSettings *settings, const QString &section)
{
    CalAvgConfig newConf;
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    readQVectorFromIni(settings, "nent", newConf.nent);
    readQVectorFromIni(settings, "nf", newConf.nf);

    int size = settings->beginReadArray("data");
    for(int i=0; i<size; ++i){
        settings->setArrayIndex(i);
        int ch = settings->value("ch").toInt();
        readQVectorFromIni(settings, "val", newConf.data[ch]);
    }
    settings->endArray();

    readQVectorFromIni(settings, "step", newConf.step);

    settings->endGroup();
    if(newConf.isValid())
        *this = newConf;
    else
        qWarning() << "CalAvgConfig in ini-file is invalid";
}

//void writeQVectorToIni(QSettings settings, QString name, QVector<int> values)
//QVector<int> readQVectorFromIni(QSettings &settings, const QString &name, const int type)

void DecoderConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    settings->setValue("mode", mode);

    calzero.writeToIni(settings);
    calavg.writeToIni(settings);

    writeQVectorToIni(settings,"fwinf", fwinf);

    writeQVectorToIni(settings,"ppos", ppos);
    writeQVectorToIni(settings,"pwid", pwid);
    writeQVectorToIni(settings,"pnf", pnf);
    writeQVectorToIni(settings,"tadj", tadj);

    writeQVectorToIni(settings,"fir", fir);


    writeQVectorToIni(settings,"blr1", blr1);
    writeQVectorToIni(settings,"blr2", blr2);
    writeQVectorToIni(settings,"int1", int1);
    writeQVectorToIni(settings,"int2", int2);
    writeQVectorToIni(settings,"expfall", expfall);
    writeQVectorToIni(settings,"puthr", puthr);
    writeQVectorToIni(settings,"pumindist", pumindist);

    settings->setValue("fprof", fprof);
    settings->setValue("fititer", fititer);
    settings->setValue("fitside", fitside);
    settings->setValue("fitfixpar", fitfixpar);
    settings->setValue("fitalgo", fitalgo);

    writeQVectorToIni(settings,"baserc", baserc);

    settings->setValue("zero_thr", zero_thr);
    settings->setValue("dt_cut_en", dt_cut_en);
    settings->setValue("dt_cut_l", dt_cut_l);
    settings->setValue("dt_cut_r", dt_cut_r);

    settings->endGroup();
}

void DecoderConfig::readFromIni(QSettings *settings, const QString &section)
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    setValueIfPresent(settings, "mode", mode);

    calzero.readFromIni(settings);
    calavg.readFromIni(settings);

    readQVectorFromIni(settings,"fwinf", fwinf);

    readQVectorFromIni(settings,"ppos", ppos);
    readQVectorFromIni(settings,"pwid", pwid);
    readQVectorFromIni(settings,"pnf", pnf);
    readQVectorFromIni(settings,"tadj", tadj);

    readQVectorFromIni(settings,"fir", fir);

    readQVectorFromIni(settings,"blr1", blr1);
    readQVectorFromIni(settings,"blr2", blr2);
    readQVectorFromIni(settings,"int1", int1);
    readQVectorFromIni(settings,"int2", int2);
    readQVectorFromIni(settings,"expfall", expfall);
    readQVectorFromIni(settings,"puthr", puthr);
    readQVectorFromIni(settings,"pumindist", pumindist);

    setValueIfPresent(settings, "fprof", fprof);
    setValueIfPresent(settings, "fititer", fititer);
    setValueIfPresent(settings, "fitside", fitside);
    setValueIfPresent(settings, "fitfixpar", fitfixpar);
    setValueIfPresent(settings, "fitalgo", fitalgo);

    readQVectorFromIni(settings, "baserc", baserc);

    setValueIfPresent(settings, "zero_thr", zero_thr);
    setValueIfPresent(settings, "dt_cut_en", dt_cut_en);
    setValueIfPresent(settings, "dt_cut_l", dt_cut_l);
    setValueIfPresent(settings, "dt_cut_r", dt_cut_r);

    settings->endGroup();
}

void HistConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    settings->setValue("enable", enable);

    settings->endGroup();
}

void HistConfig::readFromIni(QSettings *settings, const QString &section)
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    setValueIfPresent(settings, "enable", enable);

    settings->endGroup();
}

void RunConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    settings->setValue("stopCondition", stopCondition);
    settings->setValue("stopByTimer", stopByTimer);
    settings->setValue("stopByEventCount", stopByEventCount);
    settings->setValue("stopTimerValue", stopTimerValue);
    settings->setValue("stopEventCountValue", stopEventCountValue);

    settings->endGroup();
}

void RunConfig::readFromIni(QSettings *settings, const QString &section)
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    setValueIfPresent(settings, "stopCondition", stopCondition);
    setValueIfPresent(settings, "stopByTimer", stopByTimer);
    setValueIfPresent(settings, "stopByEventCount", stopByEventCount);
    setValueIfPresent(settings, "stopTimerValue", stopTimerValue);
    setValueIfPresent(settings, "stopEventCountValue", stopEventCountValue);

    settings->endGroup();
}

void AllConfig::writeToIni(QSettings *settings, const QString &section) const
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    settings->setValue("version", version);

    adc.writeToIni(settings);
    stor_raw.writeToIni(settings, "StorageConfig_Raw");
    stor_dev.writeToIni(settings, "StorageConfig_Dev");
    dec.writeToIni(settings);
    hist.writeToIni(settings);
    run.writeToIni(settings);

    settings->endGroup();
}

void AllConfig::readFromIni(QSettings *settings, const QString &section)
{
    settings->beginGroup(!section.isEmpty() ? section : sectionName());

    setValueIfPresent(settings, "version", version);

    adc.readFromIni(settings);
    stor_raw.readFromIni(settings, "StorageConfig_Raw");
    stor_dev.readFromIni(settings, "StorageConfig_Dev");
    dec.readFromIni(settings);
    hist.readFromIni(settings);
    run.readFromIni(settings);

    settings->endGroup();
}

void intfir_t::writeToIni(QSettings *settings, const QString &name) const
{
    settings->setValue("klen", klen);
    settings->setValue("offs", offs);
    settings->setValue("bits", bits);
    writeQVectorToIni(settings, "coef", coef);
}

void intfir_t::readFromIni(QSettings *settings, const QString &name)
{
    setValueIfPresent(settings, "klen", klen);
    setValueIfPresent(settings, "offs", offs);
    setValueIfPresent(settings, "bits", bits);
    readQVectorFromIni(settings, "coef", coef);
}
