//
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "decoderconfig.h"

#include <cstdio>
#include <cmath>
#include <QDebug>

#include "adcm16.h"
#include "dsplib/fir.h"
#include "dsplib/fitter.h"
#include "dsplib/dsplib.h"

DecoderConfig::DecoderConfig()
{
    initialized_nch = ADC_NCH;
    fwinf.resize(nch());
    ppos.resize(nch());
    pwid.resize(nch());
    pnf.resize(nch());
    tadj.resize(nch());
    fir.resize(nch());
    blr1.resize(nch());
    blr2.resize(nch());
    int1.resize(nch());
    int2.resize(nch());
    expfall.resize(nch());
    puthr.resize(nch());
    pumindist.resize(nch());
    baserc.resize(nch());

    mode = 0;
    fprof = FIT_PROFILE_DEFAULT;
    fititer = 2; // even one is enough. 2 for safe.
    fitside = 0; // 0 - auto
    fitfixpar = 0;       // bit 0: amplitude, bit1: time, bit2: width. Default: no fixed par
    fitalgo = FIT_ALGO_ACCURATE;
    zero_thr = 10;       // mV
    dt_cut_en = 0;
    dt_cut_l = -300;
    dt_cut_r = 300;
    for (int c = 0; c < nch(); c++) {
        fwinf[c] = WF_TUKEY_050;
        pnf[c] = fft_noisefloor (1. / sqrt (12.), ADCM_MAX_N_SAMP);
        expfall[c] = 10000;
        puthr[c] = 75;     // %
        pumindist[c] = 100; // ns
        pwid[c] = 30 /*2.5*adc.bits[c] */ ;
        //    ppos[c] = PREWIN*adc.bits[c]+1.5*pwid[c];
        ppos[c] = 0 /*1.5*pwid[c] */ ;
        blr1[c] = -10;
        blr2[c] = -3;
        int1[c] = -2;
        int2[c] = 10;//adc.recwin[c]-adc.latency-(adc.matchwin-adc.prewin);
        tadj[c] = 0;
        baserc[c] = 5;
        //    search_start[c] = PREWIN-12;
        //    search_end[c] = PREWIN+32;
        for (unsigned int i = 0; i < FIRMAXLEN; i++)
            fir[c].coef[i] = 0;
        /*
// a bypass filter
          fir[c].klen=1;
          fir[c].offs=0;
          fir[c].bits=1;
          fir[c].coef[0]=1;
*/
        // a useful filter
        fir[c].klen = 12;
        fir[c].offs = 5;
        fir[c].bits = 10;
        fir[c].coef[0] = 11;
        fir[c].coef[1] = 37;
        fir[c].coef[2] = 183;
        fir[c].coef[3] = 449;
        fir[c].coef[4] = 511;
        fir[c].coef[5] = 133;
        fir[c].coef[6] = -349;
        fir[c].coef[7] = -455;
        fir[c].coef[8] = -235;
        fir[c].coef[9] = -57;
        fir[c].coef[10] = -20;
        fir[c].coef[11] = -6;
    }
}

bool DecoderConfig::operator==(const DecoderConfig &other) const
{
    if(DEBUG_PRINT_CONFIG_DIFFERENCE){
        if(mode != other.mode)
            qDebug() << "mode differs";
        if(calzero != other.calzero)
            qDebug() << "calzero differs";
        if(calavg != other.calavg)
            qDebug() << "calavg differs";
        if(fwinf != other.fwinf)
            qDebug() << "fwinf differs";
        if(ppos != other.ppos)
            qDebug() << "ppos differs";
        if(pwid != other.pwid)
            qDebug() << "pwid differs";
        if(pnf != other.pnf)
            qDebug() << "pnf differs";
        if(tadj != other.tadj)
            qDebug() << "tadj differs";
        if(fir != other.fir)
            qDebug() << "fir differs";
        if(blr1 != other.blr1)
            qDebug() << "blr1 differs";
        if(blr2 != other.blr2)
            qDebug() << "blr2 differs";
        if(int1 != other.int1)
            qDebug() << "int1 differs";
        if(int2 != other.int2)
            qDebug() << "int2 differs";
        if(expfall != other.expfall)
            qDebug() << "expfall differs";
        if(puthr != other.puthr)
            qDebug() << "puthr differs";
        if(pumindist != other.pumindist)
            qDebug() << "pumindist differs";
        if(fprof != other.fprof)
            qDebug() << "fprof differs";
        if(fititer != other.fititer)
            qDebug() << "fititer differs";
        if(fitside != other.fitside)
            qDebug() << "fitside differs";
        if(fitfixpar != other.fitfixpar)
            qDebug() << "fitfixpar differs";
        if(fitalgo != other.fitalgo)
            qDebug() << "fitalgo differs";
        if(baserc != other.baserc)
            qDebug() << "baserc differs";
        if(zero_thr != other.zero_thr)
            qDebug() << "zero_thr differs";
        if(dt_cut_en != other.dt_cut_en)
            qDebug() << "dt_cut_en differs";
        if(dt_cut_l != other.dt_cut_l)
            qDebug() << "differs differs";
        if(dt_cut_r != other.dt_cut_r)
            qDebug() << "dt_cut_r differs";
    }
    return true
            && mode == other.mode
            && calzero == other.calzero
            && calavg == other.calavg
            && fwinf == other.fwinf
            && ppos == other.ppos
            && pwid == other.pwid
            && pnf == other.pnf
            && tadj == other.tadj
            && fir == other.fir
            && blr1 == other.blr1
            && blr2 == other.blr2
            && int1 == other.int1
            && int2 == other.int2
            && expfall == other.expfall
            && puthr == other.puthr
            && pumindist == other.pumindist
            && fprof == other.fprof
            && fititer == other.fititer
            && fitside == other.fitside
            && fitfixpar == other.fitfixpar
            && fitalgo == other.fitalgo
            && baserc == other.baserc
            && zero_thr == other.zero_thr
            && dt_cut_en == other.dt_cut_en
            && dt_cut_l == other.dt_cut_l
            && dt_cut_r == other.dt_cut_r
            ;
}

int DecoderConfig::nch() const
{
    return initialized_nch;
}

void DecoderConfig::print() const
{
    fprintf (stderr, "dec_mode: %d\n", mode);
    fprintf (stderr, "dec_fprof: %d\n", fprof);
    fprintf (stderr, "dec_fititer: %d\n", fititer);
    fprintf (stderr, "dec_fitside: %d%s\n", fitside, fitside?"":" (auto)");
    fprintf (stderr, "dec_fitfixpar: %s%s%s\n",
             (fitfixpar & 1) ? "a " : "", (fitfixpar & 2) ? "t " : "", (fitfixpar & 4) ? "w " : "");
    fprintf (stderr, "dec_fitalgo: %d\n", fitalgo);

    fprintf (stderr, "dec_fwinf: ");
    for (int c = 0; c < fwinf.size(); c++)
        fprintf (stderr, "%u ", fwinf[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_ppos: ");
    for (int c = 0; c < ppos.size(); c++)
        fprintf (stderr, "%g ", ppos[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_pwid: ");
    for (int c = 0; c < pwid.size(); c++)
        fprintf (stderr, "%g ", pwid[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_pnf: ");
    for (int c = 0; c < pnf.size(); c++)
        fprintf (stderr, "%g ", pnf[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_tadj: ");
    for (int c = 0; c < tadj.size(); c++)
        fprintf (stderr, "%g ", tadj[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_blr: ");
    for (int c = 0; c < std::min(blr1.size(), blr2.size()); c++)
        fprintf (stderr, "%d..%d ", blr1[c], blr2[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_int: ");
    for (int c = 0; c < std::min(int1.size(), int2.size()); c++)
        fprintf (stderr, "%d..%d ", int1[c], int2[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_expfall: ");
    for (int c = 0; c < expfall.size(); c++)
        fprintf (stderr, "%d ", expfall[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_puthr: ");
    for (int c = 0; c < puthr.size(); c++)
        fprintf (stderr, "%d ", puthr[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_pumindist: ");
    for (int c = 0; c < pumindist.size(); c++)
        fprintf (stderr, "%d ", pumindist[c]);
    fprintf (stderr, "\n");

    fprintf (stderr, "dec_baserc: ");
    for (int c = 0; c < baserc.size(); c++)
        fprintf (stderr, "%g ", baserc[c]);
    fprintf (stderr, "\n");

    for (int c = 0; c < fir.size(); c++) {
        fprintf (stderr, "fir[%u]: ", c);
        print_fir (&fir[c]);
    }
    //  print_calzero (&calzero);
    calavg.print();
}

int DecoderConfig::get_dec_fititer() const
{
    const unsigned int  default_value = 2;
    unsigned int fiter;
    switch (fprof) {
    case FIT_PROFILE_SUPERFAST:
        fiter = 1;
        break;
    case FIT_PROFILE_FAST:
        fiter = 1;
        break;
    case FIT_PROFILE_OPTIMAL:
        fiter = 2;
        break;
    case FIT_PROFILE_PRECISION:
        fiter = 26;
        break;
    case FIT_PROFILE_CUSTOM:
        fiter = fititer;
        break;
    default:
        fiter = default_value;
        break;
    }

    if (fiter > 1000)
        fiter = default_value;
    return fiter;
}

/*fit_algo_t*/
int DecoderConfig::get_dec_fitalgo() const
{
    int default_value = FIT_ALGO_ACCURATE;

    int algo = default_value;
    if (fprof == FIT_PROFILE_CUSTOM) algo = fitalgo;

    return algo;
}

int DecoderConfig::get_dec_fitfixpar() const
{
    int default_value = 0;

    int fixpar = default_value;
    if (fprof == FIT_PROFILE_CUSTOM) fixpar = fitfixpar;

    return fixpar;
}
