//
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "waveformdecoderpeaks.h"

#include <cmath>

#include "dsplib/dsplib.h"
#include "dsplib/fitter.h"
#include "adcm_ws_t.h"
#include "config/allconfig.h"
#include "adcmraw/adcm_mem.h"
#include "dsplib/fir.h"
#include "adcmlib/adcm_dws.h"
#include "adcmlib/adcm_ev.h"
#include "adcmlib/adcm_dec.h"
#include "decoder_profiling_macros.h"
#include "qlog/qlog.h"

WaveformDecoderPeaks::WaveformDecoderPeaks ()
    : WaveformDecoder (),
      fitterGaus(new FitterGaus(MAX_FITGAUS_POINTS))
{
}

WaveformDecoderPeaks::~WaveformDecoderPeaks ()
{
    delete fitterGaus;
}

int WaveformDecoderPeaks::convolve_ch(AdcmEvent * ev, adcm_decstat_t * decst, int c)
{
    Q_ASSERT(ev);

    int fside;
    int SEARCH_START, SEARCH_END;
    int search_extend = 4;

    int cin1, cin2, cblen;
    int i1, i2;
    int inp_len;
    wbc_t *wb = &ws->wb[c];
    cb_t *cb = &ws->cb[c];

    if (wb->len < 3)
        return(0);

    fside = AllConfig::get_dec_fitside(*conf_mem, *dec, c);
    // First, convolver is run over signal

    SEARCH_START = adc->latency - search_extend;
    SEARCH_END = SEARCH_START + 2 * search_extend;
    int trig = -1;
    if (adc->size() > c)
        trig = adc->trig[c];
    if (trig & ADCM_CCTRL_BIT_S) {
        SEARCH_START -= adc->prewin;
        SEARCH_END += adc->matchwin;
    }

    if (SEARCH_START < 1)
        SEARCH_START = 1;
    if (SEARCH_END >= wb->len)
        SEARCH_END = wb->len - 1;

    //  printf("peak search: %d-%d\n", SEARCH_START, SEARCH_END);

    int firlen=0, fir_dt=0;
    if (dec->fir.size() > c) {
        // find FIR delay
        fir_dt = dec->fir[c].offs;
        firlen = dec->fir[c].klen;
    }

    // find interesting convolution region

    // Left margin = 0. Uncomment. FIXME.
    //    cin1 = 0;
    cin1 = SEARCH_START - firlen;
    if (cin1 < 0)
        cin1 = 0;
    cin2 = SEARCH_END + firlen;
    if (cin2 > int(CONVLEN))
        cin2 = CONVLEN;

    // do convolution
    //  LOG("ch %d, FIR: %d taps\n", c, dec->fir[c].klen);
    inp_len = cin2-cin1;
    if (inp_len > 0 && firlen > 0) {
        fir_convolve_s (&dec->fir[c], &wb->d[cin1], inp_len, cb->d, CONVLEN);
        cblen = inp_len + firlen - 1;
    } else {
        for (int i = 0; i < inp_len; i++) { // direct copy
            //        ws->cb[c].len = ws->wb[c].len;
            ws->cb[c].len = inp_len;
            if (i < wb->len)
                cb->d[i] = wb->d[i+cin1];
            else
                cb->d[i] = 0;
        }
        cblen = inp_len;
    }
    cb->len = cblen;
    //printf("fir_dt=%d,  cin1=%d\n", fir_dt, cin1);
    cb->offs = fir_dt - cin1;
    //    cb->offs = cin1;

    // peak search range i1<=t<i2
    //    int icentr = fir_dt;
    i1 = SEARCH_START + cb->offs;
    i2 = SEARCH_END + cb->offs;
    if (i1 < fside)
        i1 = fside;
    if (i2 > cblen - fside)
        i2 = cblen - fside;

    cb->i1 = i1;
    cb->i2 = i2;

    return 0;
}

int WaveformDecoderPeaks::make_convolve(AdcmEvent * ev, adcm_decstat_t * decst)
{
    Q_ASSERT(ev);

    for (unsigned int c = 0; c < ADC_NCH; c++) {

        if (ws->wb[c].len < 3)
            continue;

        convolve_ch(ev, decst, c);
        //    print_cb(adc_latency (cfg), &ws->cb[c]);
    }
    return (0);
}

// find largest peak with 3x moving average filter
// in range i1+1 <= i < i2+1
// return peak position ip, alplitude ap
int WaveformDecoderPeaks::find_peak (AdcmEvent * ev, adcm_decstat_t *decst, int c)
{
    Q_ASSERT(ev);

    int ip, ap;
    cb_t *cb = &ws->cb[c];
    const int *d = &cb->d[0];
    int i1 = cb->i1+1;
    int i2 = cb->i2-1;

    if (i2 - i1 < 1) {
        if (decst)
            decst->dsp_pk_range++;
        print_decoder_dsp_err (decst, "ch %d, peak search range bad: %d - %d", c, i1, i2);
        adcm_ev_ch_flag_setbit(ev, c, ADCM_CH_PKS_NODATA);
        //      ev->ch[c].flags |= ADCM_CH_PKS_NODATA;
        //      de->flags |= F_TERR;
        return(0);
    }

    //  print_decoder_dsp_err (decst, "ch %d, peak not found in %d - %d", c, i1, i2);

    ip = i1;                      // default
    /*
  ap = cb_peak_rule (d, ip);
  for (int i = i1; i < i2; i++) {
    if (cb_peak_rule (d, i) > ap) {
      ip = i;
      ap = cb_peak_rule (d, i);
    }
  }
*/
    // Fast moving average filter
    {
        int tmp;
        tmp = d[i1-1]+d[i1]+d[i1+1];
        ap = tmp;
        for (int i = i1; i < i2; i++) {
            if (tmp>ap) {
                ip=i;
                ap=tmp;
            }
            tmp += -d[i-1]+d[i+2];
        }
    }
    cb->ip = ip;
    cb->ap = ap;
    //  printf ("ch %d, peak found in %d - %d at %d\n", c, i1, i2, ip);
    return (0);
}

int WaveformDecoderPeaks::find_pileup (AdcmEvent * ev, adcm_decstat_t * decst)
{
    Q_ASSERT(ev);

    for (unsigned int c = 0; c < ADC_NCH; c++)
    {
        int pu_code;
        if (ws->wb[c].len < 3)
            continue;
        pu_code = cb_find_pileup (c, &ws->cb[c], decst);
        if (pu_code == 1)
            adcm_ev_ch_flag_setbit(ev, c, ADCM_CH_HAS_PU);
    }
    return 0;
}

int WaveformDecoderPeaks::find_fasttime (int method, AdcmEvent * ev, adcm_decstat_t * decst)
{
    Q_ASSERT(ev);

    const int lat = adc->latency;

    for (unsigned int c = 0; c < ADC_NCH; c++)
    {
        adcm_chbuf_t *evc = ev->get_ev_chbuf(c);
        if (!evc) continue;

        if ( ! (evc->flags & ADCM_CH_HAS_THR))
            continue;

        if (ws->wb[c].len < 3)
            continue;

        if ( ! (evc->flags & ADCM_CH_EVAL_TFAST)) {
            double t=0, a=0, w=0;
            //    double chisq=0;
            cb_t *cb = &ws->cb[c];
            fb_t *fb = &ws->fb[c];
            int fb_offset;
            int fb_len;

            //    find_peak (cfg, cb);

            //    print_cb(lat, &ws->cb[c]);


            // Fill fitbuffer - fb is durty in enter.
            {
                const int fside = AllConfig::get_dec_fitside(*conf_mem, *dec, c);
                int fi1, fi2, flen;
                fi1 = cb->ip - fside;
                fi2 = cb->ip + fside + 1;
                flen = fi2-fi1;
                if (flen > MAX_FITGAUS_POINTS)
                    flen = MAX_FITGAUS_POINTS;
                for(int i=0;i<flen;i++)
                    fb->d[i] = cb->d[fi1+i];
                fb->offs = -fi1;
                fb->len = flen;
                fb_offset = fi1;
                fb_len = flen;
            }

            //    print_fb(lat + cb->offs, fb);

            {
                double par[3];
                // initial conditions
                //    par[1] = (fside + fside) / 2;
                //    par[2] = DEFAULT_SIGMA;

                //      FindGausApproxD (fb->d, fb->len, par);
                FindGausApprox (&cb->d[fb_offset], fb_len, par);
                t = par[1] - lat - fb->offs - cb->offs;
                w = par[2];
                a = par[0];
                evc->pchi2 = evc->fastnorm = FindGausDevNorm (&cb->d[fb_offset], fb_len, par) / a / fb_len;
                par[2] = dec->pwid[c]/GausFWHM;
            }

            evc->flags |= ADCM_CH_EVAL_TFAST;
            evc->pa = a;
            evc->fasttime = t;
            evc->pt = t;
            evc->pw = w*GausFWHM;
            //    evc->pchi2 = chisq;
            //    printf("t=%f, ip=%d\n", t, cb->ip);
        } // if ADCM_CH_EVAL_TFAST

    }

    return (0);
}

int WaveformDecoderPeaks::find_finetime (int method, AdcmEvent * ev, adcm_decstat_t * decst)
{
    Q_ASSERT(ev);

    const int lat = adc->latency;

    // Measure fine time

    for (unsigned int c = 0; c < ADC_NCH; c++)
    {
        adcm_chbuf_t *evc = ev->get_ev_chbuf(c);
        if (!evc) continue;

        if ( ! (evc->flags & ADCM_CH_EVAL_TFAST)) continue;
        if ((evc->flags & ADCM_CH_DT_CUT)) continue;

        {
            int tok = 1;
            int tpk;
            double t=0, a=0, w=0, chi2=0, niter=0;
            tpk = evc->pt;
            cb_t *cb = &ws->cb[c];
            fb_t *fb = &ws->fb[c];

            //      if (method == FIT_ALGO_ACCURATE)
            {
                int ret = 0;
                double par[3]={0,0,0};
                fitstatus_t fitstatus;
                memset(&fitstatus, 0, sizeof(fitstatus));
                fitstatus.niter = dec->get_dec_fititer();

                ret = fitterGaus->fit (dec->get_dec_fitfixpar(), fb->d, fb->len, par, &fitstatus);

                if (ret != 0) {
                    if (decst)
                        decst->dsp_fit_fail++;
                    print_decoder_dsp_err (decst, "Ev %d; ch %d: fit failed, code=%d", ev->getEventNumber(), c, ret);
                    evc->flags |= ADCM_CH_TFIT_ERR;
                    tok = 0;
                }
                t = par[1] - lat - cb->offs - fb->offs;
                w = par[2];
                a = par[0];
                chi2 = fitstatus.fnorm / a / fb->len;
                niter = fitstatus.niter;
                //    	  printf("par %f,  tpk %f  t %f,  cb.offs %d,  fb.offs %d\n", par[1], tpk, t, cb->offs, fb->offs);
            }

            //    log_debug (3, " * ch=%d, ip=%d, a=%g, t=%.2f, w=%.2f\n", c, ip, a, t, w);
            if ((t < tpk - 3) || (t > tpk + 3)) {
                if (DEC_DEBUG >= 1) {
                    if (decst)
                        decst->dsp_fit_bad++;
                    print_decoder_dsp_err (decst, "Ev %d; ch %d: fit is bad, tpk=%d, t=%.2f", ev->getEventNumber(), c, tpk, t);
                }
                evc->flags |= ADCM_CH_TFIT_OFF;
                tok = 0;
            }
            if (t == 0) {
                tok = 0;
                if (decst)
                    decst->dsp_fit_bad++;
                print_decoder_dsp_err (decst, "Ev %d; ch %d: fit is nul, t=%.2f", ev->getEventNumber(), c, t);
            }
            if (tok) {
                if (decst)
                    decst->dsp_fit_ok++;
                //      log_debug(3, " + ch %d: t=%.2f", c, t);
            } else {
                evc->flags |= F_TERR;
            }
            evc->pa = a;
            evc->pt = t;
            evc->pw = w*GausFWHM;
            // debug data
            //printf("pp=%d, fasttime=%f, t=%f\n", ev->ch[c].pp, ev->ch[c].fasttime, evc->pt);
            evc->pchi2 = chi2;
            evc->pniter = niter;
            if (tok) evc->flags |= ADCM_CH_EVAL_TFINE;

            /*
// DEBUG DUMP
        printf("ev %6d, ch %2d, thr at %3d -- %s\n", ev->ec, c, evc->tp, tok?"ok":"BAD");
        if (!tok)
        if (adc->trig[c] & ADCM_CCTRL_BIT_S)
        {
          printf("ev %6d, ch %2d, thr at %3d\n", ev->ec, c, evc->tp);
//      print_wb(ev->ch[c].tp, &ws->wb[c]);
      print_cb(lat, cb);
      print_fb(lat + cb->offs, fb);
        }
*/
        }
    }

    return (0);
}

int WaveformDecoderPeaks::find_tfast (AdcmEvent * ev, adcm_decstat_t * decst)
{
    return (find_fasttime (FIT_ALGO_COARSE, ev, decst));
}

int WaveformDecoderPeaks::find_tfine (AdcmEvent * ev, adcm_decstat_t * decst)
{
    return (find_finetime (dec->get_dec_fitalgo(), ev, decst));
}

void WaveformDecoderPeaks::decode_work (AdcmEvent * ev, adcm_decstat_t * decst,
                                        decst_t_t &pt, ftimer_t &t)
{
    WaveformDecoder::decode_work(ev, decst, pt, t);

    if (stages & (ADCM_CH_EVAL_TFAST | ADCM_CH_EVAL_TFINE)) {

        make_convolve(ev, decst);
        PROF_TMR_MARK(pt, t, decst_n_conv);

        peaksearch (ev, decst);
        PROF_TMR_MARK(pt, t, decst_n_pksrch);

        find_pileup (ev, decst);
        PROF_TMR_MARK(pt, t, decst_n_pu);

        find_tfast (ev, decst);
        PROF_TMR_MARK(pt, t, decst_n_tfast);

        bool pulseFound = find_dt_cut_condition(ev, decst); // TODO
        PROF_TMR_MARK(pt, t, decst_n_dtcut);
        if (!pulseFound) { // no pulses in event - drop event
            ev->setFlag(AdcmEvent::ADCM_EV_DT_CUT);
            if (decst)
                decst->ev_dt_cut++;
        } else {
            if ((stages & ADCM_CH_EVAL_TFINE)) {
                find_tfine (ev, decst);
                PROF_TMR_MARK(pt, t, decst_n_fit);
            }
        }
    }
}


int WaveformDecoderPeaks::simulate(AdcmEvent * ev)
{
    Q_ASSERT(ev);

    //Q_ASSERT(0); // uncomment next code

    double a, t, rms, base;

    for (unsigned int c = 0; c < ADC_NCH; c++) {
        adcm_chbuf_t *ch = ev->get_ev_chbuf(c);
        if (!ch) continue;
        ch->flags &= ~ADCM_CH_HAS_DATA;
        if (!(adc->trig[c] & (ADCM_CCTRL_BIT_M | ADCM_CCTRL_BIT_S)))
            continue;
        //    if (c > 1) break;
//        base = dec->calzero.get(c);
        base = 0;
        ch->w_bits = adc->bits[c];
        a = ch->pa;
        rms = ch->rms;
        unsigned int len = adc->recwin[c];
        if (len > ADCM_MAX_N_SAMP) len = ADCM_MAX_N_SAMP;
        ch->w.resize(len);
#ifdef FIT_SINE
        t = ch->pt;
        double f = ch->pw;
        for (unsigned int i = 0; i < len; i++)
            ch->w[i] = (int) (0.5 + rangaus (rms) + base + a * sin(t + f * i));
#else
        t = ch->pt + adc->latency;
        const double rt = ch->pw;
        const double ft = rt * 10;
        for (unsigned int i = 0; i < len; i++)
            ch->w[i] = (int) (0.5 + rangaus (rms) + base + detpulse (a, i - t, rt / GausFWHM, ft));
#endif
        ch->flags = ADCM_CH_HAS_DATA | ADCM_CH_FINISHED;
    }

    /*
{
  FILE *f;
  f = fopen("sim_ev.txt", "w+");
  if (f) {
  for(int c=0;c<ADC_NCH;c++)
        for(int i=0;i<ev->ch[c].len;i++)
          fprintf(f, "%d\t%d\t%d\n", c, i, ev->ch[c].w[i]);
  fclose(f);
  }
}
*/
    return (0);
}

int WaveformDecoderPeaks::peaksearch (AdcmEvent * ev, adcm_decstat_t * decst)
{
    Q_ASSERT(ev);

    for (unsigned int c = 0; c < ADC_NCH; c++)
    {
        if (ws->wb[c].len < 3)
            continue;
        find_peak (ev, decst, c);
    }
    return 0;
}

// return true if any useful pulses found
bool WaveformDecoderPeaks::find_dt_cut_condition(AdcmEvent *ev, adcm_decstat_t *decst)
{
    Q_ASSERT(ev);

    bool pulseFound = false;

    if (!dec->dt_cut_en)
        return true;

    const unsigned int nch = conf_mem->nch;
    Q_ASSERT(nch <= ADC_NCH);
    double dtMin=dec->dt_cut_l;
    double dtMax=dec->dt_cut_r;
    //printf("dt cut: l=%f, r=%f\n", l, r);
    // clear 'match' flags
    bool chUsed[ADC_NCH];
    for(unsigned int i=0; i<nch; i++)
        chUsed[i]=false;

    for(unsigned int c=0; c<nch; c++) {
        adcm_chbuf_t *cb = ev->get_ev_chbuf(c);
        if (!cb) continue;
        cb->flags &= ~ADCM_CH_DT_CUT;
    }

    for(unsigned int mCh=0; mCh<nch; mCh++) {
        //	int mc_match = 0;
        if (!(adc->trig[mCh] & ADCM_CCTRL_BIT_M)) continue;
        adcm_chbuf_t *mcb = ev->get_ev_chbuf(mCh);
        if (!mcb) continue;
        if (!(mcb->flags & ADCM_CH_EVAL_TFAST)) continue;

        for(unsigned int sCh=0; sCh<nch; sCh++) {
            //  	  int sc_match = 0;
            if (!(adc->trig[sCh] & ADCM_CCTRL_BIT_S)) continue;
            adcm_chbuf_t *scb = ev->get_ev_chbuf(sCh);
            if (!scb) continue;
            if (!(scb->flags & ADCM_CH_EVAL_TFAST)) continue;

            double dt =  scb->pt*adc->step[sCh] - mcb->pt*adc->step[mCh];
            int cond = dtMin <= dt && dt <= dtMax;
            //		printf("m:%2d, s:%2d  dt=%f  -- %s\n", mc, sc, dt, cond?"MATCH":"discard");
            if (cond) {
                pulseFound = true;
                chUsed[mCh] = true;
                chUsed[sCh] = true;
            }
        }
    }
    for (unsigned int c=0; c<nch; c++) {
        adcm_chbuf_t *cb = ev->get_ev_chbuf(c);
        if (!cb) continue;
        if ((cb->flags & ADCM_CH_EVAL_TFAST) && !chUsed[c])
            cb->flags |= ADCM_CH_DT_CUT;
    }

    //  for(i=0;i<nch;i++)
    //	printf("ch_use[%2d]: %d\n", i, ch_use[i]);
    //  for(i=0;i<nch;i++)
    //	if (ev->ch[i].flags & ADCM_CH_DT_CUT)
    //	  printf("[%2d]: discarded by dT cut\n", i);
    return pulseFound;
}
