
#include <fcntl.h> // FIXME: O_RDONLY, open()

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <vector>

#include "fir.h"
#include "dsplib/dsplib.h"
#include "qlog/qlog.h"

const int FIR_COEFF_BITS = 10;
const int FIR_OPT_MIN_VAL = 5;
const double FIR_MAX_AMPLIF = 10.;

void print_fir (const intfir_t * fir)
{
    assert(fir);
    printf ("FIR: %d bits, %d taps, offset %d: {\n", fir->bits, fir->klen, fir->offs);
    for (unsigned int i = 0; i < fir->klen; i++) {
        printf (" %d", fir->coef[i]);
        if ((i + 1) < fir->klen)
            printf (",");
    }
    printf ("\n}\n");

}

void print_dblfir (const dblfir_t * fir)
{
    assert(fir);
    printf ("FIR: %d taps, offset %d: {\n", fir->klen, fir->offs);
    for (unsigned int i = 0; i < fir->klen; i++) {
        printf (" %e", fir->coef[i]);
        if ((i + 1) < fir->klen)
            printf (",");
        if (i%5==4) printf("\n");
    }
    printf ("\n}\n");

}

int fir_valid (const intfir_t * fir)
{

    if (!fir)
        return 1;
    if (fir->klen > FIRMAXLEN) {
        fprintf (stderr, "Invalid FIR klen=%d\n", fir->klen);
        return (0);
    }
    if (fir->offs > fir->klen) {
        fprintf (stderr, "Invalid FIR offs=%d (klen=%d\n", fir->offs, fir->klen);
        return (0);
    }
    if ((fir->bits < 1) || (fir->bits > 32)) {
        fprintf (stderr, "Invalid FIR bits=%d\n", fir->bits);
        return (0);
    }
    return (1);
}

int readfir (const char *fn, intfir_t * fir)
{
    // FIXME: use QFile
 /*
    int fd, sig;

    assert(fir);

    if ((fd = open (fn, O_RDONLY)) == -1) {
        perror ("open FIR file failed");
        return (-1);
    }
    if (read (fd, &sig, sizeof (int)) == -1) {
        goto err;
    }
    if (sig != FIRFILE_MAGIC) {
        fprintf (stderr, "bad FIR file magic\n");
        goto err;
    }
    if (read (fd, &fir->klen, sizeof (int)) == -1) {
        goto err;
    }
    if (fir->klen > FIRMAXLEN) {
        fprintf (stderr, "bad FIR file: klen invalid\n");
        goto err;
    }
    if (read (fd, &fir->offs, sizeof (int)) == -1) {
        goto err;
    }
    if (fir->offs > fir->klen) {
        fprintf (stderr, "bad FIR file: offs invalid\n");
        goto err;
    }
    if (read (fd, &fir->bits, sizeof (int)) == -1) {
        goto err;
    }
    if (fir->bits > 32) {
        fprintf (stderr, "bad FIR file: bits invalid\n");
        goto err;
    }
    if (read (fd, fir->coef, fir->klen * sizeof (int)) == -1) {
        goto err;
    }
    close (fd);

    printf ("Read FIR file '%s': klen=%d, offs=%d, bits=%d\n", fn, fir->klen, fir->offs, fir->bits);
    printf ("coef = {");
    for (unsigned int i = 0; i < fir->klen; i++)
        printf (" %d,", fir->coef[i]);
    printf ("}\n");

    return 0;
    err:
        close(fd);
    perror ("read FIR file failed");
*/
    return (-1);
}

int writefir (const char *fn, const intfir_t * fir)
{
    // FIXME: use QFile
/*
    const int sig = FIRFILE_MAGIC;

    FILE *fp = fopen (fn, "w");
    if (!fp) {
        perror ("open FIR file failed");
        return (-1);
    }

    fwrite (&sig, 1, sizeof (int), fp);
    fwrite (&fir->klen, 1, sizeof (int), fp);
    fwrite (&fir->offs, 1, sizeof (int), fp);
    fwrite (&fir->bits, 1, sizeof (int), fp);
    fwrite (fir->coef, fir->klen, sizeof (int), fp);
    if (fclose (fp) == EOF) {
        perror ("close FIR file failed");
        return (-1);
    }
*/
    return 0;
}

/*
void deletefir(intfir_t *fir) {
  if (fir) {
    delete(fir->coef);
    fir->coef=NULL;
  }
}
*/

int create_deconvolution_dblFIR (dblfir_t * engine, const std::vector<double> &dsrc, const std::vector<double> &dreq, double orig_thr, double req_thr)
{
    // Return codes:
    //  -1 - fatal error
    //   0 - success
    //   1,2,... - warning, bad result may occur

    int ret = 0;
    assert(engine);

    const size_t minCommonSize = std::min(dsrc.size(), dreq.size());
    if (minCommonSize > FIRMAXLEN) {
        logger.err (QString("input pulse too large: %1 truncated to %2").arg(minCommonSize).arg(FIRMAXLEN));
        ret++;
    }
    const size_t inlen = std::min(size_t(FIRMAXLEN), minCommonSize);
    assert(inlen >= 2);


    if (!engine || (inlen > 10000)) {
        logger.err("Bad input");
        return (-1);
    }


    const unsigned int deconvkernlen = inlen;
    const unsigned int deconvkernhalflen = deconvkernlen / 2 + 1;

    double *dout = new double[deconvkernlen];
    double *powsrc = new double[deconvkernhalflen];
    double *powreq = new double[deconvkernhalflen];
    double *powout = new double[deconvkernhalflen];
    double *sinsrc = new double[deconvkernhalflen];
    double *cossrc = new double[deconvkernhalflen];
    double *sinreq = new double[deconvkernhalflen];
    double *cosreq = new double[deconvkernhalflen];
    double *sinout = new double[deconvkernhalflen];
    double *cosout = new double[deconvkernhalflen];

    // Calculate signal DC component
    const double dcsrc = dcdbl (dsrc);
    const double dcreq = dcdbl (dreq);
    // Calculate signal RMS power
    const double rmssrc = rmsdbl (dsrc, dcsrc);
    const double rmsreq = rmsdbl (dreq, dcreq);

    if (rmsreq == 0) {
        logger.err("Source signal is zero. Cannot make FIR");
        ret = -1;
        goto cleanup;
    }

    // Amplification Limiter
    {
        const double ratio = rmsreq / rmssrc;
        //  log_pri(LOG_CRIT, "DC: src %e, dreq %e", dcsrc, dcreq);
        //  log_pri(LOG_CRIT, "RMS: src %e, dreq %e", rmssrc, rmsreq);
        //  log_pri(LOG_CRIT, "Required power amplification: %e", ratio);
        const double maxspamp = ratio * FIR_MAX_AMPLIF;

        DFT (&dsrc[0], deconvkernlen, sinsrc, cossrc);
        DFT (&dreq[0], deconvkernlen, sinreq, cosreq);

        // Find power spectra and peaks
        RECT2POW (cossrc, sinsrc, deconvkernhalflen, powsrc);
        //  double srcspmax = maxdbl(powsrc, deconvkernhalflen);
        RECT2POW (cosreq, sinreq, deconvkernhalflen, powreq);
        //  double reqspmax = maxdbl(powreq, deconvkernhalflen);

        for (unsigned int i = 0; i < deconvkernhalflen; i++) {
            cosout[i] = 0;
            sinout[i] = 0;
            if (powsrc[i] < orig_thr)
                continue;
            if (powreq[i] < req_thr)
                continue;
            Compdiv (cosreq[i], sinreq[i], cossrc[i], sinsrc[i], &cosout[i], &sinout[i]);
        }

        RECT2POW (cosout, sinout, deconvkernhalflen, powout);
        // Limiter

        for (unsigned int i = 0; i < deconvkernhalflen; i++) {
            if (powout[i] > maxspamp) {
                //      log_pri(LOG_CRIT, "SRC: %2d: c %e, s %e:  a %e, ph %e", i, cossrc[i], sinsrc[i], powsrc[i], atan2(cossrc[i], sinsrc[i]));
                //      log_pri(LOG_CRIT, "REQ: %2d: c %e, s %e:  a %e, ph %e", i, cosreq[i], sinreq[i], powreq[i], atan2(cosreq[i], sinreq[i]));
                //      log_pri(LOG_CRIT, "TGT: %2d: c %e, s %e:  a %e, ph %e", i, cosout[i], sinout[i], powout[i], atan2(cosout[i], sinout[i]));
                logger.warn(QString("Spectral Amplification too big for freq %1, DFT size %2: reduced to %3").arg(i).arg(deconvkernlen).arg(maxspamp));
                ret++;
                const double k = powout[i]/maxspamp;
                cosout[i] /= k;
                sinout[i] /= k;
            }
        }
    }

    /*
  for(i=0;i<deconvkernhalflen;i++) {
      log_pri(LOG_CRIT, "%2d\t%e\t%e\t%e", i, hypot(cossrc[i], sinsrc[i]), hypot(cosreq[i], sinreq[i]), hypot(cosout[i], sinout[i]));
  }
  for(i=0;i<deconvkernhalflen;i++) {
      log_pri(LOG_CRIT, "SRC: %2d: c %e, s %e:  a %e, ph %e", i, cossrc[i], sinsrc[i], hypot(cossrc[i], sinsrc[i]), atan2(cossrc[i], sinsrc[i]));
  }
  for(i=0;i<deconvkernhalflen;i++) {
      log_pri(LOG_CRIT, "REQ: %2d: c %e, s %e:  a %e, ph %e", i, cosreq[i], sinreq[i], hypot(cosreq[i], sinreq[i]), atan2(cosreq[i], sinreq[i]));
  }
  for(i=0;i<deconvkernhalflen;i++) {
      log_pri(LOG_CRIT, "TGT: %2d: c %e, s %e:  a %e, ph %e", i, cosout[i], sinout[i], hypot(cosout[i], sinout[i]), atan2(cosout[i], sinout[i]));
  }
*/
    IDFT (dout, deconvkernlen, sinout, cosout);

    for (unsigned int i = 0; i < deconvkernlen; i++) {
        engine->coef[i] = dout[(i + deconvkernhalflen) % deconvkernlen];
    }

    engine->klen = deconvkernlen;
    engine->offs = deconvkernhalflen;
    /*
  printf("Deconvolution FIR: %d taps: {", engine->klen);
  for(i=0;i<engine->klen;i++) {
    printf(" %d", engine->coef[i]);
    if ((i+1)<engine->klen) printf(",");
  }
  printf("}\n");
*/
    cleanup:
        delete [] dout;
    delete [] powsrc;
    delete [] powreq;
    delete [] powout;
    delete [] sinsrc;
    delete [] cossrc;
    delete [] sinreq;
    delete [] cosreq;
    delete [] sinout;
    delete [] cosout;
    return (ret);
}

void fir_convolve_s (const intfir_t * fir, const short int *in, unsigned int inlen, int *out, unsigned int outlen)
{
    assert(fir);
    assert(in);
    assert(out);
    assert(inlen > 0);
    assert(outlen >= inlen);
    if (!in || !out || (inlen <= 0) || (outlen < inlen)) {
        logger.err("fir_convolve(): Bad input parameters");
        return;
    }
    if (fir_valid (fir)) {
        unsigned int workin, workout;
        workin = inlen;
        workout = inlen + fir->klen - 1;
        if (workin > outlen) {
            logger.err(QString("workin=%1 > outlen=%2").arg(workin).arg(outlen));
            workin = outlen - fir->klen;
        }
        if (workout > outlen) {
            logger.err(QString("workout=%1 > outlen=%2").arg(workout).arg(outlen));
            workout = outlen;
        }
        if ((workin <= 0) || (workout <= 0)) {
            logger.err("fir_convolve(): Bad input parameters");
            return;
        }
        /*
                printf("FIR[%d]: ", fir->klen);
                for (int i=0;i<fir->klen;i++) printf("%d, ", fir->coef[i]);
                printf("\n");
                printf("INP[%d]: ", workin);
                for (int i=0;i<workin;i++) printf("%d, ", fir->coef[i]);
                printf("\n");
*/
        if (fir->klen > 0)
            convolve_i_s_i (&fir->coef[0], fir->klen, in, workin, out);
        else {
            //      log_pri(LOG_CRIT, "NULL FIR - copying data");
            for (unsigned int i = 0; i < workin; i++)
                out[i] = in[i] * ((1 << fir->bits) - 1);
        }
        /*
                printf("OUT[%d]: ", workout);
                for (int i=0;i<workout;i++) printf("%d, ", out[i]);
                printf("\n");
*/
    } else {
        logger.err("fir_convolve(): Invalid FIR - copying data");
        for (unsigned int i = 0; i < inlen; i++)
            out[i] = in[i] * ((1 << fir->bits) - 1);
        for (unsigned int i = inlen; i < outlen; i++)
            out[i] = 0;
    }
}

/*
// same, with data type: double
void fir_convolve_d (const intfir_t * fir, const double *in, unsigned int inlen, double *out, unsigned int outlen)
{
    assert(fir);
    assert(in);
    assert(out);
    assert(inlen > 0);
    assert(outlen >= inlen);
  if (!in || !out || (inlen <= 0) || (outlen < inlen)) {
    logger.err("fir_convolve(): Bad input parameters");
    return;
  }
  if (fir_valid (fir)) {
    unsigned int workin, workout;
    workin = inlen;
    workout = inlen + fir->klen;
    if (workin > outlen) {
      logger.err(QString("workin=%1 > outlen=%2").arg(workin).arg(outlen));
      workin = outlen - fir->klen;
    }
    if (workout > outlen) {
      logger.err(QString("workout=%1 > outlen=%2").arg(workout).arg(outlen));
      workout = outlen;
    }
    if ((workin <= 0) || (workout <= 0)) {
      logger.err("fir_convolve(): Bad input parameters");
      return;
    }
    if (fir->klen > 0)
      convolve_id (fir->coef, fir->klen, in, workin, out);
    else {
//      log_pri(LOG_CRIT, "NULL FIR - copying data");
      for (unsigned int i = 0; i < workin; i++)
        out[i] = in[i] * ((1 << fir->bits) - 1);
    }
  } else {
    logger.err("fir_convolve_d(): Invalid FIR - copying data");
    for (unsigned int i = 0; i < inlen; i++)
      out[i] = in[i] * ((1 << fir->bits) - 1);
    for (unsigned int i = inlen; i < outlen; i++)
      out[i] = 0;
  }
}
*/
