
#include "Jtag_Impl_CCBPCIe.h"

#include <iostream>
#include <iomanip>
#include <cassert>

#include "exc.h"
#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"

#undef foreach
#include "stl_ext.h"

namespace mlink {

    const unsigned int CCBPCIe_BIT_TDI = 0x00000001;
    const unsigned int CCBPCIe_BIT_TMS = 0x00000002;
    const unsigned int CCBPCIe_BIT_TDO = 0x00000001;

    Jtag_Impl_CCBPCIe::~Jtag_Impl_CCBPCIe()
    {
        printStats();
    }

    void Jtag_Impl_CCBPCIe::clearStats()
    {
        count_pkt = 0;
        count_io = 0;
        count_io_b.clear();
        count_io_b32.clear();
    }

    void Jtag_Impl_CCBPCIe::printStats() const
    {
        std::cout << "JTAG packets: " << std::dec << count_pkt << std::endl;
        std::cout << "JTAG io:      " << count_io << std::endl;
        foreach_const(count_io_b, it) {
            std::cout << "JTAG io_b["<<(*it).first<<"]:    " << (*it).second << std::endl;
        }
        foreach_const(count_io_b32, it) {
            std::cout << "JTAG io_b32["<<(*it).first<<"]:    " << (*it).second << std::endl;
        }
    }

    RegIoPacket Jtag_Impl_CCBPCIe::packetIo(const RegIoPacket &sendPkt)
    {
        count_pkt++;
        RegIoPacket pkt(sendPkt);
        pkt.src = 0;
        pkt.dst = p_destAddress;
        try {
            link.rawSend(pkt);
            const PacketList &pl = link.getCtrlPackets(1, 1.0);
            if (pl.size() != 1 || pl.front().data.size() != pkt.data.size()) {
                pl.front().dump();
                throw exc_hwif("bad jtag reply size");
            }
            return pl.front();
        }
        catch (exc &e) {
            std::cerr << e.what() << std::endl;
            throw;
        }
    }

    bool Jtag_Impl_CCBPCIe::jtag_io (bool tms, bool tdi)
    {
        //  addTraceComment("// jtag_io\n");
        addTracePoint(tms, tdi);
        count_io++;
        RegIoPacket sendPkt;
        sendPkt.type = FRAME_TYPE_JTAG_SINGLE;
        sendPkt.data.push_back((tms?CCBPCIe_BIT_TMS:0) | (tdi?CCBPCIe_BIT_TDI:0));
        const RegIoPacket &recvPkt = packetIo(sendPkt);
        bool tdo = (recvPkt.data[0] & CCBPCIe_BIT_TDO) ? 1:0;
        return tdo;
    }

    std::vector<bool> Jtag_Impl_CCBPCIe::jtag_io_b(const std::vector<bool> &btms, const std::vector<bool> &btdi)
    {
        assert(btms.size() == btdi.size());
        size_t count = btms.size();
        std::vector<bool> btdo(count);
        //  addTraceComment("// jtag_io_b\n");
        count_io_b[count]++;
        RegIoPacket sendPkt;
        sendPkt.type = FRAME_TYPE_JTAG_SINGLE;
        size_t bitsSent = 0;
        while(bitsSent < count) {
            size_t bitsLeft = count - bitsSent;
            size_t sendCount = std::min(bitsLeft, (size_t)ML_MAX_FRAME_DATA_WORDS);
            sendPkt.data.clear();
            sendPkt.data.resize(sendCount);
            for(size_t i=0; i<sendCount; i++) {
                bool tdi = btdi[bitsSent+i];
                bool tms = btms[bitsSent+i];
                addTracePoint(tms, tdi);
                sendPkt.data[i] = ((tms?CCBPCIe_BIT_TMS:0) | (tdi?CCBPCIe_BIT_TDI:0));
            }
            const RegIoPacket &recvPkt = packetIo(sendPkt);
            for(size_t i=0; i<sendCount; i++)
                btdo[bitsSent+i] = (recvPkt.data[i] & CCBPCIe_BIT_TDO) ? 1:0;
            bitsSent += sendCount;
        }
        return btdo;
    }

    void Jtag_Impl_CCBPCIe::jtag_io_b32_blk (unsigned int count, const unsigned int *tdi32Data, unsigned int *tdo32Data)
    {
        //  addTraceComment("// jtag_io_b32_blk\n");
        count_io_b32[count]++;
        RegIoPacket sendPkt;
        sendPkt.type = FRAME_TYPE_JTAG_BLOCK;
        size_t wordsSent = 0;
        while(wordsSent < count) {
            size_t wordsLeft = count - wordsSent;
            size_t sendCount = std::min(wordsLeft, (size_t)ML_MAX_FRAME_DATA_WORDS);
            sendPkt.data.clear();
            sendPkt.data.resize(sendCount);
            for(size_t i=0; i<sendCount; i++) {
                unsigned int tdi32 = tdi32Data?tdi32Data[wordsSent+i]:0;
                for(int j=0; j<32; j++) {
                    addTracePoint(0, tdi32 & (1<<j));
                }
                sendPkt.data[i] = tdi32;
            }
            const RegIoPacket &recvPkt = packetIo(sendPkt);
            if (tdo32Data) {
                for(size_t i=0; i<sendCount; i++)
                    tdo32Data[wordsSent+i] = recvPkt.data[i];
            }
            wordsSent += sendCount;
        }
    }

};
