
#include "jtag_engine.h"

#include <QString>
#include <QDebug>
#include <cstdio>
#include <cerrno>
#include <cstdarg>

#include "colorlog/log.h"

#include "jtag_impl.h"
#include "mlink_bitop.h"
//#include "timer.h"

#include "jtag_state.h"

namespace mlink {

    using colorlog::debug_level_ge;

    JTAG_Engine::JTAG_Engine(Jtag_Impl *p)
    {
        impl = p;
        jtagState = JS_TEST_LOGIC_RESET;
        irStopState = JS_RUN_TEST_IDLE;
    }

    void JTAG_Engine::_internal_jtag_reset () // reset to the Test-Logic-Reset state
    {
        // set TMS high, TDI doesn't matter, clock them out five times
        std::vector<bool> btms(5, 1);
        std::vector<bool> btdi(5, 0);
        impl->jtag_io_b(btms, btdi);
        jtagState = JS_TEST_LOGIC_RESET;

        TRACE(8, QString("JTAG TEST LOGIC RESET"));
    }

    bool JTAG_Engine::_perform_scan(size_t count, const char *tdiData, char *tdoData)
    {
        bool sendBit;

        if (count==0) return true;

        if (tdoData)
            for(unsigned int n=0; n < (count>>3); n++)
                tdoData[n >> 3] =0;

        const unsigned int *tdi32Data = 0;
        unsigned int *tdo32Data = 0;
        if (tdiData)
            tdi32Data = (const unsigned int *)tdiData;
        if (tdoData)
            tdo32Data = (unsigned int *)tdoData;


        //
        size_t blen = ((int)count - 1) / 32; // count-1, don't send last bit in b32 instruction: TMS=1
        //  if (blen<0) blen=0;

        // optimize: large counts
        enum { J32_BLOCK_XFER_MAX_SIZE = 100 }; // 3200 bits

        if (blen > J32_BLOCK_XFER_MAX_SIZE) {
            qWarning() << QString("FIXME: bit count exceeds J32_BLOCK_XFER_MAX_SIZE=%1").arg(J32_BLOCK_XFER_MAX_SIZE);
            return(false);
        }

        unsigned int tdibuf[J32_BLOCK_XFER_MAX_SIZE];
        if (blen > J32_BLOCK_XFER_MAX_SIZE) {
        }
        for(size_t n=0; n<blen; n++) {
            if (tdi32Data) tdibuf[n] = tdi32Data[n];
            else tdibuf[n] = 0;
        }

        unsigned int tdobuf[J32_BLOCK_XFER_MAX_SIZE];
        impl->jtag_io_b32_blk(blen, tdibuf, tdobuf);

        for(size_t n=0; n<blen; n++) {
            if (tdo32Data) {
                tdo32Data[n] = tdobuf[n];
            }
        }


        const size_t start_count = blen*32;
        const size_t remain_count = count-start_count; // okay
        std::vector<bool> btms(remain_count);
        std::vector<bool> btdi(remain_count);

        for(size_t i=start_count; i<count; i++)  // shift through the data register
        {
            bool tmsState = (i+1 == count); // TMS = 1 if this is the last bit to be sent for now, otherwise 0
            if (tdiData)                /* not NULL */
                sendBit = tdiData[i >> 3] & (1 << (i & 7)) ? true : false;  // read a bit from the buffer
            else
                sendBit = false;
            btms[i-start_count] = tmsState;
            btdi[i-start_count] = sendBit;
        }                             // when done we will be in JS_EXIT1_DR

        const std::vector<bool> &btdo = impl->jtag_io_b(btms, btdi);

        if (tdoData)
            for(size_t p=0; p < remain_count; p++) {
            size_t tdo_p = start_count + p;
            if (btdo[p])
                tdoData[tdo_p >> 3] |= (1 << (p & 7));
            else
                tdoData[tdo_p >> 3] &= ~(1 << (p & 7));
        }
        return true;
    }

    // count is number of BYTES available to transfer (scanCount is number of bits to actually transfer)
    // data is shifted lsb first
    //
    bool JTAG_Engine::_internal_jtag_drscan (const unsigned int count, const char *tdiData, char *tdoData)
    {
        std::vector<bool> btms;
        switch (jtagState)            // move to JS_SHIFT_DR state
        {
        case JS_TEST_LOGIC_RESET:    // start in JS_TEST_LOGIC_RESET
            btms.push_back(0);             // move to IDLE, fall into the idle case

        case JS_RUN_TEST_IDLE: {       // start in JS_RUN_TEST_IDLE
                btms.push_back(1);             // move to JS_SELECT_DR_SCAN
                btms.push_back(0);             // move to JS_CAPTURE_DR
                btms.push_back(0);             // move to JS_SHIFT_DR
                break;
            }
        case JS_PAUSE_DR: {           // start in JS_PAUSE_DR
                btms.push_back(1);             // move to JS_EXIT2_DR
                btms.push_back(0);              // move to JS_SHIFT_DR
                break;
            }
        case JS_PAUSE_IR: {            // start in JS_PAUSE_IR
                btms.push_back(1);              // move to JS_EXIT2_IR
                btms.push_back(1);              // move to JS_UPDATE_IR
                btms.push_back(1);              // move to JS_SELECT_DR_SCAN
                btms.push_back(0);              // move to JS_CAPTURE_DR
                btms.push_back(0);              // move to JS_SHIFT_DR
                break;
            }
        default:
            _internal_jtag_reset ();
            return (false);             // not in a stable state, reset and return fail
        }

        impl->jtag_io_b(btms, btms);

        _perform_scan(count, tdiData, tdoData);

        btms.clear();
        btms.push_back(1);           // move to JS_UPDATE_DR
        btms.push_back(0);            // move to JS_RUN_TEST_IDLE
        impl->jtag_io_b(btms, btms);

        {
            if (debug_level_ge(8)) {
                if (count > 32) {
                    printf ("DR scan TDI[%u]: \n", count);
                    bitprint_groups (tdiData, count, 32);
                    printf ("DR scan TDO[%u]: \n", count);
                    bitprint_groups (tdoData, count, 32);
                } else {
                    printf ("DR scan TDI[%u]: ", count);
                    bitprint (tdiData, count);
                    printf ("DR scan TDO[%u]: ", count);
                    bitprint (tdoData, count);
                }
            }
        }
        return (true);                // completed okay
    }


    bool JTAG_Engine::_internal_jtag_irscan (unsigned int count, const char *tdiData, char *tdoData)
    {
        std::vector<bool> btms;
        switch (jtagState)            // move to JS_SHIFT_DR state
        {
        case JS_TEST_LOGIC_RESET:    // start in JS_TEST_LOGIC_RESET
            btms.push_back(0);             // move to IDLE, fall into the idle case

        case JS_RUN_TEST_IDLE:       // start in JS_RUN_TEST_IDLE
            btms.push_back(1);             // move to JS_SELECT_DR_SCAN
            btms.push_back(1);             // move to JS_SELECT_IR_SCAN
            btms.push_back(0);             // move to JS_CAPTURE_IR (instruction register is parallel loaded into the shift register, 01 into lsb's)
            btms.push_back(0);             // move to JS_SHIFT_IR
            break;

        case JS_PAUSE_DR:            // start in JS_PAUSE_DR
            btms.push_back(1);             // move to JS_EXIT2_DR
            btms.push_back(1);             // move to JS_UPDATE_DR
            btms.push_back(1);             // move to JS_SELECT_DR_SCAN
            btms.push_back(1);             // move to JS_SELECT_IR_SCAN
            btms.push_back(0);             // move to JS_CAPTURE_IR
            btms.push_back(0);             // move to JS_SHIFT_IR
            break;

        case JS_PAUSE_IR:            // start in JS_PAUSE_IR
            btms.push_back(1);             // move to JS_EXIT2_IR
            btms.push_back(0);             // move to JS_SHIFT_IR
            break;

        default:
            _internal_jtag_reset ();
            return (false);             // not in a stable state, reset and return fail
        }

        impl->jtag_io_b(btms, btms);

        _perform_scan(count, tdiData, tdoData);

        btms.clear();
        switch (irStopState) {
        case JS_RUN_TEST_IDLE:
            btms.push_back(1);           // move to JS_UPDATE_IR
            btms.push_back(0);           // move to JS_RUN_TEST_IDLE
            break;

        case JS_PAUSE_IR:
            btms.push_back(0);           // move to JS_PAUSE_IR
            break;

        case JS_TEST_LOGIC_RESET:
        default:
            btms.push_back(1);           // move to JS_UPDATE_IR
            btms.push_back(1);           // move to JS_SELECT_DR_SCAN
            btms.push_back(1);           // move to JS_SELECT_IR_SCAN
            btms.push_back(1);           // move to JS_TEST_LOGIC_RESET
            break;
        }
        impl->jtag_io_b(btms, btms);

        {
            if (debug_level_ge(8)) {
                printf ("IR scan TDI: ");
                bitprint ((void *) tdiData, count);
                printf ("IR scan TDO: ");
                bitprint ((void *) tdoData, count);
            }
        }
        return true;                // completed okay
    }

    int JTAG_Engine::jtag_reset ()               // resets TAP controller, clears pending scan counters
    {
        _internal_jtag_reset ();      // reset the interface
        //  irStopState = drStopState = JS_RUN_TEST_IDLE; // default end state is idle
        irStopState = JS_RUN_TEST_IDLE; // default end state is idle
        return 0;
    }

    void JTAG_Engine::jtag_pause(unsigned int count)
    {
        std::vector<bool> btms(count, 0);
        impl->jtag_io_b(btms, btms);
    }

    int JTAG_Engine::jtag_drscan (unsigned int count, const void *tdiData, void *tdoData)
    {
        if (_internal_jtag_drscan (count, (char *) tdiData, (char *) tdoData) != true)
            return (-1);

        return 0;
    }

    int JTAG_Engine::jtag_drscan_chk (unsigned int count, const void *tdiData, void *tdoData)
    {

        int ret = 0;

        // shift and update DR
        if ((ret = jtag_drscan (count, tdiData, tdoData))) {
            return (ret);
        }
        // check if DR has changed
        if (bitcmp (tdiData, tdoData, count)) {
            // shift and update DR again
            if ((ret = jtag_drscan (count, tdiData, tdoData))) {
                return (ret);
            }
            // if DR has changed again, it cannot be loaded, fail
            if (bitcmp (tdiData, tdoData, count)) {
                qWarning() << "jtag_drscan_chk: data still changing";
                if (debug_level_ge(1)) {
                    bitprintdiff(tdiData, tdoData, count);
                }
                return (-EIO);
            } else {
                return 1;
            }
        } else {
            return 0;
        }
        return 0;
    }

    int JTAG_Engine::jtag_irscan (unsigned int count, const void *tdiData, void *tdoData)
    {
        //  irscanCount = count;
        if (_internal_jtag_irscan (count, (const char*)tdiData, (char*)tdoData) != true)
            return (-1);

        return 0;
    }

    void JTAG_Engine::TRACE(int level, const QString &text) const  // used only with -DDEBUG
    {
        if (debug_level_ge(level))
            qDebug() << "JTAG: " << text;
        //  mod->debug_mod (level, "%s%s", "JTAG ", buf);
    }
};
