
#include "regiomlink.h"

#include <QDebug>
#include <string>
#include <sstream>
#include <iomanip>
#include <cstring>

#include "mlink.h"
#include "frame_hw.h"
#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"
#include "exc.h"
#include "util/gettimeofday.h"
#include "regio/TextFormatter.h"

namespace mlink {

    const int send_timeout = 1000; // use -1 to disable wait, return immediately
    const int receive_timeout = 1000; // ms

    RegIOMLink::RegIOMLink(const std::string &name)
        : RegIOBase(), txPacketSequenceNumber(0)
    {
        bufSizeLimit = 10000; // maximum number of packets
        devcaps = 0;

        if (name.empty())
            setDevName(defaultDevName());
        else
            setDevName(name);
        devOpen();
    }

    RegIOMLink::~RegIOMLink()
    {
        devClose();
    }

    void RegIOMLink::devOpen()
    {
        int rc = 0;
        {
            rc = ml_open(devName.c_str());
        }
        if (rc) {
            throw exc_hwif(std::string("Device open failed: ")+ml_strerror(-rc));
        }
        ml_get_driver_version(&driverVersion);
        ml_get_driver_revision(&driverRevision);
        mlink_caps_t caps;
        memset(&caps, 0, sizeof(caps));
        ml_get_caps (&caps);
        devcaps = caps.all;
    }

    void RegIOMLink::devClose()
    {
        int rc = 0;
        {
            rc = ml_close();
        }
        if (rc) {
            throw exc_hwif(std::string("Device close failed: ")+ml_strerror(-rc));
        }
    }

    std::string RegIOMLink::defaultDevName()
    {
        return std::string("/dev/")+std::string(ML_DEVICE_NAME)+std::string("0");
    }

    void RegIOMLink::reset()
    {
        RegIOBase::reset();
        txPacketSequenceNumber = 0;
        int rc = 0;
        {
            rc = ml_reset();
        }
        if (rc) {
            throw exc_hwif(std::string("Device reset failed: ")+ml_strerror(-rc));
        }
    }

    //void RegIOMLink::setLocalJtagMode(bool state)
    //{
    //    int rc = 0;
    //    {
    //        rc = ml_set_jtag_mode(state);
    //    }
    //    if (rc) {
    //        throw exc_hwif(std::string("Device jtag mode set failed: ")+ml_strerror(-rc));
    //    }
    //}

    ///// @warning data.seq is overwritten
    void RegIOMLink::rawSend(const RegIoPacket &data)
    {
        //    RegIoPacket data(dataIn);
        //    data.seq = nextTxPacketSequenceNumber();
        // send packet
        int rc = 0;
        {
            rc = ml_send_data_frame (reinterpret_cast<const int *>(&data.data[0]), data.data.size(), data.type, data.src, data.dst, data.seq, send_timeout);
            //  printf("tx: size=%d\n", data.data.size());
            //        qDebug() << "RegIOMLink::rawSend() " << data;
        }
        if (rc) {
            QString str = QString("Failed frame send: %1").arg(ml_strerror(-rc));
            throw exc_hwif(str.toStdString());
        }
    }

    RawPacketList RegIOMLink::getRxPackets(double timeout)
    {
        RawPacketList rxBuf;
        rawRxBuf.clear();
        rawRxBuf.resize(ML_RX_BUFFER_WORDS, 0);
        const size_t bufSize = sizeof(rawRxBuf[0]) * rawRxBuf.size();
        //  memset(&rawRxBuf[0], 0, bufSize);
        // initialize frameset
        struct frameset_t frameset;// frameset.maxdesc = ML_FRAMESET_NUMDESC;
        frameset.count = 0;
        int ret = ml_recv_data_frame_multi(&rawRxBuf[0], bufSize, &frameset, int(timeout*1000));
//        qDebug() << "ml_recv_data_frame_multi returned " << ret;
//        std::cerr.flush();
        if (ret != 0)
            return rxBuf;
        // copy frame data
        for(size_t i=0; i<frameset.count; i++) {
            const struct framedesc_t *fdesc = &frameset.fdesc[i];
            const size_t datawords = fdesc->len;// - (ML_FRAME_HEADER_WORDS+ML_FRAME_TRAILER_WORDS);
            QByteArray rawPacket(reinterpret_cast<const char *>(fdesc->rawdata), 4*datawords);
//            qDebug() << "rawPacket size " << rawPacket.size();
            //            const int *frame_data = (const int *) &fdesc->rawdata[ML_FRAME_HEADER_WORDS];
            //            const frame_hdr_t *fh = (const frame_hdr_t *) fdesc->rawdata;
            //            PacketData data;
            //            data.assign(&frame_data[0], &frame_data[datawords]);
            //            RegIoPacket pkt(fh->type, fh->seq, fh->src, fh->dst, data);
            //            rxBuf.push_back(pkt);
            rxBuf.push_back(rawPacket);
        }
        return rxBuf;
    }

    double RegIOMLink::getDeviceThroughput() const
    {
        return 160e6;
    }

    unsigned long RegIOMLink::getMaxDataSize() const
    {
        return ML_MAX_FRAME_DATA_WORDS;
    }

};
