
#include "deviceemu.h"

#include <QDebug>
#include <stdexcept>
#include <sstream>

#include "mlink/frame_types.h"

namespace mlink {

    DeviceEmu::DeviceEmu()
    {
        reset();
    }

    DeviceEmu::~DeviceEmu()
    {
    }

    void DeviceEmu::reset()
    {
        txSeq = 0;
        txBuf.clear();
        rxBuf.clear();
        regs16.clear();
        std::fill(ram32.begin(), ram32.end(), 0); // it is a vector. zero fill, not resize.
    }

    void DeviceEmu::processPackets()
    {
        for(PacketList::const_iterator it = rxBuf.begin(); it != rxBuf.end(); ++it) {
            const RegIoPacket &rx = (*it);
            //    log_info("DeviceEmu: RX type %04X, seq #%d, %d -> %d, size %zu", rx.type, rx.seq, rx.src, rx.dst, rx.data.size());
            //	rx.dump();
            int src = rx.src;
            int dst = rx.dst;
            int type = rx.type;
            //	rx.dump();
            switch (type) {
            case FRAME_TYPE_CTRL_REQ: {
                    RegIoPacket tx;
                    tx.src = dst;
                    tx.dst = src;
                    tx.type = FRAME_TYPE_CTRL_ACK;
                    const std::vector<quint32> &rxdata = rx.data;
                    for(std::vector<quint32>::const_iterator jt = rxdata.begin(); jt != rxdata.end(); ++jt) {
                        int code = (*jt);
                        bool write = (code & 0x80000000) == 0;
                        unsigned int addr = (code & 0x7FFF0000) >> 16;
                        unsigned int data = code & 0xFFFF;
                        if (write) {
                            writeReg(addr, data);
                        } else {
                            readReg(addr);
                            code &= 0xFFFF0000;
                            code |= (readReg(addr) & 0x0000FFFF);
                        }
                        tx.data.push_back(code);
                    }
                    tx.seq = txSeq++;
                    //        log_info("DeviceEmu: TX type %04X, seq #%d, %d -> %d, size %zu", tx.type, tx.seq, tx.src, tx.dst, tx.data.size());
                    //		tx.dump();
                    txBuf.push_back(tx);
                    break;
                }
            case FRAME_TYPE_RAM_REQ: {
                    RegIoPacket tx;
                    tx.src = dst;
                    tx.dst = src;
                    tx.type = FRAME_TYPE_RAM_ACK;
                    const PacketData &rxdata = rx.data;
                    if (rxdata.size() < 1) {
                        throw std::runtime_error("Bad RAM I/O request: zero data size");
                    }
                    int code = rxdata.at(0);
                    bool write = (code & 0x80000000) == 0;
                    unsigned int blockSize = (code >> 22) & 0x1FF;
                    unsigned int startAddr = (code & 0x3FFFFF);
                    //        tx.data[0] = ( 0x80000000 |(| ((startWord+totalRead)&0x3FFFFF) )
                    if (write && (rxdata.size() != 1+blockSize)) {
                        std::ostringstream os;
                        os << "Bad RAM write request: supplied data size " << rxdata.size()-1 << ", requested " << blockSize;
                        throw std::runtime_error(os.str());
                    }
                    tx.data.push_back(code);
                    if (write) {
                        std::vector<quint32> data;
                        data.assign(++rxdata.begin(), rxdata.end());
                        writeRam(startAddr, data);
                        // mirror data back
                        tx.data.resize(1+blockSize);
                        std::copy(data.begin(), data.end(), ++tx.data.begin());
                    } else {
                        const std::vector<quint32> &data = readRam(startAddr, blockSize);
                        tx.data.resize(1+blockSize);
                        std::copy(data.begin(), data.end(), ++tx.data.begin());
                    }
                    tx.seq = txSeq++;
                    //        log_info("DeviceEmu: TX type %04X, seq #%d, %d -> %d, size %zu", tx.type, tx.seq, tx.src, tx.dst, tx.data.size());
                    //		tx.dump();
                    txBuf.push_back(tx);
                    break;
                }


            default:
                qWarning() << QString("DeviceEmu: Unhandled packet type %1").arg(type, 4, 16, QChar('0'));
            }
        }
        //  log_info("DeviceEmu: processed RX %zu -> TX %zu", rxBuf.size(), txBuf.size());
        rxBuf.clear();
    }

    void DeviceEmu::writeReg(int addr, int data)
    {
        if (regs16ReadOnly.contains(addr)) {
            std::ostringstream os;
            os << "Write to read-only register 0x" << std::hex << addr;
            throw std::runtime_error(os.str());
            //    std::cerr << os.str() << std::endl;
        }
        regs16[addr] = data & 0xFFFF;
        //  log_info("DeviceEmu: Write [%04X] = %04X", addr, regs16[addr]);
    }

    int DeviceEmu::readReg(int addr) const
    {
        //  log_info("DeviceEmu: Read [%04X] = %04X", addr, regs16[addr]);
        if (regs16ReadOnly.contains(addr))
            return regs16ReadOnly[addr];
        else
            if (regs16.contains(addr))
                return regs16[addr];
        else
            return 0;
    }

    std::vector<quint32> DeviceEmu::readRam(unsigned int startAddr, unsigned int blockSize) const
    {
        //    range check
        if (startAddr+blockSize > ram32.size()) {
            std::ostringstream os;
            os << "Read access beyound end of RAM: " << startAddr+blockSize << ", available " << ram32.size();
            throw std::runtime_error(os.str());
        }
        std::vector<quint32> v;
        v.assign(ram32.begin()+startAddr, ram32.begin()+startAddr+blockSize);
        return v;
    }

    void DeviceEmu::writeRam(int startAddr, const std::vector<quint32> &data)
    {
        //    range check
        if (startAddr+data.size() > ram32.size()) {
            std::ostringstream os;
            os << "Write access beyound end of RAM: " << startAddr+data.size() << ", available " << ram32.size();
            throw std::runtime_error(os.str());
        }
        //    ram32.resize(startAddr+data.size());
        std::copy(data.begin(), data.end(), ram32.begin()+startAddr);
    }

    bool DeviceEmu::isPPAddressEnabled() const
    {
        return 0 == (readReg(0x40) & (1<<12));
    }

} // namespace mlink
