//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <qwt_plot_marker.h>

#include "ui_WaveformViewAdvancedUi.h"
#include "WaveformViewAdvanced.h"
#include "WaveformViewAdvanced_p.h"
#include "dsplib/dsplib.h"
#include "dsplib/Convolver.h"
#include "qxw/WaveformPlot.h"
#include "qlog/qlog.h"

WaveformViewAdvanced::WaveformViewAdvanced(QWidget *parent)
    : QWidget(parent),
      ui(new Ui::WaveformViewAdvancedUi),
      d(new WaveformViewAdvancedPrivate),
      currentTimeMarker(new QwtPlotMarker())
{
    ui->setupUi(this);
    //    currentTimeMarker->setLineStyle(QwtPlotMarker::VLine);
    //    currentTimeMarker->setLinePen(QPen(QColor(Qt::green)));
    //    currentTimeMarker->setTitle("Current Time");
}

WaveformViewAdvanced::~WaveformViewAdvanced()
{
    delete currentTimeMarker;
    delete d;
    delete ui;
}

void WaveformViewAdvanced::setModel(WaveformModel *p)
{
    d->model = p;
    connect(d->model, SIGNAL(dataChanged()), this, SLOT(update()));
    update();
}

void WaveformViewAdvanced::setAutoscale( bool b )
{
    setup.autoscale = b;
}

void WaveformViewAdvanced::setBipolarDisplay( bool b )
{
    setup.bipolarDisplay = b;
}

void WaveformViewAdvanced::setOffset( int d )
{
    setup.offset = d;
}

void WaveformViewAdvanced::update()
{
    if (!d->model)
        return;

    WaveformDataAll wda(d->model->get());
    if (d->offsetComp) {
        for (WaveformDataAll::iterator it = wda.begin(); it != wda.end(); ++it) {
            WaveformData &wd = (*it).second;
            for(size_t i=0; i<wd.p.size(); i++)
                wd.p[i].y -= wd.baseline;
        }
    }
    if (d->selectFilter) {
        for (WaveformDataAll::iterator it = wda.begin(); it != wda.end(); ++it) {
            WaveformData &wd = (*it).second;
            size_t len = wd.p.size();
            if (len<1) continue;
            size_t c = wd.c;
            if (c >= d->filter.size()) continue;
            const intfir_t &fir = d->filter[c];
            double step = wd.p[1].x - wd.p[0].x;
            double normfactor = 1. / (1 << fir.bits);
            std::vector<double> v(len);
            for(size_t i=0; i<len; i++)
                v[i] = wd.p[i].y;
            const std::vector<int> firCoef(&fir.coef[0], &fir.coef[fir.klen]);
            std::vector<double> u;
            if (firCoef.empty())
                u.assign(v.begin(), v.end());
            else {
                normfactor = 1. / (1 << fir.bits);
                u = Convolver::convolve(v, firCoef);
            }
            for(size_t i=0; i<len; i++) {
                wd.p[i].x -= fir.offs * step;
                wd.p[i].y = u[i] * normfactor;
            }
            //  		wd[n].p[i].y -= wd[n].baseline;
        }
    }
    if (d->enableFFT) {
        wda = d->waveformFilterFFT.filterAll(wda);
    }
    if (d->enableAverager) {
        d->averager.add(wda);
        wda = d->averager.get();
    }
    if (d->enableFFT) {
        for (WaveformDataAll::iterator it = wda.begin(); it != wda.end(); ++it) {
            WaveformData &wd = (*it).second;
            const double fullScaleSigned = d->fullScale[wd.c];
            const double fullScaleDb = val2db(fullScaleSigned);
            for(size_t i=0; i< wd.p.size(); i++) {
                wd.p[i].y = val2db(wd.p[i].y) - fullScaleDb;
            }
        }
    }
    ui->waveformPlot->setAllData(wda);
    QString yTitle;
    if (d->enableFFT) {
        yTitle = tr("Amplitude, dBFS");
    } else {
        yTitle = tr("Amplitude, bins");
    }
    QwtText yText(yTitle);
    yText.setFont(font());
    ui->waveformPlot->setAxisTitle(QwtPlot::yLeft, yText);
    ui->waveformPlot->replot();
}

void WaveformViewAdvanced::clear()
{
    d->averager.clear();
    ui->waveformPlot->clear();
    ui->waveformPlot->replot();
}


void WaveformViewAdvanced::on_checkBox_OffsetComp_toggled( bool value )
{
    d->offsetComp = value;
    d->averager.clear();
    update();
}

void WaveformViewAdvanced::on_checkBox_Filter_toggled( bool value )
{
    d->selectFilter = value;
    d->averager.clear();
    update();
}

void WaveformViewAdvanced::setFilterData( int channel, const intfir_t &fir )
{
    //    if (d->filter.size() < size_t(channel+1)) d->filter.resize(channel+1);
    d->filter[channel] = fir;
}

void WaveformViewAdvanced::setFullScale( int channel, double value )
{
    d->fullScale[channel] = value;
}

void WaveformViewAdvanced::on_checkBox_Averaging_toggled( bool value )
{
    d->enableAverager = value;
    d->averager.clear();
    update();
}

void WaveformViewAdvanced::on_checkBox_FFT_toggled( bool value )
{
    d->enableFFT = value;
    d->averager.clear();
    update();
}

WaveformPlot *WaveformViewAdvanced::getPlot()
{
    return ui->waveformPlot;
}

void WaveformViewAdvanced::setWaveformData( const WaveformDataAll &w )
{
    getPlot()->setAllData(w);
}

void WaveformViewAdvanced::setCurrentTimeMarker(double)
{
    //    currentTimeMarker->setXValue(v);
    //    if (! currentTimeMarker->plot())
    //        currentTimeMarker->attach(getPlot());
}

void WaveformViewAdvanced::setXMarkers(const std::vector<double> &v)
{
    getPlot()->setXMarkers(v);
}
